/*	$Kolabore: cache.cpp,v 1.11 2007/02/09 20:58:49 maxx Exp $	*/
/*
 * Copyright (c) 2006, 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include "compat/typesetup.h"

extern "C" {
#include <sys/types.h>

#include <pwd.h>
}

#include <ctime>

#include "cache.h"
#include "timekeeper.h"

using namespace std;

XSI_UserCache::XSI_UserCache(bool nolookup, uint32_t timeout_secs)
{
	hits = 0;
	misses = 0;
	failures = 0;
	timeouts = 0;
	no_lookups = nolookup;
	timeout = timeout_secs;
}

/*
 * Clear the cache and reset counters.
 */
void
XSI_UserCache::reset(void)
{
	cache.clear();
	hits = 0;
	misses = 0;
	failures = 0;
	timeouts = 0;
}

/*
 * Lookup UID by username, checking the cache first before querying the system
 * userdb. Returns the appropriate UID on success or UINT32_MAX on failure.
 */
uint32_t
XSI_UserCache::getUID(const char *user)
{
	vector<struct userdb_cache>::iterator it;
	struct userdb_cache new_entry;
	struct passwd  *pw;
	uint32_t	cur_time = XSI_TimeKeeper::get()->updateTimestampInt();
	bool		had_timeout = false;

	if (no_lookups)
		return (UINT32_MAX);

	for (it = cache.begin(); it != cache.end();) {
		if (cur_time && it->add_time < (cur_time - timeout)) {
			timeouts++;
			had_timeout = true;
			cache.erase(it);
			continue;
		}
		if (it->user == user) {
			hits++;
			return (it->uid);
		}
		it++;
	}

	if (!had_timeout)
		misses++;

	new_entry.user = user;
	sanitizeString(new_entry.user, true);
	if ((pw = getpwnam(user)) == NULL) {
		failures++;
		return (UINT32_MAX);
	}
	new_entry.uid = pw->pw_uid;
	new_entry.add_time = (uint32_t)time(NULL);
	cache.push_back(new_entry);

	return (new_entry.uid);
}

/*
 * Lookup username by UID, checking the cache first before querying the system
 * userdb. Returns the appropriate username on success or an empty string on
 * failure.
 */
string
XSI_UserCache::getUsername(uint32_t uid)
{
	vector<struct userdb_cache>::iterator it;
	struct userdb_cache new_entry;
	struct passwd  *pw;
	uint32_t	cur_time = XSI_TimeKeeper::get()->updateTimestampInt();
	bool		had_timeout = false;

	if (no_lookups)
		return (string(""));

	for (it = cache.begin(); it != cache.end();) {
		if (cur_time && it->add_time < (cur_time - timeout)) {
			timeouts++;
			had_timeout = true;
			cache.erase(it);
			continue;
		}
		if (it->uid == uid) {
			hits++;
			return (it->user.c_str());
		}
		it++;
	}

	if (!had_timeout)
		misses++;

	new_entry.uid = uid;
	if ((pw = getpwuid(uid)) == NULL) {
		failures++;
		return (string(""));
	}
	new_entry.user = pw->pw_name;
	sanitizeString(new_entry.user, true);
	new_entry.add_time = (uint32_t)time(NULL);
	cache.push_back(new_entry);

	return (new_entry.user);
}

unsigned long
XSI_UserCache::getHits(void)
{
	return (hits);
}

unsigned long
XSI_UserCache::getMisses(void)
{
	return (misses);
}

unsigned long
XSI_UserCache::getFailures(void)
{
	return (failures);
}

unsigned long
XSI_UserCache::getTimeouts(void)
{
	return (timeouts);
}
