/*	$Kolabore: cfg.cpp,v 1.42 2007/02/09 13:21:57 maxx Exp $	*/
/*
 * Copyright (c) 2006, 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "log.h"

extern "C" {
#include <sys/param.h>
#ifdef HAVE_LINUX_PARAM_H
#  include <linux/param.h>
#endif
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
}

#include <iostream>
#include <list>
#include <string>
#include <vector>

#include <cstdlib>

#include "cfg.h"

#ifndef CHROOT_DIR
#  define CHROOT_DIR		"/var/empty"
#endif
#ifndef PRIVDROP_USER
#  define PRIVDROP_USER		"nobody"
#endif
#ifndef PRIVDROP_GROUP
#  define PRIVDROP_GROUP	"nobody"
#endif

extern char    *optarg;
extern int	opterr;
extern int	optind;
extern int	optopt;
extern int	optreset;

using namespace std;

Config::Config(int argc, char **argv)
{
	int		ch, error;
	const char     *errStr;
	string		lAddr;
	struct addrinfo hints;
	string		listenService = DEFAULT_PORT;
	bool		aFlag = false, CFlag = false, iFlag = false,
			pFlag = false;

	bool log_syslog = false;

	GFlag = UFlag = false;

	update_interval = 2;

	compress_xml    = false;
	dont_detach     = false;
	format_output   = false;
	log_syslog      = false;
	no_names        = false;
	one_client_mode = false;
	one_shot        = false;
	show_pf         = false;
	show_raid       = false;

	while ((ch = getopt(argc, argv, "1a:C:cdFfG:hi:lNop:r:U:")) != -1) {
		switch (ch) {
		case '1':
			one_client_mode = true;
			xsi_options.push_back("one_client_mode");
			break;
		case 'a':
			if (aFlag) {
				cerr << "Error: multiple -a arguments given."
					<< endl;
				usage();
				exit(2);
			}
			lAddr = optarg;
			aFlag = true;
			break;
		case 'C':
			if (CFlag) {
				cerr << "Error: multiple -C arguments given."
					<< endl;
				usage();
				exit(2);
			}
			CFlag = true;
			chrootDir = optarg;
			if (chrootDir == "/") {
				cerr << "Error: the filesystem root is not a "
					"valid target to chroot into." << endl;
				exit(2);
			}
			break;
		case 'c':
			compress_xml = true;
			xsi_options.push_back("compress_output");
			break;
		case 'd':
			dont_detach = true;
			break;
		case 'F':
			show_pf = true;
			xsi_options.push_back("show_packet_filter");
			break;
		case 'f':
			format_output = true;
			xsi_options.push_back("format_output");
			break;
		case 'G':
			if (GFlag) {
				cerr << "Error: multiple -G arguments given."
					<< endl;
				usage();
				exit(2);
			}
			GFlag = true;
			unprivGroup = optarg;
			break;
		case 'h':
			usage();
			usageHelp();
			exit(0);
		case 'i':
			if (iFlag) {
				cerr << "Error: multiple -i arguments given."
					<< endl;
				usage();
				exit(2);
			}
			iFlag = true;
			update_interval = (int)strtonum(optarg, 1, 3600,
							&errStr);
			if (errStr) {
				cerr << "Update interval '" << optarg
					<< "' is " << errStr
					<< " (min 1, max 3600)" << endl;
				exit(2);
			}
			break;
		case 'l':
#ifdef HAVE_SYSLOG_SUPPORT
			log_syslog = true;
			break;
#else
			cerr << "Sorry, logging to syslog is not available on "
				"your system." << endl;
			exit(2);
#endif /* HAVE_SYSLOG_SUPPORT */
		case 'N':
			no_names = true;
			xsi_options.push_back("no_names");
			break;
		case 'o':
			one_shot = true;
			xsi_options.push_back("one_shot");
			break;
		case 'p':
			if (pFlag) {
				cerr << "Error: multiple -p arguments given."
					<< endl;
				usage();
				exit(2);
			}
			pFlag = true;
			listenService = optarg;
			break;
		case 'r':
			show_raid = true;
			{
				string tmp = optarg;
				raid_controllers.push_back(tmp);
				raid_controllers.sort();
				raid_controllers.unique();
			}
			xsi_options.push_back("show_raid");
			break;
		case 'U':
			if (UFlag) {
				cerr << "Error: multiple -U arguments given."
					<< endl;
				usage();
				exit(2);
			}
			UFlag = true;
			unprivUser = optarg;
			break;
		case '?':
			usage();
			exit(2);
		default:
			break;
		}
	}
	argc -= optind;
	argv += optind;

	if (xsi_options.empty())
		xsi_options.push_back("defaults");
	else {
		xsi_options.sort();
		xsi_options.unique();
	}

	memset(&hints, 0, sizeof(hints));
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;

	error = getaddrinfo(lAddr.empty() ? NULL : lAddr.c_str(),
			    listenService.c_str(), &hints, &listen_addr);
	if (error) {
		cerr << "address " << (lAddr.empty() ? "any" : lAddr)
			<< ", service/port " << listenService << ": "
			<< gai_strerror(error) << endl;
		exit(2);
	}

	XSI_Log::log()->init(log_syslog, dont_detach);
}

Config::~Config(void)
{
	freeaddrinfo(listen_addr);
}

void
Config::usage(void)
{
	extern char *__progname;

	cerr << "usage: "
		<< __progname
		<< " "
		<< "[-1cdFfhlNo] [-a address] [-C chrootdir] [-G group] [-i interval]\n"
		<< "           [-p port] [-r device] [-U user]"
		<< endl;
}

void
Config::usageHelp(void)
{
	cerr << "\n"
		"    -1              One-Client Mode\n"
		"    -a address      listen on address instead of all interfaces\n"
#if defined(INSANE_OS)
		"    -C chrootdir    (no-op on this platform)\n"
#else
		"    -C chrootdir    use chrootdir instead of " CHROOT_DIR "\n"
#endif  /* INSANE_OS */
		"    -c              compress output\n"
		"    -d              debug mode / do not detach\n"
		"    -F              include packet filter stats\n"
		"    -f              format output\n"
		"    -G group        drop group privileges to group instead of " PRIVDROP_GROUP "\n"
		"    -h              print this additional help and exit\n"
		"    -i interval     set update interval to interval seconds ([1..3600])\n"
		"    -l              enable logging to syslog"
#ifdef HAVE_SYSLOG_SUPPORT
		"\n"
#else
		" (N/A)\n"
#endif /* HAVE_SYSLOG_SUPPORT */
		"    -N              no usernames in output\n"
		"    -o              One-Shot Mode\n"
		"    -p port         TCP port/service to listen on (default: " DEFAULT_PORT ")\n"
#if defined(OPENBSD)
		"    -r device       RAID monitoring on controller device\n"
		"                    (may be supplied multiple times)\n"
#else
		"    -r device       RAID monitoring (N/A on this platform)\n"
#endif
		"    -U user         drop user privileges to user instead of " PRIVDROP_USER "\n"
		"\n"
		"See the xsi(8) manual for detailed information."
		<< endl;
}

bool
Config::oneClientMode(void)
{
	return (one_client_mode);
}

bool
Config::compressXml(void)
{
	return (compress_xml);
}

bool
Config::dontDetach(void)
{
	return (dont_detach);
}

bool
Config::formatOutput(void)
{
	return (format_output);
}

bool
Config::noNames(void)
{
	return (no_names);
}

bool
Config::oneShot(void)
{
	return (one_shot);
}

bool
Config::showRaid(void)
{
	return (show_raid);
}

bool
Config::showPacketFilter(void)
{
	return (show_pf);
}

bool
Config::userRequestsPrivdrop()
{
	return (GFlag || UFlag);
}

int
Config::getUpdateInterval(void)
{
	return (update_interval);
}

const char *
Config::getUpdateIntervalString(void)
{
	return (makeIntStr((uint32_t)update_interval));
}

struct addrinfo *
Config::getListenAddr(void)
{
	return (listen_addr);
}

list<string>
Config::getRaidControllers(void)
{
	return (raid_controllers);
}

list<string>
Config::getXSIOptionsList(void)
{
	return (xsi_options);
}

const char *
Config::getChrootDir(void)
{
	if (chrootDir.empty())
		return (CHROOT_DIR);
	else
		return (chrootDir.c_str());
}

const char *
Config::getPrivdropUser(void)
{
	if (unprivUser.empty())
		return (PRIVDROP_USER);
	else
		return (unprivUser.c_str());
}

const char *
Config::getPrivdropGroup(void)
{
	if (unprivGroup.empty()) {
		if (unprivUser.empty())
			return (PRIVDROP_GROUP);
		else
			return (NULL);	 // use passwd.pw_gid
	} else
		return (unprivGroup.c_str());
}
