/*	$Kolabore: darwin.cpp,v 1.17 2007/01/05 14:16:18 maxx Exp $	*/
/*
 * Copyright (c) 2006 Timothy Goh <timgoh0-at-gmail-dot-com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * darwin.cpp:
 * Gathering information and stats on Darwin OS 10.4 or newer.
 */

#include "log.h"

extern "C" {
	#include <sys/param.h>
	#include <sys/socket.h>
	#include <sys/ioctl.h>
	#include <sys/mount.h>
	
	#include <net/if.h>
	#include <net/route.h>
	#include <ifaddrs.h>
	
	#include <sys/sysctl.h>
	#include <CoreFoundation/CoreFoundation.h>
	#include <IOKit/IOKitLib.h>
	
	#include <mach/mach.h>
	#include <mach/machine.h>
	#include <mach/mach_host.h>
	#include <mach/host_info.h>
}

// C++ headers

#include "darwin.h"

using namespace std;

/*
 * XSI Constructor
 *
 * !! WARNING !!
 * !! Runs before dropping privileges when running as root
 * !! WARNING !!
 */
XSI::XSI(void)
{
	// initialization
}

XSI::~XSI(void)
{
	// cleanup
}

/*
 * Information read once on startup
 */
void
XSI::collectSysInfo(void)
{
	// list of XSI methods
	readSystemInfo();
	readSystemCPUFreq();
	readSystemNumCPUs();
	readSystemOSInfo();
}

/*
 * Stats to be updated regularly
 */
void
XSI::updateStats(void)
{
	// list of XSI update-methods
	updateCPUStats();
	updateVMStats();
	updateOSUptime();
	updateLoadAvg();
	updateNetStats();
	updateFSStats();

}

string
XSI::getSysctlString(int mib0, int mib1, int mib2, unsigned int len,
		     bool must_exist, const char *caller)
{
	int	mib[] = { mib0, mib1, mib2 };
	size_t	siz;
	string	ret;
	char   *buf;

	if (sysctl(mib, len, NULL, &siz, NULL, 0) == -1) {
		if (errno != EOPNOTSUPP || must_exist)
			XSI_Log::log()->err(caller, errno, "sysctl");
		return ("");
	}

	buf = new char[siz];
	if (buf == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	if (sysctl(mib, len, buf, &siz, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");
	ret = buf;
	delete[] buf;
	return (ret);
}

void
XSI::readSystemInfo(void)
{
	string buf;

	buf = getSysctlString(CTL_HW, HW_MODEL, 0, 2, false, __func__);
	if (!buf.empty())
		system.setModel(buf.c_str());
}

void
XSI::readSystemCPUFreq(void)
{
	int mib[] = { CTL_HW, HW_CPU_FREQ };
	int cpu_freq;
	size_t siz = sizeof(cpu_freq);

	if (sysctl(mib, 2, &cpu_freq, &siz, NULL, 0) == -1) {
		if (errno != EOPNOTSUPP)
			XSI_Log::log()->err(__func__, errno, "sysctl");
		return;
	}
	system.setCPUFreq((int32_t)((double)cpu_freq / 1000000.0));
}

void
XSI::readSystemNumCPUs(void)
{
	int mib[] = { CTL_HW, HW_NCPU };
	int ncpu;
	size_t siz = sizeof(ncpu);

	if (sysctl(mib, 2, &ncpu, &siz, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");
	system.setNumCPUs(ncpu);
	num_cpu = ncpu;
}

void
XSI::readSystemOSInfo(void)
{
	char	       *buf, *bp;
	size_t		size, pos, end;
	int		mib[2];
	string		os_type, os_release, os_kernel,
			os_version_string, tmp;
	struct tm	tm;

	os_type = getSysctlString(CTL_KERN, KERN_OSTYPE, 0, 2, true, __func__);
	os_release = getSysctlString(CTL_KERN, KERN_OSRELEASE, 0, 2, true, __func__);
	os_kernel = getSysctlString(CTL_KERN, KERN_VERSION, 0, 2, true, __func__);

	mib[0] = CTL_KERN;
	mib[1] = KERN_VERSION;
	if (sysctl(mib, 2, NULL, &size, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");
	buf = new char[size];
	if (buf == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	if (sysctl(mib, 2, buf, &size, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");
	bp = buf;
	while (*bp != '\n' && *bp != '\0')
		bp++;
	*bp = '\0';
	os_version_string = buf;
	delete[] buf;

	pos = os_version_string.find(':') + 2;
	end = os_version_string.find(';');
	tmp = os_version_string.substr(pos, pos+end);

	{
		char	buf2[BUFSIZ];
		char	wkday[5], month[5];
		int	day, hour, minute, second, year;

		/* Need to reformat default date(1) string for strptime() */

		/*
		 * XXX
		 * Parse timezone and convert kernel buid timestamp to UTC?
		 */
		if (sscanf(tmp.c_str(), "%s %s %d %d:%d:%d %*s %d",
			   (char *)&wkday, (char *)&month, &day,
			   &hour, &minute, &second, &year) != 7)
			XSI_Log::log()->err(__func__, "sscanf");

		snprintf(buf2, sizeof(buf2), "%s %s %2d %0d:%0d:%0d %d",
			 wkday, month, day, hour, minute, second, year);

		if (strptime(buf2, "%a %b %e %H:%M:%S %Y", &tm) == NULL)
			XSI_Log::log()->err(__func__, "strptime");

		if (strftime(buf2, sizeof(buf2), "%Y%m%d%H%M", &tm) == 0)
			XSI_Log::log()->err(__func__, "strftime");

		tmp = buf2;
	}
	os_kernel = os_version_string.substr(0, pos-2);

	system.setOSInfo(NULL, os_kernel.c_str(),
			 os_version_string.c_str());
}

void
XSI::updateCPUStats(void)
{
	vector<struct cpu>	cpus;
	struct cpu		cpu; 
	
	mach_msg_type_number_t count;
	kern_return_t status;
	double time;

	/*
	 * Get CPU usage counters.
	 */
	count = HOST_CPU_LOAD_INFO_COUNT;
	status = host_statistics(mach_host_self(), HOST_CPU_LOAD_INFO,
		(host_info_t)&cur, &count);
	if (status != KERN_SUCCESS)
		XSI_Log::log()->err(__func__, "Couldn't fetch CPU states");

	time = 0.0;
	cur.cpu_ticks[CPU_STATE_USER]
		-= last.cpu_ticks[CPU_STATE_USER];
	last.cpu_ticks[CPU_STATE_USER]
		+= cur.cpu_ticks[CPU_STATE_USER];
	time += cur.cpu_ticks[CPU_STATE_USER];
	cur.cpu_ticks[CPU_STATE_SYSTEM]
		-= last.cpu_ticks[CPU_STATE_SYSTEM];
	last.cpu_ticks[CPU_STATE_SYSTEM]
		+= cur.cpu_ticks[CPU_STATE_SYSTEM];
	time += cur.cpu_ticks[CPU_STATE_SYSTEM];
	cur.cpu_ticks[CPU_STATE_IDLE]
		-= last.cpu_ticks[CPU_STATE_IDLE];
	last.cpu_ticks[CPU_STATE_IDLE]
		+= cur.cpu_ticks[CPU_STATE_IDLE];
	time += cur.cpu_ticks[CPU_STATE_IDLE];
	
		cpu.num = -1;
		cpu.system = (100. * cur.cpu_ticks[CPU_STATE_SYSTEM])/time;
		cpu.user = (100. * cur.cpu_ticks[CPU_STATE_USER])/time;
		cpu.nice = 0.0;
		cpu.interrupt = 0.0;
		cpu.idle = (100. * cur.cpu_ticks[CPU_STATE_IDLE])/time;
		cpus.push_back(cpu);
	
	system.setCPUs(cpus);



}


void
XSI::updateVMStats(void)
{
	struct vm_info vminfo;
	struct vm_statistics vm_stat;
	mach_msg_type_number_t count;
	kern_return_t status;
	vm_size_t pagesize;
	int mib[] = { CTL_HW, HW_PHYSMEM };
	int totalmem;
	struct xsw_usage xsu_usage;
	
	//total memory

	size_t siz = sizeof(totalmem);

	if (sysctl(mib, 2, &totalmem, &siz, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");

	
	//vm free, pagesizes
	count = sizeof(vm_stat) / sizeof(natural_t);
	status = host_statistics(mach_host_self(), HOST_VM_INFO,
	    (host_info_t)&vm_stat, &count);
	if (status != KERN_SUCCESS)
		XSI_Log::log()->err(__func__, "host_statistics"); // XXX
	host_page_size(mach_host_self(), &pagesize);
	
	mib[1] = VM_SWAPUSAGE;


	//page file usage -- XXX: "xsw_usage is specific to 10.4.1 and extremely unportable"
	siz = sizeof(xsu_usage);
	if (sysctl(mib, 2, &xsu_usage, &siz, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");


		vminfo.ram_total=totalmem;
		vminfo.ram_free=vm_stat.free_count*pagesize;
		vminfo.swap_total=xsu_usage.xsu_total;
		vminfo.swap_free=xsu_usage.xsu_avail;
		system.setVMInfo(vminfo);

}

void
XSI::updateOSUptime(void)
{
	int		mib[] = { CTL_KERN, KERN_BOOTTIME };
	time_t		now, uptime;
	struct timeval	tv;
	size_t		size = sizeof(tv);

	if (sysctl(mib, 2, &tv, &size, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");
	time(&now);
	uptime = now - tv.tv_sec;

	system.setOSUptime(uptime);
}

void
XSI::updateLoadAvg(void)
{
	double v[3];

	if (getloadavg(v, 3) == -1)
		XSI_Log::log()->err(__func__, errno, "getloadavg");

	system.setLoadAvg(v);
}

void
XSI::readNetIfInfo(XSI_NetIfStat &netIfStat, struct ifaddrs * ifaddr)
{
	struct ifreq	ifr, *ifrp;
	int		sock;

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
		XSI_Log::log()->err(__func__, errno, "socket");

	bzero(&ifr, sizeof(ifr));
	memcpy(&ifr.ifr_addr, ifaddr->ifa_addr,
	       MIN(sizeof(ifr.ifr_addr), ifaddr->ifa_addr->sa_len));
	ifrp = &ifr;
	strlcpy(ifrp->ifr_name, ifaddr->ifa_name, sizeof(ifrp->ifr_name));
	if (ioctl(sock, SIOCGIFMETRIC, (caddr_t)&ifr) != -1)
		netIfStat.setMetric(ifr.ifr_metric);
	if (ioctl(sock, SIOCGIFMTU, (caddr_t)&ifr) != -1)
		netIfStat.setMTU(ifr.ifr_mtu);

	close(sock);
}

void
XSI::readNetIfAddrs(XSI_NetIfStat &netIfStat, struct ifaddrs *ifap, string ifName)
{
	struct ifaddrs	       *ifa;
	vector<XSI_IfAddress>	addrs;
	static bool		have_warned = false;

	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
		XSI_IfAddress addr;

		if (ifName != ifa->ifa_name
		    || !ifa->ifa_addr
		    || ifa->ifa_addr->sa_family == AF_LINK)
			continue;

		switch (ifa->ifa_addr->sa_family) {
		case AF_INET:
		case AF_INET6:
			addr.setAddress((struct sockaddr *)ifa->ifa_addr,
					(size_t)ifa->ifa_addr->sa_len,
					(struct sockaddr *)ifa->ifa_netmask,
					ifa->ifa_addr->sa_family);
			if (ifa->ifa_flags & IFF_POINTOPOINT && ifa->ifa_dstaddr)
				netIfStat.setPTPDestAddress((struct sockaddr *)ifa->ifa_dstaddr,
							    ifa->ifa_dstaddr->sa_len);
			break;
		case AF_APPLETALK:
			if (!have_warned) {
				XSI_Log::log()->warn(NULL, "Appletalk address found, but currently unhandled. This is still in development, sorry.");
				have_warned = true;
			}
			break;
		default:
			if (!have_warned) {
				XSI_Log::log()->warn(__func__,
						     "Unhandled address family %d found, please report this!",
						     ifa->ifa_addr->sa_family);
				have_warned = true;
			}
			break;
		}

		addrs.push_back(addr);
	}

	netIfStat.setIfAddresses(addrs);
}

void
XSI::updateNetStats(void)
{
	struct ifaddrs	       *ifap, *ifa;
	vector<struct ifaddrs *> link_ifs;
	vector<struct ifaddrs *>::iterator it;
	vector<XSI_NetIfStat>	old_nstats = network.getNetIfStats();
	bool			drop_old_stats = false, stats_changed;
	int			i;

	if (getifaddrs(&ifap) == -1)
		XSI_Log::log()->err(__func__, errno, "getifaddrs");

	/* Convenience; make a list of pointers to link-layer interfaces */
	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
		if (ifa->ifa_addr->sa_family == AF_LINK)
			link_ifs.push_back(ifa);
	}

	/* Loopback should always be there, right ...? */
	if (link_ifs.size() == 0) {
		freeifaddrs(ifap);
		return;
	}

	if (link_ifs.size() != old_nstats.size())
		drop_old_stats = true;

	do {
		stats_changed = false;
		if (drop_old_stats) {
			vector<XSI_NetIfStat> new_nstats;

			for (it = link_ifs.begin(); it != link_ifs.end(); it++) {
				XSI_NetIfStat	netIfStat((*it)->ifa_name,
							  (*it)->ifa_flags & IFF_UP);
				struct if_data *ifd;

				netIfStat.setIsRunning((*it)->ifa_flags & IFF_RUNNING);
				netIfStat.setIsPromiscuous((*it)->ifa_flags & IFF_PROMISC);
				netIfStat.setIsLoopback((*it)->ifa_flags & IFF_LOOPBACK);
				readNetIfInfo(netIfStat, (*it));
				readNetIfAddrs(netIfStat, ifap, (*it)->ifa_name);

				ifd = (struct if_data *)(*it)->ifa_data;

				netIfStat.setPacketsIn(ifd->ifi_ipackets);
				netIfStat.setPacketsOut(ifd->ifi_opackets);
				netIfStat.setErrorsIn(ifd->ifi_ierrors);
				netIfStat.setErrorsOut(ifd->ifi_oerrors);
				netIfStat.setBytesIn(ifd->ifi_ibytes);
				netIfStat.setBytesOut(ifd->ifi_obytes);
				netIfStat.setPacketsDropped(ifd->ifi_iqdrops);
				netIfStat.setCollisions(ifd->ifi_collisions);

				new_nstats.push_back(netIfStat);
			}
			network.setNetIfStats(new_nstats);
		} else {
			for (it = link_ifs.begin(), i = 0; it != link_ifs.end(); it++, i++) {
				struct if_data	       *ifd;

				if (strcmp(old_nstats[i].getIfName(), (*it)->ifa_name) != 0) {
					drop_old_stats = true;
					stats_changed = true;
					break;
				}

				old_nstats[i].setIsUp((*it)->ifa_flags & IFF_UP);
				old_nstats[i].setIsRunning((*it)->ifa_flags & IFF_RUNNING);
				old_nstats[i].setIsPromiscuous((*it)->ifa_flags & IFF_PROMISC);
				old_nstats[i].setIsLoopback((*it)->ifa_flags & IFF_LOOPBACK);
				readNetIfInfo(old_nstats[i], (*it));
				readNetIfAddrs(old_nstats[i], ifap, old_nstats[i].getIfName());

				ifd = (struct if_data *)(*it)->ifa_data;

				old_nstats[i].setPacketsIn(ifd->ifi_ipackets);
				old_nstats[i].setPacketsOut(ifd->ifi_opackets);
				old_nstats[i].setErrorsIn(ifd->ifi_ierrors);
				old_nstats[i].setErrorsOut(ifd->ifi_oerrors);
				old_nstats[i].setBytesIn(ifd->ifi_ibytes);
				old_nstats[i].setBytesOut(ifd->ifi_obytes);
				old_nstats[i].setPacketsDropped(ifd->ifi_iqdrops);
				old_nstats[i].setCollisions(ifd->ifi_collisions);
			}
			if (!stats_changed)
				network.setNetIfStats(old_nstats);
		}
	} while (stats_changed);

	freeifaddrs(ifap);
}

void
XSI::updateFSStats(void)
{
	struct statfs	       *sfs;
	int			num_mounts, i;
	size_t			size;
	vector<struct fs>	filesystems;

	if ((num_mounts = getfsstat(NULL, 0, MNT_NOWAIT)) == -1)
		XSI_Log::log()->err(__func__, errno, "getfsstat");

	size = num_mounts * sizeof(struct statfs);
	sfs = new struct statfs[num_mounts];
	if (sfs == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");

	if ((num_mounts = getfsstat(sfs, size, MNT_NOWAIT)) == -1)
		XSI_Log::log()->err(__func__, errno, "getfsstat");

	for (i = 0; i < num_mounts; i++) {
		struct fs fs;

		fs.mountpoint = sfs[i].f_mntonname;
		fs.size_total = (long long)sfs[i].f_blocks * (long long)sfs[i].f_bsize;
		/* weed out uninteresting filesystems */
		if (fs.mountpoint.find("/dev") == 0 || fs.size_total == 0)
			continue;
		fs.fs_name = sfs[i].f_mntfromname;
		fs.fs_type = sfs[i].f_fstypename;
		fs.size_free = (long long)sfs[i].f_bfree * (long long)sfs[i].f_bsize;
		fs.size_avail = (long long)sfs[i].f_bavail * (long long)sfs[i].f_bsize;
		fs.nodes_total = sfs[i].f_files;
		fs.nodes_free = sfs[i].f_ffree;

		filesystems.push_back(fs);
	}

	storage.setFilesystems(filesystems);

	delete[] sfs;
}

