/*	$Kolabore: ipf.cpp,v 1.17 2007/02/09 13:06:06 maxx Exp $	*/
/*
 * Copyright (c) 2006, 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * ipf.cpp:
 * Gather statistics from the ipf(4) packet filter.
 */

#include "log.h"

extern "C" {
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <netinet/in.h>
#ifdef HAVE_IPFILTER
#  include <netinet/ip_compat.h>
#  include <netinet/ip_fil.h>
#  include <netinet/ip_state.h>
#  include <netinet/ipl.h>
#endif

#include <fcntl.h>
#include <unistd.h>
}

#include <cerrno>
#include <iostream>
#include <cstring>

#include <list>
#include <vector>

#include "ipf.h"

using namespace std;

IPF::IPF(void)
{
	ipl_fd = -1;
	ipstate_fd = -1;
	initialized = false;
}

IPF::~IPF(void)
{
	if (ipl_fd != -1)
		close(ipl_fd);
	if (ipstate_fd != -1)
		close(ipstate_fd);
}

/*
 * Initialize the IPF class and return 1 on success, 0 on failure.
 */
int
IPF::init(void)
{
#ifdef HAVE_IPFILTER
	if ((ipl_fd = open(IPL_NAME, O_RDONLY, 0)) == -1) {
		cerr << "Error: Unable to open " IPL_NAME
			" for ipf(4) stats: " << strerror(errno) << endl;
		return (0);
	}

	if ((ipstate_fd = open(IPSTATE_NAME, O_RDONLY, 0)) == -1) {
		cerr << "Error: Unable to open " IPSTATE_NAME
			" for ipf(4) stats: " << strerror(errno) << endl;
		return (0);
	}

	initialized = true;
	return (1);
#else
	return (0);
#endif /* HAVE_IPFILTER */
}

void
IPF::updateStats(XSI_PFilter &packetFilter)
{
#ifdef HAVE_IPFILTER
	ipfobj_t	ipfo;
	friostat_t	frio;
	ips_stat_t	iss;
	bool		running;
	uint64_t	state_inserts;
	vector<FilteredIfStat>	stats;
	stats.push_back(FilteredIfStat(NULL));
	stats[0].setIPv4IsSummaryTraffic(true);

	if (!initialized || ipl_fd == -1 || ipstate_fd == -1)
		return;

	packetFilter.setPFilterType(PFILTER_IPFILTER);
	packetFilter.setAvailable(true);

	memset(&ipfo, 0, sizeof(ipfo));
	memset(&frio, 0, sizeof(frio));
	ipfo.ipfo_rev = IPFILTER_VERSION;
	ipfo.ipfo_size = sizeof(frio);
	ipfo.ipfo_ptr = &frio;
	ipfo.ipfo_type = IPFOBJ_IPFSTAT;
	if (ioctl(ipl_fd, SIOCGETFS, &ipfo) == -1)
		XSI_Log::log()->err(__func__, errno, "SIOCGETFS (ipl)");

	running = (frio.f_running == 1) ? true : false;
	packetFilter.setEnabled(running);
	if (!running)
		return;

	packetFilter.setRuntime(frio.f_ticks / 2);

	stats[0].setPktsV4OutBlocked(frio.f_st[IPF_OUT].fr_block);
	stats[0].setPktsV4OutPassed(frio.f_st[IPF_OUT].fr_pass);
	stats[0].setPktsV4InBlocked(frio.f_st[IPF_IN].fr_block);
	stats[0].setPktsV4InPassed(frio.f_st[IPF_IN].fr_pass);
	packetFilter.setFilteredIfStats(stats);

	memset(&ipfo, 0, sizeof(ipfo));
	memset(&iss, 0, sizeof(iss));
	ipfo.ipfo_rev = IPFILTER_VERSION;
	ipfo.ipfo_size = sizeof(iss);
	ipfo.ipfo_ptr = &iss;
	ipfo.ipfo_type = IPFOBJ_STATESTAT;
	if (ioctl(ipstate_fd, SIOCGETFS, &ipfo) == -1)
		XSI_Log::log()->err(__func__, errno, "SIOCGETFS (ipstate)");

	/*
	 * The number of rule matches is:
	 * (all packets - packets that don't match a rule) - state matches
	 */
	packetFilter.setRuleMatches(((uint64_t)frio.f_st[IPF_OUT].fr_block +
				     (uint64_t)frio.f_st[IPF_IN].fr_block +
				     (uint64_t)frio.f_st[IPF_OUT].fr_pass +
				     (uint64_t)frio.f_st[IPF_IN].fr_pass)
				    - ((uint64_t)frio.f_st[IPF_OUT].fr_nom +
				       (uint64_t)frio.f_st[IPF_IN].fr_nom)
				    - iss.iss_hits);
	/*
	 * What's the difference between iss_inuse and iss_active, except that
	 * the latter appears to have numbers that appear more correct?
	 */
	packetFilter.setStates(iss.iss_active);
	packetFilter.setStateSearches((uint64_t)iss.iss_hits
				      + (uint64_t)iss.iss_miss);
	state_inserts = (uint64_t)frio.f_st[IPF_IN].fr_ads
		+ (uint64_t)frio.f_st[IPF_OUT].fr_ads;
	packetFilter.setStateInserts(state_inserts);
	packetFilter.setStateRemovals(state_inserts - iss.iss_active);
	/*
	 * iss.iss_miss appears to include packets that create state as well.
	 * That's silly, as we already know the number of state inserts, so
	 * substract state_inserts to get the real number of mismatches.
	 * As a safeguard, set iss.iss_miss only if the substraction results
	 * in an unexpected negative value, as that shouldn't kill off XSI.
	 */
	if ((iss.iss_miss - state_inserts) < 0)
		packetFilter.setStateMismatches(iss.iss_miss);
	else
		packetFilter.setStateMismatches(iss.iss_miss - state_inserts);
#endif /* HAVE_IPFILTER */
}
