/*	$Kolabore: linux.cpp,v 1.98 2007/02/09 15:50:43 maxx Exp $	*/
/*
 * Copyright (c) 2006, 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * linux.cpp:
 * Gathering information and stats on Linux 2.6.14 or newer.
 */

#include "log.h"

extern "C" {
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/utsname.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/statvfs.h>
#include <sys/sysinfo.h>

/*
 * Note to self: The following would be required to allow inclusion of
 * <linux/ethtool.h>. We can't use it to get network interface media
 * information, because it's restricted to root.
 */
/*
#ifndef u8
#  define u8 uint8_t
#endif
#ifndef u16
#  define u16 uint16_t
#endif
#ifndef u32
#  define u32 uint32_t
#endif
#ifndef u64
#  define u64 uint64_t
#endif
*/

// #include <linux/fs.h> // NTS; required for ioctl(BLKSSZGET)
// #include <linux/ethtool.h> // NTS; required for struct ethtool_cmd
#include <linux/if.h>
#include <linux/major.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
// #include <linux/sockios.h> // NTS; required for ioctl(SIOCETHTOOL)

#include <net/if_arp.h>

#include <netinet/ether.h>
#include <netinet/in.h>
#include <arpa/inet.h> // This one should go away eventually. IPv4-only == bad.

#include <dirent.h>
#include <fcntl.h>
#include <fts.h>
#include <ifaddrs.h>
#include <mntent.h>
#include <unistd.h>
}

#include <cctype>
#include <cerrno>
#include <climits>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>

#include <fstream>
#include <list>
#include <string>
#include <vector>

#include "linux.h"

#ifndef MIN
#  define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#endif /* MIN */

#ifndef SA_LEN
#  ifdef HAVE_STRUCT_SOCKADDR_SA_LEN
#    define SA_LEN(x)	((x)->sa_len)
#  else
#    define SA_LEN(x)	((x)->sa_family == AF_INET6 ? \
			sizeof(struct sockaddr_in6) : \
			sizeof(struct sockaddr_in))
#  endif /* HAVE_STRUCT_SOCKADDR_SA_LEN */
#endif /* SA_LEN */

#define SEQNUM				0xdeadbeef
#define FILESYSTEM_DESCRIPTION_FILE	"/etc/mtab"
#define SMALLBUFSIZ			128
#ifndef PROC_PATH
#  define PROC_PATH			"/proc"
#endif /* PROC_PATH */
#ifndef SYS_PATH
#  define SYS_PATH			"/sys"
#endif /* SYS_PATH */

#define XSI_CP_USER	0
#define XSI_CP_NICE	1
#define XSI_CP_SYS	2
#define XSI_CP_IDLE	3
#define XSI_CP_IOWAIT	4
#define XSI_CP_IRQ	5
#define XSI_CP_SOFTIRQ	6
#define XSI_CPUSTATES	7

/* Default sector size (correct for at least harddisks): */
#define LX_DEFAULT_SECTOR_SIZE	512

#define SENSOR_ALARM_UNSPEC	0x00
#define SENSOR_ALARM_OK 	0x01
#define SENSOR_ALARM_WARN	0x02
#define SENSOR_ALARM_CRIT	0x04

using namespace std;

/*
 * Structs to locally remember some things ...
 */
struct linux_io_data {
	char		name[32];
	uint64_t	rbytes;
	uint64_t	wbytes;
	uint64_t	rxfers;
	uint64_t	wxfers;
	uint32_t	major;
};

struct xsi_intr_data {
	string		source;
	uint64_t	count;
};

struct linux_proc_info {
	uint32_t	pid;
	uint64_t	cp_time;
};

struct linux_private {
	uint64_t	pagesize;
	uint64_t	last_cp_time_total;
	uint64_t	cp_time_diff;
	vector<struct linux_proc_info> ps;
};

/*
 * XSI Constructor
 *
 * !! WARNING !!
 * !! Runs before dropping privileges when running as root
 * !! WARNING !!
 */
XSI::XSI(void)
{
	num_cpu = -1;
	rtnl_fd = -1;
	priv = NULL;
}

XSI::~XSI(void)
{
	int i;

	/* allocated in linuxInit() */
	for (i = 0; i < num_cpu; i++) {
		delete[] cur_cp_time[i];
		delete[] last_cp_time[i];
	}
	delete[] cur_cp_time;
	delete[] last_cp_time;
	if (rtnl_fd != -1)
		close(rtnl_fd);
	if (priv != NULL)
		delete priv;
}

/*
 * Information read once on startup
 */
void
XSI::collectSysInfo(void)
{
	linuxInit();
	readSystemOSInfo();
}

/*
 * Stats to be updated regularly
 */
void
XSI::updateStats(void)
{
	updateCPUStats(); /* !!! This _must_ run before updateProcInfo()! */
	updateSystemCPUFreq();
	updateSystemSysInfo();
	updateLoadAvg();
	updateProcInfo();
	updateInterrupts();
	updateSensors();
	updateFSStats();
	updateDiskStats();
	updateNetStats();
	updateDefaultRoutes();
}

void
XSI::linuxInit(void)
{
	static bool firstrun = true;

	if (firstrun) {
		int i, j, flags;

		firstrun = false;

		if ((rtnl_fd = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE)) == -1)
			XSI_Log::log()->err(__func__, errno, "socket");
		flags = fcntl(rtnl_fd, F_GETFL);
		if (flags < 0 || fcntl(rtnl_fd, F_SETFL, O_NONBLOCK | flags) == -1)
			XSI_Log::log()->err(__func__, errno, "fcntl");
		readSystemNumCPUs();
		cur_cp_time = new int64_t*[num_cpu];
		last_cp_time = new int64_t*[num_cpu];
		if (cur_cp_time == NULL || last_cp_time == NULL)
			XSI_Log::log()->err(__func__, "Out of memory");
		for (i = 0; i < num_cpu; i++) {
			cur_cp_time[i] = new int64_t[XSI_CPUSTATES];
			last_cp_time[i] = new int64_t[XSI_CPUSTATES];
			if (cur_cp_time[i] == NULL || last_cp_time[i] == NULL)
				XSI_Log::log()->err(__func__, "Out of memory");
			for (j = 0; j < XSI_CPUSTATES; j++) {
				cur_cp_time[i][j] = 0;
				last_cp_time[i][j] = 0;
			}
		}

		if ((priv = new struct linux_private) == NULL)
			XSI_Log::log()->err(__func__, "Out of memory");
#if defined(_SC_PAGESIZE)
		priv->pagesize = (uint64_t)sysconf(_SC_PAGESIZE);
#else
		priv->pagesize = (uint64_t)getpagesize();
#endif

		priv->cp_time_diff = priv->last_cp_time_total = 0;
	}
}

void
XSI::readSystemNumCPUs(void)
{
	DIR	       *dirp;
	FTS	       *ftsp;
	FTSENT	       *p;
	const char     *path[2] = { SYS_PATH "/devices/system/cpu/", NULL };
	int		ncpu = 0;

	if ((dirp = opendir(path[0])) == NULL)
		XSI_Log::log()->err(__func__, errno, "%s", path[0]);
	else
		closedir(dirp);

	if ((ftsp = fts_open(const_cast<char *const *>(path), 0, NULL)) == NULL)
		XSI_Log::log()->err(__func__, errno, "fts_open: %s", path[0]);
	errno = 0;
	while ((p = fts_read(ftsp)) != NULL) {
		switch (p->fts_info) {
		case FTS_D:
			if (strlen(p->fts_name) < 4
			    || strncmp(p->fts_name, "cpu", 3) != 0
			    || !isdigit(p->fts_name[3]))
				break;
			ncpu++;
			break;
		case FTS_ERR:
			XSI_Log::log()->err(__func__, p->fts_errno, "fts_read");
		default:
			break;
		}
	}
	if (errno != 0)
		XSI_Log::log()->err(__func__, errno, "fts_read");
	fts_close(ftsp);

	if (ncpu == 0)
		XSI_Log::log()->err(__func__, "Internal error: No CPUs found (possibly a bug)");
	num_cpu = ncpu;
	system.setNumCPUs(ncpu);
}

void
XSI::readSystemOSInfo(void)
{
	struct utsname  name;
	string          tmp, os_type, os_release, os_version, os_kernel,
		        os_version_string;
	vector<string>	words;
	size_t		i;
	char		buf[BUFSIZ];
	struct tm	tm;

	if (uname(&name) == -1)
		XSI_Log::log()->err(__func__, errno, "uname");

	os_type = name.sysname;
	os_release = name.release;
	os_version = name.version;

	/*
	 * Here goes, the nasty part of trying to normalize a Linux
	 * version string: Split it up into words, take the last
	 * 7 words, throw away the timezone string, put them back
	 * together and parse the date.
	 * The first words are the actual version, which also
	 * need processing and finally put it all back together.
	 * The reason for this is that Linux kernels usually don't
	 * have a name ... except when they do. And then, the name
	 * appears as a separate word - unlike in BSD-land where it's
	 * always <name>#<number> - after the version number.
	 */

	/* split string into a vector of words */
	words = split(os_version);

	/* Let's make an assertion here ... */
	if (words.size() < 7)
		XSI_Log::log()->err(__func__, "Internal error (words.size): Assertion failed: name.version is %s",
				    name.version);

	/* reassemble a string parseable by strptime() */
	tmp = "";
	for (i = words.size() - 6; i < words.size(); i++) {
		if (i != words.size() - 2) // ignore timezone string
			tmp += words[i];
		if (i < words.size() - 2)
			tmp += " ";
	}

	/* parse and reformat, keep the result in buf */
	if (strptime(tmp.c_str(), "%a %b %e %H:%M:%S %Y", &tm) == NULL)
		XSI_Log::log()->err(__func__, "Internal error (strptime): Assertion failed: name.version is %s",
				    name.version);

	if (strftime(buf, sizeof(buf), "%Y%m%d%H%M", &tm) == 0)
		XSI_Log::log()->err(__func__, "strftime");

	/* concatenate all as-of-yet unprocessed words, except for the first */
	tmp = "";
	for (i = 1; i < words.size() - 6; i++) {
		tmp += words[i];
		if (i < words.size() - 7)
			tmp += "_";
	}

	/* finally, put it all back together */
	os_kernel = tmp;
	os_kernel += words[0];
	os_kernel += " ";
	os_kernel += buf;

	/* assemble an "uname -a"-like version string */
	os_version_string = os_type;
	os_version_string += " ";
	os_version_string += os_release;
	os_version_string += " ";
	os_version_string += os_version;
	os_version_string += " ";
	os_version_string += name.machine;

	system.setOSInfo(NULL, os_kernel.c_str(), os_version_string.c_str());
}

void
XSI::updateCPUStats(void)
{
	int64_t 	       *total;
	int64_t 	      **used;
	double		       *percent;
	vector<struct cpu>	cpus;
	int			i, j;
	static bool		load_per_cpu = true;
	vector<string>		proc_stat = getFileVector(PROC_PATH "/stat");
	uint64_t		cp_time_total;

	total = new int64_t[num_cpu];
	used = new int64_t*[num_cpu];
	percent = new double[num_cpu];
	if (total == NULL || used == NULL || percent == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	for (i = 0; i < num_cpu; i++) {
		used[i] = new int64_t[XSI_CPUSTATES];
		if (used[i] == NULL)
			XSI_Log::log()->err(__func__, "Out of memory");
	}

	cp_time_total = 0;
	for (i = 0; i < num_cpu; i++) {
		struct cpu	cpu;
		vector<string>::iterator it;
		char		cpustr[16];
		bool		found_load_per_cpu = false;

		if (load_per_cpu) {
			snprintf(cpustr, sizeof(cpustr), "cpu%d ", i);
			for (it = proc_stat.begin(); it != proc_stat.end(); it++) {
				if (it->find(cpustr) == 0) {
					if (sscanf(it->c_str(),
						   "%*s %" PRIi64 " %" PRIi64
						   " %" PRIi64 " %" PRIi64
						   " %" PRIi64 " %" PRIi64
						   " %" PRIi64 " %*s",
						   &cur_cp_time[i][XSI_CP_USER],
						   &cur_cp_time[i][XSI_CP_NICE],
						   &cur_cp_time[i][XSI_CP_SYS],
						   &cur_cp_time[i][XSI_CP_IDLE],
						   &cur_cp_time[i][XSI_CP_IOWAIT],
						   &cur_cp_time[i][XSI_CP_IRQ],
						   &cur_cp_time[i][XSI_CP_SOFTIRQ])
					    != 7)
						XSI_Log::log()->err(__func__, "sscanf");
					found_load_per_cpu = true;
				}
			}
			if (!found_load_per_cpu) {
				load_per_cpu = false;
			}
		}
		if (!load_per_cpu) {
			for (it = proc_stat.begin(); it != proc_stat.end(); it++) {
				if (it->find("cpu ") == 0) {
					if (sscanf(it->c_str(),
						   "%*s %" PRIi64 " %" PRIi64
						   " %" PRIi64 " %" PRIi64
						   " %" PRIi64 " %" PRIi64
						   " %" PRIi64 " %*s",
						   &cur_cp_time[0][XSI_CP_USER],
						   &cur_cp_time[0][XSI_CP_NICE],
						   &cur_cp_time[0][XSI_CP_SYS],
						   &cur_cp_time[0][XSI_CP_IDLE],
						   &cur_cp_time[0][XSI_CP_IOWAIT],
						   &cur_cp_time[0][XSI_CP_IRQ],
						   &cur_cp_time[0][XSI_CP_SOFTIRQ])
					    != 7)
						XSI_Log::log()->err(__func__, "sscanf");
					break;
				}
			}
		}

		if (load_per_cpu) {
			total[i] = 0;
			for (j = 0; j < XSI_CPUSTATES; j++) {
				cp_time_total += cur_cp_time[i][j];
				used[i][j] = cur_cp_time[i][j] - last_cp_time[i][j];
				last_cp_time[i][j] = cur_cp_time[i][j];
				total[i] += used[i][j];
			}
			if (total[i] > 0)
				percent[i] = 100 / (double)total[i];
			else
				percent[i] = 0;
			cpu.num = i;
			cpu.system = (used[i][XSI_CP_SYS] + used[i][XSI_CP_IOWAIT])
				* percent[i];
			cpu.user = used[i][XSI_CP_USER] * percent[i];
			cpu.nice = used[i][XSI_CP_NICE] * percent[i];
			cpu.interrupt = (used[i][XSI_CP_IRQ] + used[i][XSI_CP_SOFTIRQ])
				* percent[i];
			cpu.idle = used[i][XSI_CP_IDLE] * percent[i];
			cpus.push_back(cpu);
		} else {
			total[0] = 0;
			for (j = 0; j < XSI_CPUSTATES; j++) {
				cp_time_total += cur_cp_time[0][j];
				used[0][j] = cur_cp_time[0][j] - last_cp_time[0][j];
				last_cp_time[0][j] = cur_cp_time[0][j];
				total[0] += used[0][j];
			}
			if (total[0] > 0)
				percent[0] = 100 / (double)total[0];
			else
				percent[0] = 0;
			cpu.num = -1;
			cpu.system = (used[0][XSI_CP_SYS] + used[0][XSI_CP_IOWAIT]) * percent[0];
			cpu.user = used[0][XSI_CP_USER] * percent[0];
			cpu.nice = used[0][XSI_CP_NICE] * percent[0];
			cpu.interrupt = (used[0][XSI_CP_IRQ] + used[0][XSI_CP_SOFTIRQ]) * percent[0];
			cpu.idle = used[0][XSI_CP_IDLE] * percent[0];
			cpus.push_back(cpu);
			break;
		}
	}

	system.setCPUs(cpus);

	/*
	 * This is required later in updateProcInfo() to get the CPU usage of a
	 * process.
	 */
	if (priv->last_cp_time_total == 0)
		priv->cp_time_diff = 0;
	else
		priv->cp_time_diff = cp_time_total - priv->last_cp_time_total;
	priv->last_cp_time_total = cp_time_total;

	for (i = 0; i < num_cpu; i++)
		delete[] used[i];
	delete[] used;
	delete[] total;
	delete[] percent;
}

/*
 * Enter the 3rd word of the 4th paragraph on page 7 of your manual to
 * find out the CPU frequency on Linux.
 */
void
XSI::updateSystemCPUFreq(void)
{
	static bool	have_cpufreq = true;
	vector<string>	cpuinfofile;
	vector<string>::iterator i;
	string		freq_line;
	double		divisor = 1.0;
	static bool	have_warned = false;

	if (!have_cpufreq)
		return;

	cpuinfofile = getFileVector(PROC_PATH "/cpuinfo");
	for (i = cpuinfofile.begin(); i != cpuinfofile.end(); i++) {
		size_t pos = i->size();

		if (i->find("cpu MHz", 0) < pos) {	// most systems, I hope
			freq_line = *i;
			break;
		} else if (i->find("cycle frequency [Hz]", 0) < pos) {	// alpha
			freq_line = *i;
			divisor = 1000000;
			break;
		} else if (i->find("clock", 0) < pos) {	// ppc
			freq_line = *i;
			break;
		}
	}

	if (!freq_line.empty()) {
		size_t pos = freq_line.find(':') + 2;
		double freq;
		int32_t iFreq;

		freq_line = freq_line.substr(pos);
		freq = strtod(freq_line.c_str(), NULL) / divisor;
		if (freq == 0 || freq == HUGE_VAL || freq == -HUGE_VAL) {
			if (!have_warned) {
				XSI_Log::log()->warn(__func__,
						     "Error converting `%s' to a number",
						     freq_line.c_str());
				have_warned = true;
			}
		} else {
			if ((iFreq = (int32_t)freq) == INT_MAX) {
				if (!have_warned) {
					XSI_Log::log()->warn(__func__,
							     "CPU frequency %f is unreasonably large",
							     freq);
					have_warned = true;
				}
			} else {
				double d;

				if (modf(freq, &d) >= 0.5)
					iFreq++;
				system.setCPUFreq(iFreq);
				return;
			}
		}
	}

	have_cpufreq = false;
	system.setCPUFreq(-1);
}

void
XSI::updateSystemSysInfo(void)
{
	struct sysinfo si;
	struct vm_info vm;

	if (sysinfo(&si) == -1)
		XSI_Log::log()->err(__func__, errno, "sysinfo");

	system.setOSUptime(si.uptime);
	vm.ram_total = (uint64_t)si.totalram * (uint64_t)si.mem_unit;
	vm.ram_free = (uint64_t)si.freeram * (uint64_t)si.mem_unit;
	vm.swap_total = (uint64_t)si.totalswap * (uint64_t)si.mem_unit;
	vm.swap_free = (uint64_t)si.freeswap * (uint64_t)si.mem_unit;
	system.setVMInfo(vm);
}

void
XSI::updateLoadAvg(void)
{
	double v[3];

	if (getloadavg(v, 3) == -1)
		XSI_Log::log()->err(__func__, errno, "getloadavg");

	system.setLoadAvg(v);
}

void
XSI::updateProcInfo(void)
{
	vector<struct process_info> processes;
	vector<struct linux_proc_info> curprocs;
	const char     *procdir = PROC_PATH;
	DIR	       *dirp;
	struct dirent  *d;

	if ((dirp = opendir(procdir)) == NULL)
		XSI_Log::log()->err(__func__, errno, "%s", procdir);

	while ((d = readdir(dirp)) != NULL) {
		char			path[SMALLBUFSIZ];
		string			statl, statml;
		vector<string>		statv, statmv;
		bool			error;
		struct process_info	process;
		const char	       *errstr;
		uint64_t		usrtime, systime;
		struct linux_proc_info  save_proc;
		uint64_t		new_cp_time, old_cp_time;
		vector<string>		tmp;
		struct stat		st;

		/*
		 * Every time an error occurs, just skip to the next proc
		 * entry. The process disappeared right under our collective
		 * behinds.
		 */
		if (sscanf(d->d_name, "%" PRIu32, &process.pid) != 1)
			continue;

		/*
		 * The /stat file contains tons of undocumented numbers.
		 * Luckily, the ... haha, this is funny. The source code
		 * at least gives a hint of what it means and what types
		 * it returns.
		 */
		error = false;
		snprintf(path, sizeof(path), "%s/%" PRIu32 "/stat", procdir,
			 process.pid);
		tmp = getFileVector(path, &error);
		if (error)
			continue;
		else
			statl = tmp[0];

		/*
		 * The /statm file *is* documented, and the document says that
		 * almost all its values have a history of being severly
		 * busted. We're lucky again, because the one thing that is
		 * interesting (resident size of a process in memory) seems to
		 * be okay.
		 */
		error = false;
		snprintf(path, sizeof(path), "%s/%" PRIu32 "/statm", procdir,
			 process.pid);
		tmp = getFileVector(path, &error);
		if (error)
			continue;
		else
			statml = tmp[0];

		/*
		 * The following horrible thing gives us a process' EUID,
		 * because the /proc/<pid> directory itself (but nothing else)
		 * is owned by the EUID.
		 */
		snprintf(path, sizeof(path), "%s/%" PRIu32, procdir,
			 process.pid);
		bzero(&st, sizeof(st));
		if (stat(path, &st) == -1) {
			if (errno == ENOENT)
				continue;
			else
				XSI_Log::log()->err(__func__, errno, "stat");
		}

		process.name = getProcnameFromProcStat(statl);
		statv = split(statl);
		statmv = split(statml);
		process.euid = st.st_uid;
		process.username = user_cache->getUsername(process.euid);
		process.login_user = "";

		/*
		 * statl content, after stripping PID (NAME), has 40 fields:
		 *  char # <state> \
		 *   int > <ppid> <pgid> <sid> <tty_nr> <tty_pgrp> \
		 * ulong > <flags> <min_flt> <cmin_flt> <maj_flt> <cmaj_flt> \
		 *       > <usrtime> <systime> \
		 *  long > <cusrtime> <csystime> <priority> <nice> \
		 *   int > <num_threads> \
		 *       > <0> \
		 * ullng > <vsize> \
		 * ulong > <mm_rss | 0> \
		 *  long > <rsslimit> [... uninteresting crap]
		 *       # syntax error: `newline' unexpected
		 *
		 * statml content:
		 *   int # <size> <ressize> <shared> [... broken crap]
		 */

		/* Process state */
		process.running = (statv[0] == "R");
		process.zombie = (statv[0] == "Z");

		/* Jump through a few hoops to calculate pctcpu. */
		if (sscanf(statv[11].c_str(), "%" PRIu64, &usrtime) != 1)
			XSI_Log::log()->err(__func__, "sscanf: usrtime == `%s'", usrtime);
		if (sscanf(statv[12].c_str(), "%" PRIu64, &systime) != 1)
			XSI_Log::log()->err(__func__, "sscanf: systime == `%s'", systime);
		new_cp_time = usrtime + systime;
		old_cp_time = getLinuxProcCPTime(process.pid);
		save_proc.pid = process.pid;
		save_proc.cp_time = new_cp_time;
		curprocs.push_back(save_proc);
		if (priv->cp_time_diff == 0 || old_cp_time == 0)
			process.pctcpu = 0.0;
		else
			process.pctcpu = 100.0 * ((double)new_cp_time -
						  (double)old_cp_time)
				/ (double)priv->cp_time_diff;

		/* Resident size in memory. */
		process.ressize = strtonum(statmv[1].c_str(), 0, INT_MAX, &errstr);
		if (errstr)
			XSI_Log::log()->err(__func__, "strtonum: %s is %s", statmv[1].c_str(), errstr);
		process.ressize *= priv->pagesize;

		/* Nice */
		process.nice = strtonum(statv[16].c_str(), INT32_MIN, INT32_MAX, &errstr);
		if (errstr)
			XSI_Log::log()->err(__func__, "strtonum: %s is %s", statmv[16].c_str(), errstr);

		processes.push_back(process);
	}

	closedir(dirp);

	priv->ps = curprocs;
	proc_info.setProcessVector(processes);
}

/*
 * Dear Linux kernel developers. Please stop with this insane filesystem-based
 * terror. Please fix sysctl(). PLEASE!
 * This is so sick.
 */
void
XSI::updateInterrupts(void)
{
	vector<XSI_Interrupt>	old_ints;
	bool			drop_old_stats, ints_changed;
	vector<string>		intr_list;
	vector<string>::iterator si;
	vector<struct xsi_intr_data> interrupts;
	vector<struct xsi_intr_data>::iterator ii;

	if (num_cpu <= 0)
		XSI_Log::log()->err(__func__, "Internal error: num_cpu not set");

	{
		vector<string> tmp = getFileVector(PROC_PATH "/interrupts");

		/* Keep only those lines with enough columns */
		for (si = tmp.begin(); si != tmp.end(); si++) {
			if ((int)split(*si).size() > num_cpu + 2)
				intr_list.push_back(*si);
		}
	}

	if (intr_list.empty())
		return;

	for (si = intr_list.begin(); si != intr_list.end(); si++) {
		struct xsi_intr_data intr;
		vector<string>	intrs = split(*si);
		int		i;

		/* Each CPU has a counter per interrupt source */
		intr.count = 0;
		for (i = 0; i < num_cpu; i++) {
			const char     *errstr;
			int64_t 	count;

			count = strtonum(intrs[i + 1].c_str(), 0, INT64_MAX, &errstr);
			if (errstr)
				XSI_Log::log()->err(__func__, "strtonum: %s is %s",
						    intrs[i + 1].c_str(), errstr);
			else if (count < 0)
				XSI_Log::log()->err(__func__, "Internal error: Interrupt counter negative?");
			intr.count += count;
		}

		/*
		 * Interrupts can come from a list of devices. Thankfully,
		 * those lists already have a meaningful separator (', '),
		 * of which we'll only keep the comma.
		 */
		for (i = num_cpu + 2; i < (int)intrs.size(); i++)
			intr.source += intrs[i];

		interrupts.push_back(intr);
	}

	if (interrupts.empty())
		return;

	drop_old_stats = false;
	if ((old_ints = system.getInterrupts()).size() != interrupts.size())
		drop_old_stats = true;

	do {
		ints_changed = false;
		if (drop_old_stats) {
			vector<XSI_Interrupt> new_ints;

			for (ii = interrupts.begin(); ii != interrupts.end(); ii++) {
				XSI_Interrupt interrupt;

				interrupt.setSourceName(ii->source.c_str());
				interrupt.setIntsCount(ii->count);

				new_ints.push_back(interrupt);
			}
			system.setInterrupts(new_ints);
		} else {
			for (unsigned int i = 0; i < interrupts.size(); i++) {
				if (interrupts[i].source != old_ints[i].getSourceName()) {
					ints_changed = true;
					drop_old_stats = true;
					break;
				}
				old_ints[i].setIntsCount(interrupts[i].count);
			}
			if (!ints_changed)
				system.setInterrupts(old_ints);
		}
	} while (ints_changed);
}

void
XSI::updateSensors(void)
{
	vector<struct linux_sensor_private> linux_sensors;
	vector<struct linux_sensor_private>::iterator si;

	sensors.clear();

	{
		const char     *i2cdevs = SYS_PATH "/bus/i2c/devices";
		DIR	       *dirp;
		struct dirent  *d;
		vector<string>	dev_paths;

		if ((dirp = opendir(i2cdevs)) == NULL)
			/* No need to look any further */
			return;
		else {
			string path = i2cdevs;

			while ((d = readdir(dirp)) != NULL) {
				if (d->d_name[0] != '.')
					dev_paths.push_back(path + "/" + d->d_name);
			}
			closedir(dirp);
			linux_sensors = getSensors(dev_paths);
		}
	}

	for (si = linux_sensors.begin(); si != linux_sensors.end(); si++) {
		XSI_Sensor s;
		bool nonsense = false;

		if (si->has_fault)
			continue;

		s.setSensorID(si->id);
		s.setSensorDevice(si->driver.c_str());

		if (si->alarm & SENSOR_ALARM_CRIT)
			s.setSensorStatus(XSI_SENSOR_STATUS_CRIT);
		else if (si->alarm & SENSOR_ALARM_WARN)
			s.setSensorStatus(XSI_SENSOR_STATUS_WARN);
		else if (si->alarm & SENSOR_ALARM_OK)
			s.setSensorStatus(XSI_SENSOR_STATUS_OK);
		else
			s.setSensorStatus(XSI_SENSOR_STATUS_UNSPEC);

		switch (si->type) {
		case VOLT:
			s.setSensorType(XSI_SENSOR_TYPE_VOLTS_DC);
			s.setSensorValue(si->value * 1000);
			break;
		case TEMP:
			s.setSensorType(XSI_SENSOR_TYPE_TEMP);
			s.setSensorValue((si->value + 273150) * 1000);
			break;
		case FAN:
			s.setSensorType(XSI_SENSOR_TYPE_FANRPM);
			/* special case: can contain `-1' as nonsense value */
			if (si->value != -1)
				s.setSensorValue(si->value);
			else
				nonsense = true;
			break;
		case CPUVOLT:
			s.setSensorType(XSI_SENSOR_TYPE_VOLTS_DC);
			s.setSensorValue(si->value * 1000);
			break;
		case AMP:
			s.setSensorType(XSI_SENSOR_TYPE_AMPS);
			s.setSensorValue(si->value * 1000);
			break;
		case OTHER:
			s.setSensorType(XSI_SENSOR_TYPE_RAW_INT);
			s.setSensorValue(si->value);
			break;
		default:
			XSI_Log::log()->err(__func__, "Internal error: Unknown sensor type");
		}

		s.setHasUnknownValue(false);

		if (!nonsense)
			sensors.push_back(s);
	}
}

void
XSI::updateFSStats(void)
{
	vector<struct fs>	filesystems;
	struct mntent	       *mnt;
	FILE		       *fp = NULL;

	if ((fp = fopen(FILESYSTEM_DESCRIPTION_FILE, "r")) == NULL)
		XSI_Log::log()->err(__func__, errno, "fopen: " FILESYSTEM_DESCRIPTION_FILE);

	while ((mnt = getmntent(fp)) != NULL) {
		struct fs	fs;
		struct statvfs	f;
		uint64_t	blocksize;

		fs.mountpoint = mnt->mnt_dir;
		fs.fs_name = mnt->mnt_fsname;
		fs.fs_type = mnt->mnt_type;

		if (statvfs(mnt->mnt_dir, &f) == -1) {
			if (errno == EACCES || errno == EIO)
				/*
				 * If the OS insists on not letting us look at
				 * something, or if a media we're trying to
				 * look at has an error, ignore and move on.
				 */
				continue;
			else
				XSI_Log::log()->err(__func__, errno, "statvfs");
		}

		if ((blocksize = f.f_frsize) == 0)
			blocksize = f.f_bsize;

		fs.size_total = (uint64_t)f.f_blocks * blocksize;
		if (fs.size_total == 0 || fs.mountpoint.find("/dev") == 0 ||
		    fs.fs_type == "tmpfs")
			continue;	// ignore; boring stuff

		fs.size_free = (uint64_t)f.f_bfree * blocksize;
		fs.size_avail = (int64_t)f.f_bavail * (int64_t)blocksize;
		fs.nodes_total = f.f_files;
		fs.nodes_free = f.f_ffree;

		filesystems.push_back(fs);
	}

	fclose(fp);

	storage.setFilesystems(filesystems);
}

void
XSI::updateDiskStats(void)
{
	vector<struct linux_io_data> io_stats;
	vector<struct linux_io_data>::iterator ii;
	vector<XSI_DiskIO>	old_disks = storage.getDisks();
	vector<string>	        diskstats = getFileVector(PROC_PATH "/diskstats");
	vector<string>::iterator si;
	bool			drop_old_stats = false, disks_changed;

	/*
	 * This is plain awful. I don't know how to figure out how many disks
	 * devices are actually in use and how they are called.
	 * This is all about making an educated guess and weeding out obviously
	 * uninteresting devices.
	 */
	for (si = diskstats.begin(); si != diskstats.end(); si++) {
		struct linux_io_data io;
		uint64_t	sects_read, sects_written;
		uint64_t	rxfers, wxfers;
		size_t		sectsiz;

		if (sscanf(si->c_str(),
			   //            1           2   3
			   "%u %*s %31s %" PRIu64 " %*s %" PRIu64
			   // 4   5           6   7
			   " %*s %" PRIu64 " %*s %" PRIu64,
			   &io.major, io.name, &rxfers, &sects_read,
			   &wxfers, &sects_written) != 6)
			continue;

		/* Ignore ramdisks and loopback devices */
		if (io.major == RAMDISK_MAJOR || io.major == LOOP_MAJOR)
			continue;

		/*
		 * The following commented code is similar to what we'd need to
		 * ensure that numbers are correct. Unfortunately, this requires
		 * superuser privileges and that is unacceptable.
		 */
		/*
		{
			int	fd;
			char	device[256];

			snprintf(device, sizeof(device), "/dev/%s", io.name);
			if ((fd = open(device, O_RDONLY)) == -1)
				XSI_Log::log()->err(__func__, errno, "%s", device);
			if (ioctl(fd, BLKSSZGET, &sectsiz) == -1)
				XSI_Log::log()->err(__func__, errno, "BLKSSZGET");
			close(fd);
		}
		*/
		sectsiz = LX_DEFAULT_SECTOR_SIZE;

		/*
		 * XXX: rbytes / wbytes only correct for devices with a
		 *      512 bytes sector size ...
		 */
		io.rbytes = sects_read * sectsiz;
		io.wbytes = sects_written * sectsiz;
		io.rxfers = rxfers;
		io.wxfers = wxfers;

		io_stats.push_back(io);
	}

	if (io_stats.size() != old_disks.size())
		drop_old_stats = true;

	do {
		disks_changed = false;
		if (drop_old_stats) {
			vector<XSI_DiskIO> new_disks;

			for (ii = io_stats.begin(); ii != io_stats.end(); ii++) {
				XSI_DiskIO disk(ii->name);

				/*
				 * Do not set bytes read/written values, as we
				 * don't know whether these are correct or not.
				 */
				// disk.setWrittenBytes(ii->wbytes);
				// disk.setReadBytes(ii->rbytes);
				disk.setWriteTransfers(ii->wxfers);
				disk.setReadTransfers(ii->rxfers);

				new_disks.push_back(disk);
			}
			storage.setDisks(new_disks);
		} else {
			size_t i;

			for (i = 0; i < io_stats.size(); i++) {
				if (strcmp(io_stats[i].name,
					   old_disks[i].getDiskName()) != 0) {
					drop_old_stats = true;
					disks_changed = true;
					break;
				}
				/*
				 * Do not set bytes read/written values, as we
				 * don't know whether these are correct or not.
				 */
				// old_disks[i].setWrittenBytes(io_stats[i].wbytes);
				// old_disks[i].setReadBytes(io_stats[i].rbytes);
				old_disks[i].setWriteTransfers(io_stats[i].wxfers);
				old_disks[i].setReadTransfers(io_stats[i].rxfers);
			}
			if (!disks_changed)
				storage.setDisks(old_disks);
		}
	} while (disks_changed);
}

/*
 * Looks like Linux has an undocumented, neutered getifaddrs().
 * Let's use it as much as possible, as it greatly simplifies
 * things.
 */
void
XSI::updateNetStats(void)
{
	struct ifaddrs	       *ifap, *ifa;
	vector<struct ifaddrs *> link_ifs;
	vector<struct ifaddrs *>::iterator it;
	vector<XSI_NetIfStat>	old_nstats = network.getNetIfStats();
	vector<string>		net_dev = getFileVector(PROC_PATH "/net/dev");
	bool			drop_old_stats = false, stats_changed;
	int			i;

	if (getifaddrs(&ifap) == -1)
		XSI_Log::log()->err(__func__, errno, "getifaddrs");

	/*
	 * Convenience; make a vector of pointers to link-layer (AF_PACKET on
	 * Linux) interfaces
	 */
	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
		if (ifa->ifa_addr->sa_family == AF_PACKET)
			link_ifs.push_back(ifa);
	}

	/* Loopback should always be there, right ...? */
	if (link_ifs.size() == 0) {
		freeifaddrs(ifap);
		return;
	}

	if (link_ifs.size() != old_nstats.size())
		drop_old_stats = true;

	do {
		stats_changed = false;
		if (drop_old_stats) {
			vector<XSI_NetIfStat> new_nstats;

			for (it = link_ifs.begin(); it != link_ifs.end(); it++) {
				XSI_NetIfStat	netIfStat((*it)->ifa_name,
							  (*it)->ifa_flags & IFF_UP);

				netIfStat.setIsRunning((*it)->ifa_flags & IFF_RUNNING);
				netIfStat.setIsPromiscuous((*it)->ifa_flags & IFF_PROMISC);
				netIfStat.setIsLoopback((*it)->ifa_flags & IFF_LOOPBACK);
				getNetIfInfo(netIfStat, (*it));
				getNetIfAddrs(netIfStat, ifap, (*it)->ifa_name);
				getNetIfCounters(netIfStat, net_dev, (*it)->ifa_name);
				getNetIfStatus(netIfStat, (*it)->ifa_name);

				new_nstats.push_back(netIfStat);
			}
			network.setNetIfStats(new_nstats);
		} else {
			for (it = link_ifs.begin(), i = 0; it != link_ifs.end(); it++, i++) {
				if (strcmp(old_nstats[i].getIfName(), (*it)->ifa_name) != 0) {
					drop_old_stats = true;
					stats_changed = true;
					break;
				}

				old_nstats[i].setIsUp((*it)->ifa_flags & IFF_UP);
				old_nstats[i].setIsRunning((*it)->ifa_flags & IFF_RUNNING);
				old_nstats[i].setIsPromiscuous((*it)->ifa_flags & IFF_PROMISC);
				old_nstats[i].setIsLoopback((*it)->ifa_flags & IFF_LOOPBACK);
				getNetIfInfo(old_nstats[i], (*it));
				getNetIfAddrs(old_nstats[i], ifap, old_nstats[i].getIfName());
				getNetIfCounters(old_nstats[i], net_dev, (*it)->ifa_name);
				getNetIfStatus(old_nstats[i], (*it)->ifa_name);
			}
			if (!stats_changed)
				network.setNetIfStats(old_nstats);
		}
	} while (stats_changed);

	freeifaddrs(ifap);
}

/*
 * XXX -- Supports IPv4 default routes only
 */
void
XSI::updateDefaultRoutes(void)
{
	size_t			len;
	uint8_t 	       *buf;
	struct in_addr		addr;
	struct sockaddr_nl	snl;
	struct {
		struct nlmsghdr nlh;
		struct rtgenmsg rtg;
	} req;
	vector<struct default_route> defRoutes;

	memset(&snl, 0, sizeof(snl));
	snl.nl_family = AF_NETLINK;

	memset(&req, 0, sizeof(req));
	req.nlh.nlmsg_len = NLMSG_LENGTH(sizeof(req.rtg));
	req.nlh.nlmsg_type = RTM_GETROUTE;
	req.nlh.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP;
	req.nlh.nlmsg_seq = SEQNUM;
	req.nlh.nlmsg_pid = 0;
	req.rtg.rtgen_family = AF_INET;

	if (sendto(rtnl_fd, &req, sizeof(req), 0,
		   (struct sockaddr *)&snl, sizeof(snl)) == -1)
		XSI_Log::log()->err(__func__, errno, "sendto");

	for (;;) {
		struct default_route	defRt;

		if ((buf = getNetlinkMessage(rtnl_fd, len)) == NULL)
			break;
		if (parseRoutes(buf, len, &addr) == 0) {
			defRt.family = XSI_RTAF_IPV4;
			defRt.gateway = inet_ntoa(addr);
			defRoutes.push_back(defRt);
		}
		delete[] buf;
	}

	network.setDefaultRoutes(defRoutes);
}

void
XSI::getNetIfInfo(XSI_NetIfStat &netIfStat, struct ifaddrs *ifaddr)
{
	struct ifreq	ifr, *ifrp;
	int		sock;

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
		XSI_Log::log()->err(__func__, errno, "socket");

	bzero(&ifr, sizeof(ifr));
	memcpy(&ifr.ifr_addr, ifaddr->ifa_addr,
	       MIN(sizeof(ifr.ifr_addr), SA_LEN(ifaddr->ifa_addr)));
	ifrp = &ifr;
	strlcpy(ifrp->ifr_name, ifaddr->ifa_name, sizeof(ifrp->ifr_name));
	if (ioctl(sock, SIOCGIFMETRIC, (caddr_t)&ifr) != -1)
		netIfStat.setMetric(ifr.ifr_metric);
	if (ioctl(sock, SIOCGIFMTU, (caddr_t)&ifr) != -1)
		netIfStat.setMTU(ifr.ifr_mtu);
	if (ioctl(sock, SIOCGIFHWADDR, (caddr_t)&ifr) != -1
	    && ifr.ifr_hwaddr.sa_family == ARPHRD_ETHER) {
		char	lladdr[24];

		memset(lladdr, 0, sizeof(lladdr));
		snprintf(lladdr, sizeof(lladdr), "%02x:%02x:%02x:%02x:%02x:%02x",
			 (ifr.ifr_hwaddr.sa_data[0] & 0xff),
			 (ifr.ifr_hwaddr.sa_data[1] & 0xff),
			 (ifr.ifr_hwaddr.sa_data[2] & 0xff),
			 (ifr.ifr_hwaddr.sa_data[3] & 0xff),
			 (ifr.ifr_hwaddr.sa_data[4] & 0xff),
			 (ifr.ifr_hwaddr.sa_data[5] & 0xff));
		netIfStat.setLladdrString(lladdr);
	}

	/*
	 * The following code to get interface media information would work if
	 * we had root privileges. That, however, will never happen. Just keep
	 * it here commented out as a reminder until some day, maybe, this
	 * oh-so-confidential data becomes available to regular users.
	 */
	/*
	{
		struct ethtool_cmd ecmd;

		bzero(&ecmd, sizeof(ecmd));
		ecmd.cmd = ETHTOOL_GSET;
		ifr.ifr_data = (caddr_t)&ecmd;
		if (ioctl(sock, SIOCETHTOOL, (caddr_t)&ifr) != -1) {
			// do stuff
		}
	}
	*/

	close(sock);
}

void
XSI::getNetIfAddrs(XSI_NetIfStat &netIfStat, struct ifaddrs *ifap, string ifName)
{
	struct ifaddrs	       *ifa;
	vector<XSI_IfAddress>	addrs;
	static bool		have_warned = false;

	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
		XSI_IfAddress addr;

		if (ifName != ifa->ifa_name
		    || !ifa->ifa_addr
		    || ifa->ifa_addr->sa_family == AF_PACKET)
			continue;

		switch (ifa->ifa_addr->sa_family) {
		case AF_INET:
		case AF_INET6:
			addr.setAddress((struct sockaddr *)ifa->ifa_addr,
					SA_LEN(ifa->ifa_addr),
					(struct sockaddr *)ifa->ifa_netmask,
					ifa->ifa_addr->sa_family);
			if (ifa->ifa_flags & IFF_POINTOPOINT && ifa->ifa_dstaddr)
				netIfStat.setPTPDestAddress((struct sockaddr *)ifa->ifa_dstaddr,
							    SA_LEN(ifa->ifa_dstaddr));
			break;
		default:
			if (!have_warned) {
				XSI_Log::log()->warn(__func__,
						     "Unhandled address family %d found, please report this!",
						     ifa->ifa_addr->sa_family);
				have_warned = true;
			}
			break;
		}

		addrs.push_back(addr);
	}

	netIfStat.setIfAddresses(addrs);
}

void
XSI::getNetIfCounters(XSI_NetIfStat &netIfStat, vector<string> net_dev, string ifName)
{
	vector<string>::iterator i;
	string		if_counters;
	uint64_t	ibytes, obytes, ipkts, opkts, ierrs,
			oerrs, idrops, odrops, ocolls;

	for (i = net_dev.begin(); i != net_dev.end(); i++) {
		if (i->find(ifName, 0) < i->size()) {
			size_t pos = i->find(':');

			if (pos < i->size()) {
				if_counters = i->substr(pos + 1);
				break;
			}
		}
	}

	if (if_counters.empty())
		return;

	/*
	 * ibytes, ipkts, ierrs, idrops, ififo, iframe, icompr, imcast,
	 *   obytes, opkts, oerrs, odrops, ofifo, ocolls, ocarr, ocompr
	 */
	if (sscanf(if_counters.c_str(),
		   "%" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64
		   " %*s %*s %*s %*s %" PRIu64 " %" PRIu64 " %" PRIu64
		   " %" PRIu64 " %*s %" PRIu64 " %*s %*s",
		   &ibytes, &ipkts, &ierrs, &idrops,
		   &obytes, &opkts, &oerrs, &odrops,
		   &ocolls) != 9)
		XSI_Log::log()->err(__func__, "sscanf");

	netIfStat.setPacketsIn(ipkts);
	netIfStat.setPacketsOut(opkts);
	netIfStat.setErrorsIn(ierrs);
	netIfStat.setErrorsOut(oerrs);
	netIfStat.setBytesIn(ibytes);
	netIfStat.setBytesOut(obytes);
	netIfStat.setPacketsDropped(idrops + odrops);
	netIfStat.setCollisions(ocolls);
}

void
XSI::getNetIfStatus(XSI_NetIfStat &netIfStat, string ifName)
{
	char	buf[SMALLBUFSIZ];
	string	val_str;
	bool	error;

	snprintf(buf, sizeof(buf), SYS_PATH "/class/net/%s/operstate",
		 ifName.c_str());

	error = false;
	val_str = getFileString(buf, &error);
	if (!error) {
		if (val_str == "up")
			netIfStat.setLinkStatus(IF_LINK_UP);
		else if (val_str == "down")
			netIfStat.setLinkStatus(IF_LINK_DOWN);
		else
			netIfStat.setLinkStatus(IF_LINK_UNKNOWN);
	} else
		netIfStat.setLinkStatus(IF_LINK_UNKNOWN);
}

uint8_t *
XSI::getNetlinkMessage(int fd, size_t &buflen)
{
	uint8_t 	       *buf;
	size_t			len;
	ssize_t 		res;
	struct sockaddr_nl	snl;
	socklen_t		slen;

	len = BUFSIZ * 2;
	if ((buf = new uint8_t[len]) == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");

	do {
		memset(&snl, 0, sizeof(snl));
		slen = sizeof(snl);
		if ((res = recvfrom(fd, buf, len, 0, (struct sockaddr *)&snl, &slen)) == -1) {
			if (errno != EAGAIN)
				XSI_Log::log()->err(__func__, errno, "recvfrom");
			delete[] buf;
			buflen = 0;
			return (NULL);
		}
		if (sizeof(snl) > slen || snl.nl_family != AF_NETLINK) {
			XSI_Log::log()->err(__func__, "Bad address");
			// delete[] buf;
			// buflen = 0;
			// return (NULL);
		}
	} while (snl.nl_pid != 0);

	buflen = res;

	return (buf);
}

int
XSI::parseRoutes(uint8_t *buf, size_t len, struct in_addr *addr)
{
	struct nlmsghdr        *nlm;
	struct nlmsgerr        *nle;
	struct rtmsg	       *rtm;
	struct rtattr	       *rta;
	int			rtalen;
	struct in_addr		gw, dst;

	nlm = (struct nlmsghdr *)buf;
	while (NLMSG_OK(nlm, len)) {
		gw.s_addr = INADDR_ANY;
		dst.s_addr = INADDR_ANY;
		if (nlm->nlmsg_type == NLMSG_ERROR) {
			nle = (struct nlmsgerr *)NLMSG_DATA(nlm);
			if (NLMSG_LENGTH(NLMSG_ALIGN(sizeof(struct nlmsgerr))) >
			    nlm->nlmsg_len)
				XSI_Log::log()->err(__func__, "Netlink message truncated");
			else
				XSI_Log::log()->err(__func__, nle->error, "Netlink error");
		} else if (nlm->nlmsg_type == RTM_NEWROUTE && nlm->nlmsg_seq == SEQNUM
			   && (pid_t) nlm->nlmsg_pid == getpid()
			   && nlm->nlmsg_len >= NLMSG_LENGTH(sizeof(struct rtmsg))) {
			rtm = (struct rtmsg *)NLMSG_DATA(nlm);
			rta = RTM_RTA(rtm);
			rtalen = RTM_PAYLOAD(nlm);

			while(RTA_OK(rta, rtalen)) {
				if (RTA_PAYLOAD( rta ) >= sizeof(struct in_addr)) {
					switch(rta->rta_type) {
					case RTA_GATEWAY:
						memcpy(&gw, RTA_DATA(rta), sizeof(gw));
						break;
					case RTA_DST:
						memcpy(&dst, RTA_DATA(rta), sizeof(dst));
						break;
					}
				}
				rta = RTA_NEXT(rta, rtalen);
			}
		}

		if (gw.s_addr != INADDR_NONE && gw.s_addr != INADDR_ANY
		    && dst.s_addr == INADDR_ANY) {
			*addr = gw;
			return (0);
		}

		nlm = NLMSG_NEXT(nlm, len);
	}

	return (1);
}

vector<struct linux_sensor_private>
XSI::getSensors(vector<string> dev_paths)
{
	vector<string>::iterator si;
	vector<char **> sensor_paths;
	vector<char **>::iterator cpi;
	vector<struct linux_sensor_private> linux_sensors;

	/*
	 * Create an array of arrays, to be used by fts_open(). While this
	 * could also be in one single array, this guarantees that we always
	 * get the sensor's device name without having to check it for each
	 * entry in the filesystem.
	 */
	for (si = dev_paths.begin(); si != dev_paths.end(); si++) {
		char **path;
		size_t siz = si->size() + 1;

		path = new char*[2];
		path[0] = new char[siz];
		strlcpy(path[0], si->c_str(), siz);
		path[1] = NULL;
		sensor_paths.push_back(path);
	}

	for (cpi = sensor_paths.begin(); cpi != sensor_paths.end(); cpi++) {
		FTS	       *ftsp;
		FTSENT	       *p;
		string		driver = getSysfsDeviceName(*cpi[0]);

		if ((ftsp = fts_open(const_cast<char *const *>(*cpi),
				     FTS_COMFOLLOW | FTS_NOSTAT, NULL)) == NULL)
			XSI_Log::log()->err(__func__, errno, "fts_open: %s", *cpi[0]);
		errno = 0;
		while ((p = fts_read(ftsp)) != NULL) {
			string file;
			struct linux_sensor_private sensor;
			bool sensor_found = false;

			sensor.type = UNKNOWN;
			sensor.value = 0;
			sensor.id = -1;
			sensor.alarm = SENSOR_ALARM_UNSPEC;
			sensor.has_fault = true;

			switch (p->fts_info) {
			case FTS_DNR:
			case FTS_ERR:
				XSI_Log::log()->err(__func__, p->fts_errno, "fts_read");
			case FTS_F:
				file = p->fts_name;
				/* *_input files contain sensor values */
				if (file.size() > 6
				    && file.find("_input") == file.size() - 6) {
					string path = getDirName(p->fts_path);
					string fault_file = path + file
						+ "_fault";
					string alarm_file = path
						+ file.substr(0, file.find('_'))
						+ "_alarm";
					string min_alarm_file = path
						+ file.substr(0, file.find('_'))
						+ "_min_alarm";
					string max_alarm_file = path
						+ file.substr(0, file.find('_'))
						+ "_max_alarm";
					string crit_alarm_file = path
						+ file.substr(0, file.find('_'))
						+ "_crit_alarm";
					int fault, alarm;
					bool error;

					sensor.value = getValueFromFile(p->fts_path);
					sensor.type = getSensorType(file);

					fault = getValueFromFile(fault_file.c_str(),
								 &error, 0, 1);
					if (error)
						sensor.has_fault = false;
					else {
						if (fault == 1)
							sensor.has_fault = true;
						else
							sensor.has_fault = false;
					}

					alarm = getValueFromFile(alarm_file.c_str(),
								 &error, 0, 1);
					if (error) {
						alarm = getValueFromFile(min_alarm_file.c_str(),
									 &error, 0, 1);
						if (!error) {
							if (alarm == 1)
								sensor.alarm |= SENSOR_ALARM_WARN;
							else
								sensor.alarm |= SENSOR_ALARM_OK;
						}
						alarm = getValueFromFile(max_alarm_file.c_str(),
									 &error, 0, 1);
						if (!error) {
							if (alarm == 1)
								sensor.alarm |= SENSOR_ALARM_WARN;
							else
								sensor.alarm |= SENSOR_ALARM_OK;
						}
						alarm = getValueFromFile(crit_alarm_file.c_str(),
									 &error, 0, 1);
						if (!error) {
							if (alarm == 1)
								sensor.alarm |= SENSOR_ALARM_CRIT;
							else
								sensor.alarm |= SENSOR_ALARM_OK;
						}
					} else {
						if (alarm == 1)
							sensor.alarm |= SENSOR_ALARM_WARN;
						else
							sensor.alarm |= SENSOR_ALARM_OK;
					}

					sensor_found = true;
				} else if (file.size() > 4 && file.find("_vid") == file.size() - 4) {
					/* For unknown reasons, voltages in CPU cores get special treatments */
					sensor.value = getValueFromFile(p->fts_path);
					sensor.type = getSensorType(file);
					sensor_found = true;
				}
				break;
			default:
				break;
			}
			if (sensor_found) {
				getSensorID(file, sensor.id);
				sensor.driver = driver;
				linux_sensors.push_back(sensor);
			}
		}
		if (errno != 0)
			XSI_Log::log()->err(__func__, errno, "fts_read");
		fts_close(ftsp);
		delete[] *cpi[0];
		delete[] *cpi;
	}

	return (linux_sensors);
}

enum linux_sensor_type
XSI::getSensorType(string name)
{
	if (name.find("in") == 0)
		return (VOLT);
	else if (name.find("temp") == 0)
		return (TEMP);
	else if (name.find("fan") == 0)
		return (FAN);
	else if (name.find("curr") == 0)
		return (AMP);
	else if (name.find("_vid") == name.size() - 4)
		return (CPUVOLT);
	else
		return (OTHER);
}

void
XSI::getSensorID(string filename, int &num)
{
	string::iterator	i;
	string			buf;
	const char	       *errstr;

	num = 0;
	for (i = filename.begin(); i != filename.end(), !isdigit(*i); i++) {}
	while (isdigit(*i))
		buf += *(i++);
	num = strtonum(buf.c_str(), INT_MIN, INT_MAX, &errstr);
	if (errstr)
		XSI_Log::log()->err(__func__, "strtonum: %s is %s",
				    buf.c_str(), errstr);
}

string
XSI::getSysfsDeviceName(const char *path)
{
	string namefile = path;

	namefile += "/name";

	return (getFileString(namefile.c_str()));
}

string
XSI::getDirName(const char *path)
{
	vector<string>	tmp = split(path, '/');
	vector<string>::iterator it;
	unsigned int	i;
	string		ret = "/";

	for (it = tmp.begin(), i = 0;
	     it != tmp.end(), (tmp.size() > 0 && i < tmp.size() - 1);
	     i++, it++)
		ret += *it + "/";

	return (ret);
}

string
XSI::getProcnameFromProcStat(string &statl)
{
	string name;
	string::iterator it;
	size_t pos1, pos2, i;

	if ((pos1 = statl.find('(')) >= (statl.size() - 1) ||
	    (pos2 = statl.rfind(')')) >= (statl.size() - 2) ||
	    pos1 > pos2)
		XSI_Log::log()->err(__func__, "Internal error: Unexpected string received");

	for (i = pos1 + 1; i < pos2; i++)
		name.push_back(statl[i]);

	statl.erase(0, pos2 + 2);

	return (name);
}

uint64_t
XSI::getLinuxProcCPTime(uint32_t pid)
{
	vector<struct linux_proc_info>::iterator it;

	for (it = priv->ps.begin(); it != priv->ps.end(); it++) {
		if (it->pid == pid)
			return (it->cp_time);
	}

	return (0);
}
