/*	$Kolabore: log.cpp,v 1.14 2007/02/10 00:43:02 maxx Exp $	*/
/*
 * Copyright (c) 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * log.cpp:
 * Provide means to log, print to stderr or discard messages from XMLSysInfo.
 */

#include "log.h"

extern "C" {
#ifdef HAVE_SYSLOG_H
#  include <syslog.h>
#endif
}

#include <iostream>
#include <string>

#include <cerrno>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#define ERRBUFSIZE	256

extern char    *__progname;


XSI_Log *
XSI_Log::log(void)
{
	static XSI_Log	instance;

	return (&instance);
}

XSI_Log::~XSI_Log(void)
{
#ifdef HAVE_SYSLOG_SUPPORT
	closelog();
#endif /* HAVE_SYSLOG_SUPPORT */
}

void
XSI_Log::init(bool log_to_syslog, bool log_to_stderr)
{
	log_stderr = log_to_stderr;
#ifdef HAVE_SYSLOG_SUPPORT
	log_syslog = log_to_syslog;

	openlog(__progname, LOG_NDELAY | LOG_PID, LOG_DAEMON);
#else
	log_syslog = false;
#endif /* HAVE_SYSLOG_SUPPORT */
	initialized = true;
}

void
XSI_Log::debug(const char *fmt, ...)
{
	va_list ap;

	if (!initialized) {
		cerr << "Internal error: XSI_Log class not initialized!"
			<< endl;
		exit(1);
	}

	if (!log_syslog && !log_stderr)
		return;

	va_start(ap, fmt);
	makeMessageAndLog(NULL, 0, fmt, ap, LOG_DEBUG);
	va_end(ap);
}

void
XSI_Log::notice(const char *fmt, ...)
{
	va_list ap;

	if (!initialized) {
		cerr << "Internal error: XSI_Log class not initialized!"
			<< endl;
		exit(1);
	}

	if (!log_syslog && !log_stderr)
		return;

	va_start(ap, fmt);
	makeMessageAndLog(NULL, 0, fmt, ap, LOG_NOTICE);
	va_end(ap);
}

void
XSI_Log::warn(const char *func, const char *fmt, ...)
{
	va_list ap;

	if (!initialized) {
		cerr << "Internal error: XSI_Log class not initialized!"
			<< endl;
		exit(1);
	}

	if (!log_syslog && !log_stderr)
		return;

	va_start(ap, fmt);
	makeMessageAndLog(func, 0, fmt, ap, LOG_WARNING);
	va_end(ap);
}

void
XSI_Log::warn(const char *func, int error, const char *fmt, ...)
{
	va_list ap;

	if (!initialized) {
		cerr << "Internal error: XSI_Log class not initialized!"
			<< endl;
		exit(1);
	}

	if (!log_syslog && !log_stderr)
		return;

	va_start(ap, fmt);
	makeMessageAndLog(func, error, fmt, ap, LOG_WARNING);
	va_end(ap);
}

void
XSI_Log::err(const char *func, const char *fmt, ...)
{
	va_list ap;

	if (!initialized) {
		cerr << "Internal error: XSI_Log class not initialized!"
			<< endl;
		exit(1);
	}

	if (!log_syslog && !log_stderr)
		exit(1);

	va_start(ap, fmt);
	makeMessageAndLog(func, 0, fmt, ap, LOG_ERR);
	va_end(ap);

	exit(1);
}

void
XSI_Log::err(const char *func, int error, const char *fmt, ...)
{
	va_list ap;

	if (!initialized) {
		cerr << "Internal error: XSI_Log class not initialized!"
			<< endl;
		exit(1);
	}

	if (!log_syslog && !log_stderr)
		exit(1);

	va_start(ap, fmt);
	makeMessageAndLog(func, error, fmt, ap, LOG_ERR);
	va_end(ap);

	exit(1);
}

void
XSI_Log::crash(int signal)
{
#define CRASH_MSG	"[ERROR] XMLSysInfo terminated abnormally: "
	if (log_syslog) {
#ifdef HAVE_STRSIGNAL
		syslog(LOG_CRIT, CRASH_MSG "signal %d (%s)", signal, strsignal(signal));
#else
		syslog(LOG_CRIT, CRASH_MSG "signal %d", signal);
#endif /* HAVE_STRSIGNAL */
	}

	if (log_stderr) {
		cerr << __progname << ": " CRASH_MSG "signal " << signal;
#ifdef HAVE_STRSIGNAL
		cerr << " (" << strsignal(signal) << ")" << endl;
#else
		cerr << endl;
#endif /* HAVE_STRSIGNAL */
	}

	abort();
}

XSI_Log::XSI_Log(void)
{
	initialized = false;
}

void
XSI_Log::makeMessageAndLog(const char *func, int error, const char *fmt,
			   va_list ap, int priority)
{
	const char	truncated[] = "... (message truncated)";
	char		errbuf[ERRBUFSIZE];
	string		msg, caller;
	int		sret;
	int		werror = error;
	int		s_errno = errno;
	bool		internal_error = false;

	caller = "";
	if (func != NULL) {
		caller = func;
		if (caller.size() > 0 && caller[caller.size() - 1] != ')')
			caller += "()";
	}
	if (!caller.empty())
		caller += ":";
	msg = caller;

	if (fmt != NULL) {
		errbuf[0] = '\0';
		sret = vsnprintf(errbuf, sizeof(errbuf), fmt, ap);
		if (sret == -1) {
			snprintf(errbuf, sizeof(errbuf),
				 "Fatal internal error in "
				 "XSI_Log::makeMessageAndLog(): vsnprintf: "
				 "Format/encoding error");
			werror = 0;
			internal_error = true;
			msg = errbuf;
		} else if (strlen(errbuf) > strlen(truncated) &&
			   (sret < 0 || sret >= (int)sizeof(errbuf))) {
			snprintf(&errbuf[strlen(errbuf) - strlen(truncated)],
				 strlen(truncated) + 1, "%s", truncated);
			if (!msg.empty())
				msg += " ";
			msg += errbuf;
		} else {
			if (!msg.empty())
				msg += " ";
			msg += errbuf;
		}
	}

	if (werror > 0) {
		if (!msg.empty() && msg[msg.size() - 1] != ':')
			msg += ":";
		if (!msg.empty())
			msg += " ";
		msg += strerror(werror);
	}

#ifdef HAVE_SYSLOG_SUPPORT
	if (log_syslog)
		syslog(priority, "%s", msg.c_str());
#endif /* HAVE_SYSLOG_SUPPORT */

	if (log_stderr)
		cerr << __progname << ": " << msg << endl;

	errno = s_errno;

	if (internal_error)
		exit(1);
}
