/*	$Kolabore: netbsd.cpp,v 1.84 2007/02/09 13:03:40 maxx Exp $	*/
/*
 * Copyright (c) 2006, 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * netbsd.cpp:
 * Gathering information and stats on NetBSD 3.0 or newer.
 */

#include "log.h"

/* For handling currently unknown sensor types. */
#ifndef ENVSYSUNITNAMES
#  define ENVSYSUNITNAMES
#endif

extern "C" {
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/disk.h>
#include <sys/envsys.h>
#include <sys/proc.h>
#include <sys/statvfs.h>
#include <sys/swap.h>
#include <sys/sysctl.h>
#include <sys/wait.h> // XXX for the ugly hack

#include <machine/cpu.h>

#include <netinet/in.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_ether.h>
#include <net/if_media.h>
#include <net/if_types.h>
#include <net/route.h>

#include <fcntl.h>
#include <ifaddrs.h>
#include <libgen.h> // XXX for the ugly hack
#include <netdb.h>
#include <paths.h>
#include <unistd.h>
}

#include <cerrno>
#include <climits> // XXX for the ugly hack
#include <csignal> // XXX for the ugly hack
#include <cstdio>
#include <cstdlib> // XXX for the ugly hack
#include <ctime>

#include <string>
#include <vector>

#include "netbsd.h"

#define SMALLBUFSIZ	128

using namespace std;

struct netbsd_private {
	uint64_t        last_cp_time[CPUSTATES];
	time_t		boottime;
	uint64_t	pagesize;
};

struct netbsd_interrupt {
	uint64_t	count;
	string		source;
};

struct netbsd_iostat {
	string name;
	uint64_t wbytes;
	uint64_t rbytes;
	uint64_t wxfers;
	uint64_t rxfers;
};

static const struct ifmedia_description ifm_type_descriptions[] =
	IFM_TYPE_DESCRIPTIONS;
static const struct ifmedia_description ifm_subtype_descriptions[] =
	IFM_SUBTYPE_DESCRIPTIONS;
static const struct ifmedia_description ifm_mode_descriptions[] =
	IFM_MODE_DESCRIPTIONS;
static const struct ifmedia_description ifm_option_descriptions[] =
	IFM_OPTION_DESCRIPTIONS;
static const int ifm_status_valid_list[] = IFM_STATUS_VALID_LIST;
static const struct ifmedia_status_description ifm_status_descriptions[] =
	IFM_STATUS_DESCRIPTIONS;

static void xsi_in6_fillscopeid(struct sockaddr_in6 *sin6);

/*
 * XSI Constructor
 *
 * !! WARNING !!
 * !! Runs before dropping privileges when running as root
 * !! WARNING !!
 */
XSI::XSI(void)
{
	priv = NULL;
}

XSI::~XSI(void)
{
	if (priv != NULL)
		delete priv;
}

/*
 * Information read once on startup
 */
void
XSI::collectSysInfo(void)
{
	netbsdInit();

	readSystemInfo();
	readSystemNumCPUs();
	readSystemOSInfo();
}

/*
 * Stats to be updated regularly
 */
void
XSI::updateStats(void)
{
	updateCPUFreq();
	updateCPUStats();
	updateVMStats();
	updateFSStats();
	updateDiskStats();
	updateOSUptime();
	updateLoadAvg();
	updateProcInfo();
	updateInterrupts();
	updateSensors();
	updateNetStats();
}

void
XSI::netbsdInit(void)
{
	int		mib[] = { CTL_KERN, KERN_BOOTTIME };
	struct timeval	tv;
	size_t		size = sizeof(tv);
	static bool	first = true;
	struct uvmexp	uvmexp;

	if (first)
		first = false;
	else
		return;

	if ((priv = new struct netbsd_private) == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	bzero(priv->last_cp_time, sizeof(priv->last_cp_time));
	if (sysctl(mib, 2, &tv, &size, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");
	priv->boottime = tv.tv_sec;
	mib[0] = CTL_VM;
	mib[1] = VM_UVMEXP;
	size = sizeof(uvmexp);
	if (sysctl(mib, 2, &uvmexp, &size, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");
	priv->pagesize = uvmexp.pagesize;
}

string
XSI::getSysctlString(int mib0, int mib1, int mib2, unsigned int len,
		     bool must_exist, const char *caller)
{
	int	mib[] = { mib0, mib1, mib2 };
	size_t	siz;
	string	ret;
	char   *buf;

	if (sysctl(mib, len, NULL, &siz, NULL, 0) == -1) {
		if (errno != EOPNOTSUPP || must_exist)
			XSI_Log::log()->err(caller, errno, "sysctl");
		return ("");
	}

	buf = new char[siz];
	if (buf == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	if (sysctl(mib, len, buf, &siz, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");
	ret = buf;
	delete[] buf;
	return (ret);
}

void
XSI::readSystemInfo(void)
{
	string buf;

	buf = getSysctlString(CTL_HW, HW_MODEL, 0, 2, false, __func__);
	if (!buf.empty())
		system.setModel(buf.c_str());
}

void
XSI::readSystemNumCPUs(void)
{
	int mib[] = { CTL_HW, HW_NCPU };
	int ncpu;
	size_t siz = sizeof(ncpu);

	if (sysctl(mib, 2, &ncpu, &siz, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");
	system.setNumCPUs(ncpu);
}

void
XSI::readSystemOSInfo(void)
{
	char *rel, *fullvers;
	char *sharp, *colon, *ident;
	string vers, kern;

	if(!sysctlStr(&rel, CTL_KERN, KERN_OSRELEASE, -1))
		XSI_Log::log()->err(__func__, errno, "sysctl");

	if(!sysctlStr(&fullvers, CTL_KERN, KERN_VERSION, -1)) {
		delete[] rel;
		XSI_Log::log()->err(__func__, errno, "sysctl");
	}

	if(NULL != (sharp = strchr(fullvers, '#')) &&
	   NULL != (colon = strchr(fullvers, ':')) &&
	   NULL != (ident = strrchr(fullvers, '/'))) {
		char *ii, *date;
		char buf[BUFSIZ];
		char wkday[5], month[5];
		int day, hour, minute, second, year;
		struct tm tm;

		/* snip entire version string in two at the newline */
		for (ii = fullvers; *ii && *ii != '\n'; ii++) {}
		*ii = '\0';

		/* learn the kernel name from the end of the 2nd half */
		ident++;
		for (ii = ident; '\0' != *ii && !isspace(*ii); ii++) {}
		*ii = '\0';

		for (date = colon + 1; isspace(*date); date++) {}

		/* reformat and parse date(1) string; %+ doesn't do it */
		if (sscanf(date, "%4s %4s %d %d:%d:%d %*s %d",
			   (char *)&wkday, (char *)&month, &day,
			   &hour, &minute, &second, &year) != 7)
			XSI_Log::log()->err(__func__, "sscanf");
		snprintf(buf, sizeof(buf), "%s %s %2d %0d:%0d:%0d %d",
			 wkday, month, day, hour, minute, second, year);
		if (strptime(buf, "%a %b %e %H:%M:%S %Y", &tm) == NULL)
			XSI_Log::log()->err(__func__, "strptime");

		/* assemble version string as in /etc/motd */
		vers = string(fullvers, sharp - fullvers) + sharp;

		/* assemble normalized kernel version */
		*colon = '\0';
		if(strftime(buf, 64, "%Y%m%d%H%M", &tm) == 0)
			XSI_Log::log()->err(__func__, "strftime");
		kern = string(ident) + sharp + " " + buf;
	} else
		XSI_Log::log()->err(__func__,
				    "Internal error: Kernel version parser broken");

	system.setOSInfo(NULL, kern.c_str(), vers.c_str());

	delete[] fullvers;
	delete[] rel;
}

void
XSI::updateCPUFreq(void)
{
	int32_t freq;
	size_t siz = sizeof(freq);

	if (sysctlbyname("machdep.powernow.frequency.current", &freq, &siz,
			 NULL, 0) != -1) {
		system.setCPUFreq(freq);
		return;
	}

	if (sysctlbyname("machdep.est.frequency.current", &freq, &siz,
			 NULL, 0) != -1) {
		system.setCPUFreq(freq);
		return;
	}

	/* XXX - Is this in Hz or MHz? */
	if (sysctlbyname("machdep.tm_longrun_frequency", &freq, &siz,
			 NULL, 0) != -1) {
		system.setCPUFreq(freq);
		return;
	}
}

void
XSI::updateCPUStats(void)
{
	uint64_t info[CPUSTATES];
	uint64_t total;
	uint64_t diff[CPUSTATES];
	double percent;
	int i;
	vector<struct cpu> cpus;
	int mib[] = { CTL_KERN, KERN_CP_TIME };
	size_t siz = sizeof(info);

	if (sysctl(mib, 2, info, &siz, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");

	cpus.resize(1);
	cpus[0].num = -1;

	total = 0;
	for (i = 0; i < CPUSTATES; i++) {
		diff[i] = info[i] - priv->last_cp_time[i];
		priv->last_cp_time[i] = info[i];
		total += diff[i];
	}

	if (total > 0)
		percent = 100 / (double)total;
	else
		percent = 0;

	cpus[0].system    = diff[CP_SYS] * percent;
	cpus[0].user      = diff[CP_USER] * percent;
	cpus[0].nice      = diff[CP_NICE] * percent;
	cpus[0].interrupt = diff[CP_INTR] * percent;
	cpus[0].idle      = diff[CP_IDLE] * percent;

	system.setCPUs(cpus);
}

#define dbtoqb(b)	dbtob((int64_t)(b))
void
XSI::updateVMStats(void)
{
	struct vm_info	vm_info;
	struct uvmexp	uvmexp;
	struct swapent *swap;
	size_t		size;
	int		mib[2], num_swapdevs, i;
	uint64_t	swap_used = 0;

	mib[0] = CTL_VM;
	mib[1] = VM_UVMEXP;
	size = sizeof(uvmexp);
	if (sysctl(mib, 2, &uvmexp, &size, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");

	vm_info.ram_total = (uint64_t)uvmexp.npages * priv->pagesize;
	vm_info.ram_free = (uint64_t)uvmexp.free * priv->pagesize;

	vm_info.swap_total = vm_info.swap_free = 0;

	if ((num_swapdevs = swapctl(SWAP_NSWAP, NULL, 0)) < 1) {
		system.setVMInfo(vm_info);
		return;
	}

	swap = new struct swapent[num_swapdevs];
	if (swap == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	if (swapctl(SWAP_STATS, swap, num_swapdevs) < num_swapdevs) {
		delete[] swap;
		XSI_Log::log()->err(__func__, errno, "swapctl");
	}

	for (i = 0; i < num_swapdevs; i++) {
		vm_info.swap_total += dbtoqb(swap[i].se_nblks);
		swap_used += dbtoqb(swap[i].se_inuse);
	}
	vm_info.swap_free = vm_info.swap_total - swap_used;

	system.setVMInfo(vm_info);

	delete[] swap;
}

/*
 * Reason #1 for not chrooting on NetBSD:
 * Outside filesystems are being hidden by the kernel from chroot()ed
 * processes.
 */
void
XSI::updateFSStats(void)
{
	vector<struct fs> filesystems;
	struct statvfs *fsp = NULL;
	int i, num;

	if ((num = getmntinfo(&fsp, ST_NOWAIT)) == 0) {
		if (fsp == NULL)
			XSI_Log::log()->err(__func__, errno, "getmntinfo");
		storage.setFilesystems(filesystems);
	}

	for (i = 0; i < num; i++) {
		struct fs	fs;
		uint64_t	blocksize;

		fs.fs_type = fsp[i].f_fstypename;
		if (fs.fs_type == "kernfs" || fs.fs_type == "procfs" ||
		    fs.fs_type == "tmpfs")
			continue;

		if ((blocksize = fsp[i].f_frsize) == 0)
			blocksize = fsp[i].f_bsize;

		fs.mountpoint = fsp[i].f_mntonname;
		fs.fs_name = fsp[i].f_mntfromname;
		fs.size_total = (uint64_t)fsp[i].f_blocks * blocksize;
		fs.size_free = (uint64_t)fsp[i].f_bfree * blocksize;
		fs.size_avail = (int64_t)fsp[i].f_bavail * (int64_t)blocksize;
		fs.nodes_total = fsp[i].f_files;
		fs.nodes_free = fsp[i].f_ffree;

		filesystems.push_back(fs);
	}

	storage.setFilesystems(filesystems);
}

/*
 * The #ifdef maze required to get I/O statistics on NetBSD is due to their
 * API change between NetBSD 3.0 and 4.0. Let's hope they don't do that again.
 */
void
XSI::updateDiskStats(void)
{
#if defined(HW_DISKSTATS) || defined(HW_IOSTATS)
	int			mib[3];
	unsigned int            ndrives, i;
	size_t			size;
#  ifdef HW_IOSTATS
	struct io_sysctl       *dstats;
#  else
	struct disk_sysctl     *dstats;
#  endif
	vector<XSI_DiskIO>	old_disks = storage.getDisks();
	bool			drop_old_stats = false, disks_changed;
	vector<netbsd_iostat>	diostats;
	vector<netbsd_iostat>::iterator ii;

	{
#  ifdef HW_IOSTATS
		string disknames = getSysctlString(CTL_HW, HW_IOSTATNAMES, 0, 2,
						   true, __func__);
#  else
		string disknames = getSysctlString(CTL_HW, HW_DISKNAMES, 0, 2,
						   true, __func__);
#  endif
		ndrives = (int)split(disknames.c_str()).size();
	}

	if (ndrives == 0) {
		old_disks.clear();
		storage.setDisks(old_disks);
		return;
	}

	mib[0] = CTL_HW;
#  ifdef HW_IOSTATS
	mib[1] = HW_IOSTATS;
	mib[2] = (int)sizeof(struct io_sysctl);
	size = ndrives * sizeof(struct io_sysctl);
	dstats = new struct io_sysctl[ndrives];
#  else
	mib[1] = HW_DISKSTATS;
	mib[2] = (int)sizeof(struct disk_sysctl);
	size = ndrives * sizeof(struct disk_sysctl);
	dstats = new struct disk_sysctl[ndrives];
#  endif
	if (dstats == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	if (sysctl(mib, 3, dstats, &size, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");

	for (i = 0; i < ndrives; i++) {
		struct netbsd_iostat ios;

#  ifdef HW_IOSTATS
		if (dstats[i].type == IOSTAT_NFS)
			continue;

		ios.name = dstats[i].name;
		ios.wbytes = dstats[i].wbytes;
		ios.rbytes = dstats[i].rbytes;
		ios.wxfers = dstats[i].wxfer;
		ios.rxfers = dstats[i].rxfer;
#  else
		ios.name = dstats[i].dk_name;
		ios.wbytes = dstats[i].dk_wbytes;
		ios.rbytes = dstats[i].dk_rbytes;
		ios.wxfers = dstats[i].dk_wxfer;
		ios.rxfers = dstats[i].dk_rxfer;
#  endif

		diostats.push_back(ios);
	}

	delete[] dstats;

	if (diostats.empty()) {
		old_disks.clear();
		storage.setDisks(old_disks);
		return;
	}

	if (diostats.size() != old_disks.size())
		drop_old_stats = true;

	do {
		disks_changed = false;
		if (drop_old_stats) {
			vector<XSI_DiskIO> new_disks;

			for (ii = diostats.begin(); ii != diostats.end(); ii++) {
				XSI_DiskIO disk(ii->name.c_str());

				disk.setWrittenBytes(ii->wbytes);
				disk.setReadBytes(ii->rbytes);
				disk.setWriteTransfers(ii->wxfers);
				disk.setReadTransfers(ii->rxfers);

				new_disks.push_back(disk);
			}
			storage.setDisks(new_disks);
		} else {
			for (i = 0; i < ndrives; i++) {
				if (diostats[i].name != old_disks[i].getDiskName()) {
					drop_old_stats = true;
					disks_changed = true;
					break;
				}
				old_disks[i].setWrittenBytes(diostats[i].wbytes);
				old_disks[i].setReadBytes(diostats[i].rbytes);
				old_disks[i].setWriteTransfers(diostats[i].wxfers);
				old_disks[i].setReadTransfers(diostats[i].rxfers);
			}
			if (!disks_changed)
				storage.setDisks(old_disks);
		}
	} while (disks_changed);
#endif /* HW_DISKSTATS || HW_IOSTATS */
}

void
XSI::updateOSUptime(void)
{
	time_t	now;

	time(&now);
	system.setOSUptime(now - priv->boottime);
}

void
XSI::updateLoadAvg(void)
{
	double v[3];

	if (getloadavg(v, 3) == -1)
		XSI_Log::log()->err(__func__, errno, "getloadavg");

	system.setLoadAvg(v);
}

void
XSI::updateProcInfo(void)
{
	int			i, num_procs, tries;
	int			mib[6];
	struct kinfo_proc2     *proc0, *proc;
	size_t			size;
	bool			retry;
	vector<struct process_info> processes;

	mib[0] = CTL_KERN;
	mib[1] = KERN_PROC2;
	mib[2] = KERN_PROC_ALL;
	mib[3] = 0;
	mib[4] = (int)sizeof(struct kinfo_proc2);
	mib[5] = 0;
	proc0 = NULL;
	tries = 0;
	do {
		retry = false;
		if (proc0 != NULL)
			delete[] proc0;

		size = 0;
		if (sysctl(mib, 6, NULL, &size, NULL, 0) == -1)
			XSI_Log::log()->err(__func__, errno, "sysctl (1)");
		if ((proc0 = new struct kinfo_proc2[size]) == NULL)
			XSI_Log::log()->err(__func__, "Out of memory");
		num_procs = (int)(size / sizeof(struct kinfo_proc2));
		mib[5] = num_procs;
		if (sysctl(mib, 6, proc0, &size, NULL, 0) == -1) {
			if (errno == ENOMEM && tries < 100) {
				delete[] proc0;
				proc0 = NULL;
				tries++;
				retry = true;
			} else
				XSI_Log::log()->err(__func__, errno, "sysctl (2)");
		}
	} while (retry);

	num_procs = (int)(size / sizeof(struct kinfo_proc2));
	for (i = 0, proc = proc0; i < num_procs; i++, proc++) {
		struct process_info process;

		if (proc->p__pgid == 0)
			continue;

		process.pid = proc->p_pid;
		process.name = proc->p_comm;
		process.euid = proc->p_uid;
		process.username = user_cache->getUsername(process.euid);
		process.login_user = (proc->p_login[0]) ? proc->p_login : "";
		process.running = (proc->p_stat == LSRUN || proc->p_stat == LSONPROC);
		process.zombie = P_ZOMBIE(proc);
		process.pctcpu = 100.0 * (double)proc->p_pctcpu / (double)FSCALE;
		process.ressize = (uint64_t)proc->p_vm_rssize * priv->pagesize;
		process.nice = proc->p_nice - 20;

		processes.push_back(process);
	}

	delete[] proc0;

	proc_info.setProcessVector(processes);
}

// BEGIN UGLY HACK

#ifdef HAVE_VMSTAT_UTILITY

static pid_t spawnOfCthulhu(int);
static void chldHandler(int);
static int checkChild(pid_t, const char *);
volatile sig_atomic_t sigChld;
volatile sig_atomic_t error;

const char vmstat_utility[] = HAVE_VMSTAT_UTILITY;

/*
 * XXXXXXXXXX!
 *
 * This is a very nasty hack to get interrupt names and -counters.
 *
 * Parse the output from spawnOfCthulhu() (which is the output of
 * `/usr/bin/vmstat -i`.) Vmstat is sgid kmem and therefore able to use KVM and
 * event counters. I don't want XSI to ever need to run with even slightly
 * elevated privileges, so this avenue is closed to XSI itself.
 *
 * There may be better alternatives, such as somehow extracting that
 * information from /dev/ksyms(4), via nlist(3) and/or ioctl(2). There,
 * interesting symbol names would be "[e]intrnames" and "[e]intrcnt". So far,
 * I tried hard to do that without any success ... I don't understand that API
 * at all and need help with it from a NetBSD guru. A working patch for this
 * would go in immediately and there's so much glory and fame to get from this!
 * Please submit a patch. :-)
 *
 * This is currently worse than all of the Linux parsing-of-files crap. :<
 */
void
XSI::updateInterrupts(void)
{
	vector<XSI_Interrupt>	old_ints = system.getInterrupts();
	bool			drop_old_stats = false, ints_changed;

	int	pipe_fds[2];
	int	pRead, pWrite;
	FILE   *pipef;
	pid_t	pid;
	int	c;
	string	buf;
	vector<netbsd_interrupt> intrbuf;
	vector<netbsd_interrupt>::iterator ii;

	if (pipe(pipe_fds) == -1)
		XSI_Log::log()->err(__func__, errno, "pipe");

	pRead = pipe_fds[0];
	pWrite = pipe_fds[1];

	pid = spawnOfCthulhu(pWrite);

	// parent process

	signal(SIGCHLD, chldHandler);

	close(pWrite);

	if ((pipef = fdopen(pRead, "r")) == NULL)
		XSI_Log::log()->err(__func__, errno, "fdopen");
	while ((c = getc(pipef)) != EOF)
		buf.push_back((char)c);
	fclose(pipef);

	signal(SIGCHLD, SIG_IGN);

	if (sigChld) {
		if (checkChild(pid, vmstat_utility) == 0) {
			kill(pid, SIGTERM);
			error = 1;
		}
		sigChld = 0;
	}

	do {
		if ((pid = wait(NULL)) == -1 &&
		    errno != EINTR && errno != ECHILD)
			XSI_Log::log()->err(__func__, errno, "wait");
	} while (pid != -1 || (pid == -1 && errno == EINTR));

	signal(SIGCHLD, SIG_DFL);

	if (error)
		XSI_Log::log()->err(__func__, "%s exited with an error. See error message above, if any.",
				    vmstat_utility);

	{
		vector<string> bufvect = split(buf, '\n');
		vector<string>::iterator si;

		for (si = bufvect.begin(); si != bufvect.end(); si++) {
			vector<string> tmp = split(*si);
			string numstr = tmp[tmp.size() - 2];
			struct netbsd_interrupt intr;
			char *ep;
			unsigned int i;

			if (tmp.size() <= 3 || tmp[0] == "interrupt" ||
			    tmp[0] == "Total")
				continue;

			intr.count = strtoull(numstr.c_str(), &ep, 10);
			if ((intr.count == ULLONG_MAX && errno == ERANGE)
			    || numstr.c_str() == ep || *ep != '\0')
				XSI_Log::log()->err(__func__, "strtoull: Error reading %s",
				     numstr.c_str());

			intr.source = "";
			for (i = 0; i < tmp.size() - 2; i++) {
				intr.source += tmp[i];
				if (i == 0)
					intr.source += ":";
				else if (i < tmp.size() - 3)
					intr.source += "_";
			}

			intrbuf.push_back(intr);
		}
	}

	if (intrbuf.size() != old_ints.size())
		drop_old_stats = true;

	do {
		ints_changed = false;
		if (drop_old_stats) {
			vector<XSI_Interrupt> new_ints;

			for (ii = intrbuf.begin(); ii != intrbuf.end(); ii++) {
				XSI_Interrupt intr(ii->source.c_str());

				intr.setIntsCount(ii->count);

				new_ints.push_back(intr);
			}
			system.setInterrupts(new_ints);
		} else {
			for (unsigned int i = 0; i < intrbuf.size(); i++) {
				if (intrbuf[i].source != old_ints[i].getSourceName()) {
					ints_changed = true;
					drop_old_stats = true;
					break;
				}
				old_ints[i].setIntsCount(intrbuf[i].count);
			}
			if (!ints_changed)
				system.setInterrupts(old_ints);
		}
	} while (ints_changed);
}

/*
 * XXX ... Uargh. :-(
 */
static pid_t
spawnOfCthulhu(int pWrite)
{
	pid_t	pid;

	switch (pid = fork()) {
	case -1:
		XSI_Log::log()->err(__func__, errno, "fork");
	case 0:
		setproctitle("vmstat(8) launcher");
		break;
	default:
		return (pid);
	}

	// child process

	close(STDOUT_FILENO);
	if (dup2(pWrite, STDOUT_FILENO) == -1)
		XSI_Log::log()->err(__func__, errno, "dup2");
	close(pWrite);

	if (execl(vmstat_utility, basename(const_cast<char *>(vmstat_utility)),
		  "-i", NULL) == -1) {
		if (errno == ENOENT)
			exit(0);
		else
			XSI_Log::log()->err(__func__, errno, "execl");
	}

	/* NOTREACHED */
	/* Squelch warnings: */
	exit(1);
}

static void
chldHandler(int sig)
{
	if (sig == SIGCHLD)
		sigChld = 1;
}

static int
checkChild(pid_t pid, const char *name)
{
	int status, ret = 0;

	if (waitpid(pid, &status, WNOHANG) > 0) {
		if (WIFEXITED(status)) {
			if (WEXITSTATUS(status)) {
				XSI_Log::log()->warn(NULL, "%s exited with error code %d",
						     name, WEXITSTATUS(status));
				error = 1;
			}
			ret = 1;
		}
		if (WIFSIGNALED(status)) {
			XSI_Log::log()->warn(NULL, "%s exited due to signal %d",
					     name, WTERMSIG(status));
			error = 1;
			ret = 1;
		}
	}

	return (ret);
}

#else

void XSI::updateInterrupts(void) {}

#endif /* HAVE_VMSTAT_UTILITY */

// END UGLY HACK

void
XSI::updateSensors(void)
{
	int			fd;
	int			i;
	int			invalid_count;
	envsys_tre_data_t	etd;
	envsys_basic_info_t	ebi;

	sensors.clear();

	if ((fd = open(_PATH_SYSMON, O_RDONLY)) == -1) {
		if (errno == ENODEV || errno == ENOENT || errno == ENXIO)
			/*
			 * /dev/sysmon does not exist or isn't configured:
			 * Ignore it.
			 */
			return;
		else
			XSI_Log::log()->err(__func__, errno, _PATH_SYSMON);
	}

	/* Cap the overall number of sensors we're trying to check at 256. */
	invalid_count = 0;
	for (i = 0; i < 256; i++) {
		XSI_Sensor s;

		bzero(&etd, sizeof(etd));
		bzero(&ebi, sizeof(ebi));
		etd.sensor = ebi.sensor = i;

		if (ioctl(fd, ENVSYS_GTREDATA, &etd) == -1)
			XSI_Log::log()->err(__func__, errno, "ENVSYS_GTREDATA");

		/*
		 * Check up to 16 additional sensors when encountering an
		 * invalid sensor, to make sure we don't miss any.
		 */
		if ((etd.validflags & ENVSYS_FVALID) != true) {
			if (invalid_count < 16) {
				invalid_count++;
				continue;
			} else
				break;
		}
		invalid_count = 0;

		if (ioctl(fd, ENVSYS_GTREINFO, &ebi) == -1)
			XSI_Log::log()->err(__func__, errno, "ENVSYS_GTREINFO");

		if ((etd.warnflags & ENVSYS_WARN_CRITUNDER) ||
		    (etd.warnflags & ENVSYS_WARN_CRITOVER))
			s.setSensorStatus(XSI_SENSOR_STATUS_CRIT);
		else if ((etd.warnflags & ENVSYS_WARN_UNDER) ||
			 (etd.warnflags & ENVSYS_WARN_OVER))
			s.setSensorStatus(XSI_SENSOR_STATUS_WARN);
		else if (etd.warnflags & ENVSYS_WARN_OK)
			s.setSensorStatus(XSI_SENSOR_STATUS_OK);
		else
			s.setSensorStatus(XSI_SENSOR_STATUS_UNSPEC);

		s.setSensorID(i);
		s.setSensorDevice("envsys");
		s.setHasUnknownValue((etd.validflags & ENVSYS_FCURVALID) == 0);
		if (ebi.desc[0] != '\0')
			s.setSensorDescription(ebi.desc);

		switch (etd.units) {
		case ENVSYS_STEMP:
			s.setSensorType(XSI_SENSOR_TYPE_TEMP);
			s.setSensorValue((int64_t)etd.cur.data_us);
			break;
		case ENVSYS_SFANRPM:
			s.setSensorType(XSI_SENSOR_TYPE_FANRPM);
			s.setSensorValue((int64_t)etd.cur.data_us);
			break;
		case ENVSYS_SVOLTS_AC:
			s.setSensorType(XSI_SENSOR_TYPE_VOLTS_AC);
			s.setSensorValue((int64_t)etd.cur.data_s);
			break;
		case ENVSYS_SVOLTS_DC:
			s.setSensorType(XSI_SENSOR_TYPE_VOLTS_DC);
			s.setSensorValue((int64_t)etd.cur.data_s);
			break;
		case ENVSYS_SOHMS:
			s.setSensorType(XSI_SENSOR_TYPE_OHMS);
			s.setSensorValue((int64_t)etd.cur.data_s);
			break;
		case ENVSYS_SWATTS:
			s.setSensorType(XSI_SENSOR_TYPE_WATTS);
			s.setSensorValue((int64_t)etd.cur.data_s);
			break;
		case ENVSYS_SAMPS:
			s.setSensorType(XSI_SENSOR_TYPE_AMPS);
			s.setSensorValue((int64_t)etd.cur.data_s);
			break;
		case ENVSYS_SWATTHOUR:
			s.setSensorType(XSI_SENSOR_TYPE_WATTHOUR);
			s.setSensorValue((int64_t)etd.cur.data_s);
			break;
		case ENVSYS_SAMPHOUR:
			s.setSensorType(XSI_SENSOR_TYPE_AMPHOUR);
			s.setSensorValue((int64_t)etd.cur.data_s);
			break;
		case ENVSYS_INDICATOR:
			s.setSensorType(XSI_SENSOR_TYPE_INDICATOR);
			s.setSensorValue((etd.cur.data_s > 0) ? 1 : 0);
			break;
		case ENVSYS_INTEGER:
			s.setSensorType(XSI_SENSOR_TYPE_RAW_INT);
			s.setSensorValue((int64_t)etd.cur.data_s);
			break;
		default:
			s.setSensorType(XSI_SENSOR_TYPE_OTHER);
			/* Best effort: Set the type to the "unit". */
			s.setOtherSensorType(envsysunitnames[(int)etd.units]);
			s.setSensorValue((int64_t)etd.cur.data_s);
			break;
		}

		sensors.push_back(s);
	}

	close(fd);
}

void
XSI::updateNetStats(void)
{
	struct ifaddrs	       *ifap, *ifa;
	vector<struct ifaddrs *> link_ifs;
	vector<struct ifaddrs *>::iterator it;
	vector<XSI_NetIfStat>	old_nstats = network.getNetIfStats();
	bool			drop_old_stats = false, stats_changed;
	int			i;

	if (getifaddrs(&ifap) == -1)
		XSI_Log::log()->err(__func__, errno, "getifaddrs");

	/* Convenience; make a vector of pointers to link-layer interfaces */
	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
		if (ifa->ifa_addr->sa_family == AF_LINK)
			link_ifs.push_back(ifa);
	}

	/* Loopback should always be there, right ...? */
	if (link_ifs.size() == 0) {
		freeifaddrs(ifap);
		return;
	}

	if (link_ifs.size() != old_nstats.size())
		drop_old_stats = true;

	do {
		stats_changed = false;
		if (drop_old_stats) {
			vector<XSI_NetIfStat> new_nstats;

			for (it = link_ifs.begin(); it != link_ifs.end(); it++) {
				XSI_NetIfStat	netIfStat((*it)->ifa_name,
							  (*it)->ifa_flags & IFF_UP);
				struct if_data *ifd;

				netIfStat.setIsRunning((*it)->ifa_flags & IFF_RUNNING);
				netIfStat.setIsPromiscuous((*it)->ifa_flags & IFF_PROMISC);
				netIfStat.setIsLoopback((*it)->ifa_flags & IFF_LOOPBACK);
				getNetLLAddr(netIfStat, (struct sockaddr_dl *)(*it)->ifa_addr);
				getNetIfInfo(netIfStat, (*it));
				getNetIfmediaInfo(netIfStat);
				getNetIfAddrs(netIfStat, ifap, (*it)->ifa_name);
				getDefaultRoutes();

				ifd = (struct if_data *)(*it)->ifa_data;

/*
 * At this point LINK_STATE_IS_UP is OpenBSD-only. Considering the code-sharing
 * between the projects, anticipate it anyways.
 */
#ifdef LINK_STATE_IS_UP
				if (LINK_STATE_IS_UP(ifd->ifi_link_state))
#else
				if (ifd->ifi_link_state == LINK_STATE_UP)
#endif /* LINK_STATE_IS_UP */
					netIfStat.setLinkStatus(IF_LINK_UP);
				else if (ifd->ifi_link_state == LINK_STATE_DOWN)
					netIfStat.setLinkStatus(IF_LINK_DOWN);
				else
					netIfStat.setLinkStatus(IF_LINK_UNKNOWN);

				netIfStat.setPacketsIn(ifd->ifi_ipackets);
				netIfStat.setPacketsOut(ifd->ifi_opackets);
				netIfStat.setErrorsIn(ifd->ifi_ierrors);
				netIfStat.setErrorsOut(ifd->ifi_oerrors);
				netIfStat.setBytesIn(ifd->ifi_ibytes);
				netIfStat.setBytesOut(ifd->ifi_obytes);
				netIfStat.setPacketsDropped(ifd->ifi_iqdrops);
				netIfStat.setCollisions(ifd->ifi_collisions);

				new_nstats.push_back(netIfStat);
			}
			network.setNetIfStats(new_nstats);
		} else {
			for (it = link_ifs.begin(), i = 0; it != link_ifs.end(); it++, i++) {
				struct if_data	       *ifd;

				if (strcmp(old_nstats[i].getIfName(), (*it)->ifa_name) != 0) {
					drop_old_stats = true;
					stats_changed = true;
					break;
				}

				old_nstats[i].setIsUp((*it)->ifa_flags & IFF_UP);
				old_nstats[i].setIsRunning((*it)->ifa_flags & IFF_RUNNING);
				old_nstats[i].setIsPromiscuous((*it)->ifa_flags & IFF_PROMISC);
				old_nstats[i].setIsLoopback((*it)->ifa_flags & IFF_LOOPBACK);
				getNetLLAddr(old_nstats[i], (struct sockaddr_dl *)(*it)->ifa_addr);
				getNetIfInfo(old_nstats[i], (*it));
				getNetIfmediaInfo(old_nstats[i]);
				getNetIfAddrs(old_nstats[i], ifap, old_nstats[i].getIfName());
				getDefaultRoutes();

				ifd = (struct if_data *)(*it)->ifa_data;

#ifdef LINK_STATE_IS_UP
				if (LINK_STATE_IS_UP(ifd->ifi_link_state))
#else
				if (ifd->ifi_link_state == LINK_STATE_UP)
#endif /* LINK_STATE_IS_UP */
					old_nstats[i].setLinkStatus(IF_LINK_UP);
				else if (ifd->ifi_link_state == LINK_STATE_DOWN)
					old_nstats[i].setLinkStatus(IF_LINK_DOWN);
				else
					old_nstats[i].setLinkStatus(IF_LINK_UNKNOWN);

				old_nstats[i].setPacketsIn(ifd->ifi_ipackets);
				old_nstats[i].setPacketsOut(ifd->ifi_opackets);
				old_nstats[i].setErrorsIn(ifd->ifi_ierrors);
				old_nstats[i].setErrorsOut(ifd->ifi_oerrors);
				old_nstats[i].setBytesIn(ifd->ifi_ibytes);
				old_nstats[i].setBytesOut(ifd->ifi_obytes);
				old_nstats[i].setPacketsDropped(ifd->ifi_iqdrops);
				old_nstats[i].setCollisions(ifd->ifi_collisions);
			}
			if (!stats_changed)
				network.setNetIfStats(old_nstats);
		}
	} while (stats_changed);

	freeifaddrs(ifap);
}

void
XSI::getNetLLAddr(XSI_NetIfStat &netIfStat, struct sockaddr_dl *sdl)
{
	if (sdl->sdl_alen) {
		switch (sdl->sdl_type) {
		case IFT_ETHER:
#if defined(IFT_CARP)
		case IFT_CARP:
#endif /* IFT_CARP */
			netIfStat.setLladdrString(ether_ntoa((struct ether_addr *)LLADDR(sdl)));
			return;
		default:
			break;
		}
	}
}

void
XSI::getNetIfInfo(XSI_NetIfStat &netIfStat, struct ifaddrs *ifaddr)
{
	struct ifreq	ifr, *ifrp;
	int		sock;

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
		XSI_Log::log()->err(__func__, errno, "socket");

	bzero(&ifr, sizeof(ifr));
	memcpy(&ifr.ifr_addr, ifaddr->ifa_addr,
	       MIN(sizeof(ifr.ifr_addr), ifaddr->ifa_addr->sa_len));
	ifrp = &ifr;
	strlcpy(ifrp->ifr_name, ifaddr->ifa_name, sizeof(ifrp->ifr_name));
	if (ioctl(sock, SIOCGIFMETRIC, (caddr_t)&ifr) != -1)
		netIfStat.setMetric(ifr.ifr_metric);
	if (ioctl(sock, SIOCGIFMTU, (caddr_t)&ifr) != -1)
		netIfStat.setMTU(ifr.ifr_mtu);

	close(sock);
}

void
XSI::getNetIfmediaInfo(XSI_NetIfStat &netIfStat)
{
	struct ifmediareq	ifmr;
	int			sock, *media_list;
	const char	       *ifname = netIfStat.getIfName();
	static bool		have_warned = false;

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
		XSI_Log::log()->err(__func__, errno, "socket");

	bzero(&ifmr, sizeof(ifmr));
	strlcpy(ifmr.ifm_name, ifname, sizeof(ifmr.ifm_name));
	if (ioctl(sock, SIOCGIFMEDIA, (caddr_t)&ifmr) == -1) {
		// Interface doesn't support media directly
		close(sock);
		return;
	} else {
		string mediaStr;

		if (ifmr.ifm_count == 0) {
			if (!have_warned) {
				XSI_Log::log()->warn(__func__,
						     "%s: no media types?",
						     ifname);
				have_warned = true;
			}
			close(sock);
			return;
		}

		media_list = new int[ifmr.ifm_count];
		ifmr.ifm_ulist = media_list;
		if (ioctl(sock, SIOCGIFMEDIA, (caddr_t)&ifmr) == -1) {
			close(sock);
			XSI_Log::log()->err(__func__, errno, "SIOCGIFMEDIA");
		}

		mediaStr = getMediaWordString(ifmr.ifm_current, 1);
		if (ifmr.ifm_active != ifmr.ifm_current) {
			mediaStr += " (";
			mediaStr += getMediaWordString(ifmr.ifm_active, 0);
			mediaStr += ")";
		}

		netIfStat.setMediaString(mediaStr.c_str());

		delete[] media_list;
	}

	close(sock);
}

void
XSI::getNetIfAddrs(XSI_NetIfStat &netIfStat, struct ifaddrs *ifap, string ifName)
{
	struct ifaddrs	       *ifa;
	vector<XSI_IfAddress>	addrs;
	static bool		have_warned = false;

	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
		XSI_IfAddress addr;

		if (ifName != ifa->ifa_name
		    || !ifa->ifa_addr
		    || ifa->ifa_addr->sa_family == AF_LINK)
			continue;

		switch (ifa->ifa_addr->sa_family) {
		case AF_INET6:
			xsi_in6_fillscopeid((struct sockaddr_in6 *)ifa->ifa_addr);
			if (ifa->ifa_flags & IFF_POINTOPOINT && ifa->ifa_dstaddr)
				xsi_in6_fillscopeid((struct sockaddr_in6 *)ifa->ifa_dstaddr);
			/* FALLTHROUGH */
		case AF_INET:
			addr.setAddress((struct sockaddr *)ifa->ifa_addr,
					(size_t)ifa->ifa_addr->sa_len,
					(struct sockaddr *)ifa->ifa_netmask,
					ifa->ifa_addr->sa_family);
			if (ifa->ifa_flags & IFF_POINTOPOINT && ifa->ifa_dstaddr)
				netIfStat.setPTPDestAddress((struct sockaddr *)ifa->ifa_dstaddr,
							    ifa->ifa_dstaddr->sa_len);
			break;
		default:
			if (!have_warned) {
				XSI_Log::log()->warn(__func__,
						     "Unhandled address family %d found, please report this!",
						     ifa->ifa_addr->sa_family);
				have_warned = true;
			}
			break;
		}

		addrs.push_back(addr);
	}

	netIfStat.setIfAddresses(addrs);
}

#define ROUNDUP(a) \
	((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
void
XSI::getDefaultRoutes(void)
{
	int			mib[6];
	size_t			siz;
	char		       *buf, *next, *ep = NULL;
	struct rt_msghdr       *rtm;
	struct sockaddr        *sa, *rti_info[RTAX_MAX];
	char			host[MAXHOSTNAMELEN];
	int			error;
	vector<struct default_route> defRoutes;

	mib[0] = CTL_NET;
	mib[1] = PF_ROUTE;
	mib[2] = 0;	// protocol number
	mib[3] = 0;	// AF
	mib[4] = NET_RT_FLAGS;
	mib[5] = RTF_GATEWAY;
	if (sysctl(mib, 6, NULL, &siz, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");

	if (siz == 0)
		return;

	buf = new char[siz];
	if (buf == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	if (sysctl(mib, 6, buf, &siz, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");

	ep = buf + siz;
	for (next = buf; next < ep; next += rtm->rtm_msglen) {
		unsigned int		i;
		struct default_route	defRt;

		rtm = (struct rt_msghdr *)next;
		sa = (struct sockaddr *)(rtm + 1);
		for (i = 0; i < RTAX_MAX; i++) {
			if (rtm->rtm_addrs & (1 << i)) {
				rti_info[i] = sa;
				sa = (struct sockaddr *)((char *)(sa) +
							 ROUNDUP(sa->sa_len));
			} else
				rti_info[i] = NULL;
		}
		if (rti_info[RTAX_DST] == NULL || rti_info[RTAX_GATEWAY] == NULL)
			continue;
		if (rti_info[RTAX_DST]->sa_family == AF_INET) {
			if (((struct sockaddr_in *)rti_info[RTAX_DST])->sin_addr.s_addr == INADDR_ANY)
				defRt.family = XSI_RTAF_IPV4;
			else
				continue;
		} else if (rti_info[RTAX_DST]->sa_family == AF_INET6) {
			struct sockaddr_in6 sin6;
			bool zero_mask = true;;

			sin6 = *((struct sockaddr_in6 *)rti_info[RTAX_DST]);
			if (rti_info[RTAX_NETMASK]) {
				struct sockaddr_in6 mask;
				unsigned int lim, final = 0, illegal = 0;
				unsigned char *p;

				mask = *((struct sockaddr_in6 *)rti_info[RTAX_NETMASK]);
				lim = mask.sin6_len - offsetof(struct sockaddr_in6, sin6_addr);
				lim = lim < sizeof(struct in6_addr) ? lim : sizeof(struct in6_addr);
				for (p = (unsigned char *)&mask.sin6_addr, i = 0; i < lim; p++) {
					int masks[9] = {0x00, 0x80, 0xc0, 0xe0,
							0xf0, 0xf8, 0xfc, 0xfe,
							0xff};
					int j;

					if (final && *p) {
						illegal++;
						sin6.sin6_addr.s6_addr[i++] = 0x00;
						continue;
					}

					for (j = 0; j < 9; j++) {
						if (*p & 0xff == masks[j])
							break;
					}

					if (j >= 1 && j <= 8) {
						zero_mask = false;
						break;
					}
					if (j == 9)
						illegal++;
					if (j != 8)
						final++;
					if (!illegal)
						sin6.sin6_addr.s6_addr[i++] &= *p;
					else
						sin6.sin6_addr.s6_addr[i++] = 0x00;
				}
				while (i < sizeof(struct in6_addr))
					sin6.sin6_addr.s6_addr[i++] = 0x00;
			} else
				zero_mask = false;

			if (zero_mask && IN6_IS_ADDR_UNSPECIFIED(&sin6.sin6_addr))
				defRt.family = XSI_RTAF_IPV6;
			else
				continue;
		} else
			continue;
		if ((error = getnameinfo(rti_info[RTAX_GATEWAY], rti_info[RTAX_GATEWAY]->sa_len,
					 host, sizeof(host), NULL, 0, NI_NUMERICHOST)) != 0)
			XSI_Log::log()->err(__func__, "getnameinfo: %s",
					    gai_strerror(error));
		defRt.gateway = host;
		defRoutes.push_back(defRt);
	}

	network.setDefaultRoutes(defRoutes);

	delete[] buf;
}

/*
 * Adapted from OpenBSD's src/sbin/ifconfig/ifconfig.c:in6_fillscopeid()
 */
static void
xsi_in6_fillscopeid(struct sockaddr_in6 *sin6)
{
	if (IN6_IS_ADDR_LINKLOCAL(&sin6->sin6_addr)) {
		sin6->sin6_scope_id =
			ntohs(*(uint16_t *)&sin6->sin6_addr.s6_addr[2]);
		sin6->sin6_addr.s6_addr[2] = sin6->sin6_addr.s6_addr[3] = 0;
	}
}

/* convenience functions for retrieving information via sysctl */
u_int
XSI::getmib(int *name, va_list ap)
{
	u_int len;

	for(len = 0; CTL_MAXNAME > len; len++) {
		name[len] = va_arg(ap, int);
		if(0 > name[len])
			break;
	}

	return len;
}

int
XSI::sysctlStr(char **val, ...)
{
	int name[CTL_MAXNAME];
	u_int namelen;
	va_list ap;
	size_t len;

	va_start(ap, val);
	namelen = getmib(name, ap);
	va_end(ap);

	if(0 > sysctl(name, namelen, NULL, &len, NULL, 0))
		return 0;
	*val = new char[len];

	if(0 > sysctl(name, namelen, *val, &len, NULL, 0)) {
		delete[] *val;
		return 0;
	}

	return 1;
}

/*
 * Adapted from OpenBSD's src/sbin/ifconfig/ifconfig.c:print_media_word()
 */
string
XSI::getMediaWordString(int ifmw, int print_type)
{
	const struct ifmedia_description *desc;
	int	seen_option = 0;
	char	tbuf[SMALLBUFSIZ];
	string	ret = "";

	tbuf[0] = '\0';

	if (print_type) {
		bool found = false;

		for (desc = ifm_type_descriptions; desc->ifmt_string != NULL;
		     desc++) {
			if (IFM_TYPE(ifmw) == desc->ifmt_word) {
				found = true;
				snprintf(tbuf, sizeof(tbuf), "%s ", desc->ifmt_string);
				break;
			}
		}
		if (!found)
			return ("unknown");
	}
	ret += tbuf;

	{
		bool found = false;

		for (desc = ifm_subtype_descriptions; desc->ifmt_string != NULL;
		     desc++) {
			if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
			    IFM_SUBTYPE(desc->ifmt_word) == IFM_SUBTYPE(ifmw)) {
				found = true;
				snprintf(tbuf, sizeof(tbuf), "%s", desc->ifmt_string);
				break;
			}
		}
		if (!found)
			strlcpy(tbuf, "unknown subtype", sizeof(tbuf));
	}
	ret += tbuf;

	/* Find mode. */
	if (IFM_MODE(ifmw) != 0) {
		bool found = false;

		for (desc = ifm_mode_descriptions; desc->ifmt_string != NULL;
		     desc++) {
			if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
			    IFM_MODE(ifmw) == IFM_MODE(desc->ifmt_word)) {
				found = true;
				snprintf(tbuf, sizeof(tbuf), " mode %s", desc->ifmt_string);
				break;
			}
		}
		if (found)
			ret += tbuf;
	}

	/* Find options. */
	for (desc = ifm_option_descriptions; desc->ifmt_string != NULL;
	     desc++) {
		if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
		    (IFM_OPTIONS(ifmw) & IFM_OPTIONS(desc->ifmt_word)) != 0 &&
		    (seen_option & IFM_OPTIONS(desc->ifmt_word)) == 0) {
			if (seen_option == 0)
				ret += " ";
			snprintf(tbuf, sizeof(tbuf), "%s%s", seen_option ? "," : "", desc->ifmt_string);
			ret += tbuf;
			seen_option |= IFM_OPTIONS(desc->ifmt_word);
		}
	}
	if (IFM_INST(ifmw) != 0) {
		snprintf(tbuf, sizeof(tbuf), " instance %d", IFM_INST(ifmw));
		ret += tbuf;
	}

	return (ret);
}
