/*	$Kolabore: openbsd.cpp,v 1.124 2007/02/09 13:06:06 maxx Exp $	*/
/*
 * Copyright (c) 2006, 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * openbsd.cpp:
 * Gathering information and stats on OpenBSD 4.0 or newer. Contains a few
 * hunks of adapted code from OpenBSD src/sbin/ifconfig/ifconfig.c
 */

#include "log.h"

extern "C" {
#include <sys/param.h>
#include <sys/disk.h>
#include <sys/dkstat.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <sys/proc.h>
#include <sys/queue.h>
#include <sys/sensors.h>
#include <sys/socket.h>
#include <sys/swap.h>
#include <sys/sysctl.h>

#include <dev/biovar.h>

#include <uvm/uvm_param.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_media.h>
#include <net/if_types.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/ip_carp.h>

#include <fcntl.h>
#include <ifaddrs.h>
#include <netdb.h>
#include <unistd.h>
}

#include <iostream>
#include <list>
#include <string>
#include <vector>

#include <cerrno>
#include <climits>
#include <cstdio>
#include <cstring>
#include <ctime>

#include "openbsd.h"

#define SMALLBUFSIZ	128
#define dbtoqb(b)	dbtob((int64_t)(b))

extern int errno;

struct openbsd_private {
	time_t		boottime;
	uint64_t	pagesize;
};

static const struct ifmedia_description ifm_type_descriptions[] =
	IFM_TYPE_DESCRIPTIONS;
static const struct ifmedia_description ifm_subtype_descriptions[] =
	IFM_SUBTYPE_DESCRIPTIONS;
static const struct ifmedia_description ifm_mode_descriptions[] =
	IFM_MODE_DESCRIPTIONS;
static const struct ifmedia_description ifm_option_descriptions[] =
	IFM_OPTION_DESCRIPTIONS;
static const int ifm_status_valid_list[] = IFM_STATUS_VALID_LIST;
static const struct ifmedia_status_description ifm_status_descriptions[] =
	IFM_STATUS_DESCRIPTIONS;
static const char *carp_states[] = { CARP_STATES };

static void xsi_in6_fillscopeid(struct sockaddr_in6 *sin6);

/*
 * XSI Constructor
 *
 * !! WARNING !!
 * !! Runs before dropping privileges when running as root
 * !! WARNING !!
 */
XSI::XSI(void)
{
	bio_fd = -1;
	priv = NULL;
}

XSI::~XSI(void)
{
	int i;

	if (priv != NULL)
		delete priv;
	for (i = 0; i < num_cpu; i++) {
		delete[] cur_cp_time[i];
		delete[] last_cp_time[i];
	}
	delete[] cur_cp_time;
	delete[] last_cp_time;
	if (bio_fd != -1)
		close(bio_fd);
}

/*
 * Silently returns 0 on success. Warns and returns >0 on error.
 */
int
XSI::tryOpenPrivilegedDevices(void)
{
	if (configuration->showRaid()) {
		if ((bio_fd = open("/dev/bio", O_RDONLY, 0)) == -1) {
			cerr << "Error: Unable to open /dev/bio for RAID "
				"monitoring: " << strerror(errno) << endl;
			return (1);
		}
	}

	return (0);
}

/*
 * Information read once on startup
 */
void
XSI::collectSysInfo(void)
{
	openbsdInit();

	readSystemInfo();
	readSystemOSInfo();
}

/*
 * Stats to be updated regularly
 */
void
XSI::updateStats(void)
{
	updateCPUFreq();
	updateCPUStats();
	updateVMStats();
	updateFSStats();
	updateDiskStats();
	updateOSUptime();
	updateLoadAvg();
	updateProcInfo();
	updateInterrupts();
	updateSensors();
	updateNetStats();
	updateBioRaidInfo();
}

void
XSI::openbsdInit(void)
{
	int		mib[] = { CTL_KERN, KERN_BOOTTIME };
	int		i, j;
	struct timeval	tv;
	size_t		size = sizeof(tv);
	static bool	first = true;
	struct uvmexp	uvmexp;

	if (first)
		first = false;
	else
		return;

	readSystemNumCPUs();
	cur_cp_time = new int64_t*[num_cpu];
	last_cp_time = new int64_t*[num_cpu];
	if (cur_cp_time == NULL || last_cp_time == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	for (i = 0; i < num_cpu; i++) {
		cur_cp_time[i] = new int64_t[CPUSTATES];
		last_cp_time[i] = new int64_t[CPUSTATES];
		if (cur_cp_time[i] == NULL || last_cp_time[i] == NULL)
			XSI_Log::log()->err(__func__, "Out of memory");
		for (j = 0; j < CPUSTATES; j++) {
			cur_cp_time[i][j] = 0;
			last_cp_time[i][j] = 0;
		}
	}
	if ((priv = new struct openbsd_private) == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	if (sysctl(mib, 2, &tv, &size, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");
	priv->boottime = tv.tv_sec;
	mib[0] = CTL_VM;
	mib[1] = VM_UVMEXP;
	size = sizeof(uvmexp);
	if (sysctl(mib, 2, &uvmexp, &size, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");
	priv->pagesize = uvmexp.pagesize;
}

void
XSI::readSystemInfo(void)
{
	string buf;

	buf = getSysctlString(CTL_HW, HW_MODEL, 0, 2, false, __func__);
	if (!buf.empty())
		system.setModel(buf.c_str());
	buf = getSysctlString(CTL_HW, HW_VENDOR, 0, 2, false, __func__);
	if (!buf.empty())
		system.setVendor(buf.c_str());
	buf = getSysctlString(CTL_HW, HW_PRODUCT, 0, 2, false, __func__);
	if (!buf.empty())
		system.setProduct(buf.c_str());
	buf = getSysctlString(CTL_HW, HW_VERSION, 0, 2, false, __func__);
	if (!buf.empty())
		system.setVersion(buf.c_str());
	buf = getSysctlString(CTL_HW, HW_SERIALNO, 0, 2, false, __func__);
	if (!buf.empty())
		system.setSerialNo(buf.c_str());
	buf = getSysctlString(CTL_HW, HW_UUID, 0, 2, false, __func__);
	if (!buf.empty())
		system.setUUID(buf.c_str());
}

void
XSI::readSystemNumCPUs(void)
{
	int mib[] = { CTL_HW, HW_NCPU };
	int ncpu;
	size_t siz = sizeof(ncpu);

	if (sysctl(mib, 2, &ncpu, &siz, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");
	system.setNumCPUs(ncpu);
	num_cpu = ncpu;
}

void
XSI::readSystemOSInfo(void)
{
	char	       *buf, *bp;
	size_t		size, pos;
	int		mib[2];
	string		os_status, os_kernel, os_version_string, tmp;

	os_kernel = getSysctlString(CTL_KERN, KERN_OSVERSION, 0, 2, true, __func__);

	mib[0] = CTL_KERN;
	mib[1] = KERN_VERSION;
	if (sysctl(mib, 2, NULL, &size, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");
	buf = new char[size];
	if (buf == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	if (sysctl(mib, 2, buf, &size, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");
	bp = buf;
	while (*bp != '\n' && *bp != '\0')
		bp++;
	*bp = '\0';
	os_version_string = buf;
	delete[] buf;

	pos = os_version_string.find(':') + 2;
	tmp = os_version_string.substr(pos);

	{
		char		buf2[BUFSIZ];
		char		wkday[5], month[5];
		int		day, hour, minute, second, year;
		struct tm	tm;

		/* Need to reformat default date(1) string for strptime() */

		if (sscanf(tmp.c_str(), "%4s %4s %d %d:%d:%d %*s %d",
			   (char *)&wkday, (char *)&month, &day,
			   &hour, &minute, &second, &year) != 7)
			XSI_Log::log()->err(__func__, "sscanf");

		snprintf(buf2, sizeof(buf2), "%s %s %2d %0d:%0d:%0d %d",
			 wkday, month, day, hour, minute, second, year);

		if (strptime(buf2, "%a %b %e %H:%M:%S %Y", &tm) == NULL)
			XSI_Log::log()->err(__func__, "strptime");

		if (strftime(buf2, sizeof(buf2), "%Y%m%d%H%M", &tm) == 0)
			XSI_Log::log()->err(__func__, "strftime");

		tmp = buf2;
	}
	os_kernel += " " + tmp;

	pos = os_version_string.size();
	if (os_version_string.find("-current", 0) < pos)
		os_status = "current";
	else if (os_version_string.find("-beta", 0) < pos)
		os_status = "beta";
	else if (os_version_string.find("-stable", 0) < pos)
		os_status = "stable";
	else
		os_status = "release";

	system.setOSInfo(os_status.c_str(), os_kernel.c_str(),
			 os_version_string.c_str());
}

void
XSI::updateCPUFreq(void)
{
	int mib[] = { CTL_HW, HW_CPUSPEED };
	int cpu_freq;
	size_t siz = sizeof(cpu_freq);

	if (sysctl(mib, 2, &cpu_freq, &siz, NULL, 0) == -1) {
		if (errno != EOPNOTSUPP)
			XSI_Log::log()->err(__func__, errno, "sysctl");
		return;
	}
	system.setCPUFreq(cpu_freq);
}

void
XSI::updateCPUStats(void)
{
	int64_t 	       *total;
	int64_t 	      **used;
	size_t			size;
	double		       *percent;
	vector<struct cpu>	cpus;
	int			i, j;

	if (num_cpu > 1) {
		size = CPUSTATES * sizeof(int64_t);
		for (i = 0; i < num_cpu; i++) {
			int mib[] = { CTL_KERN, KERN_CPTIME2, i };

			if (sysctl(mib, 3, cur_cp_time[i], &size, NULL, 0) == -1)
				XSI_Log::log()->err(__func__, errno, "sysctl");
		}
	} else {
		int mib[] = { CTL_KERN, KERN_CPTIME };
		long cp_time_tmp[CPUSTATES];

		size = sizeof(cp_time_tmp);
		if (sysctl(mib, 2, cp_time_tmp, &size, NULL, 0) == -1)
			XSI_Log::log()->err(__func__, errno, "sysctl");
		for (i = 0; i < CPUSTATES; i++)
			cur_cp_time[0][i] = cp_time_tmp[i];
	}

	total = new int64_t[num_cpu];
	used = new int64_t*[num_cpu];
	percent = new double[num_cpu];
	if (total == NULL || used == NULL || percent == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	for (i = 0; i < num_cpu; i++) {
		used[i] = new int64_t[CPUSTATES];
		if (used[i] == NULL)
			XSI_Log::log()->err(__func__, "Out of memory");
	}

	for (i = 0; i < num_cpu; i++) {
		struct cpu	cpu;

		total[i] = 0;
		for (j = 0; j < CPUSTATES; j++) {
			used[i][j] = cur_cp_time[i][j] - last_cp_time[i][j];
			last_cp_time[i][j] = cur_cp_time[i][j];
			total[i] += used[i][j];
		}
		if (total[i] > 0)
			percent[i] = 100 / (double)total[i];
		else
			percent[i] = 0;
		cpu.num = i;
		cpu.system = used[i][CP_SYS] * percent[i];
		cpu.user = used[i][CP_USER] * percent[i];
		cpu.nice = used[i][CP_NICE] * percent[i];
		cpu.interrupt = used[i][CP_INTR] * percent[i];
		cpu.idle = used[i][CP_IDLE] * percent[i];
		cpus.push_back(cpu);
	}

	system.setCPUs(cpus);

	for (i = 0; i < num_cpu; i++)
		delete[] used[i];
	delete[] used;
	delete[] total;
	delete[] percent;
}

void
XSI::updateVMStats(void)
{
	struct vm_info	vm_info;
	struct uvmexp	uvmexp;
	struct swapent *swap;
	size_t		size;
	int		mib[2], num_swapdevs, i;
	uint64_t	swap_used = 0;

	mib[0] = CTL_VM;
	mib[1] = VM_UVMEXP;
	size = sizeof(uvmexp);
	if (sysctl(mib, 2, &uvmexp, &size, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");

	vm_info.ram_total = (uint64_t)uvmexp.npages * priv->pagesize;
	vm_info.ram_free = (uint64_t)uvmexp.free * priv->pagesize;

	vm_info.swap_total = vm_info.swap_free = 0;

	if ((num_swapdevs = swapctl(SWAP_NSWAP, NULL, 0)) < 1) {
		system.setVMInfo(vm_info);
		return;
	}

	swap = new struct swapent[num_swapdevs];
	if (swap == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	if (swapctl(SWAP_STATS, swap, num_swapdevs) < num_swapdevs) {
		delete[] swap;
		XSI_Log::log()->err(__func__, errno, "swapctl");
	}

	for (i = 0; i < num_swapdevs; i++) {
		vm_info.swap_total += dbtoqb(swap[i].se_nblks);
		swap_used += dbtoqb(swap[i].se_inuse);
	}
	vm_info.swap_free = vm_info.swap_total - swap_used;

	system.setVMInfo(vm_info);

	delete[] swap;
}

void
XSI::updateFSStats(void)
{
	struct statfs	       *sfs;
	int			num_mounts, i;
	size_t			size;
	vector<struct fs>	filesystems;

	if ((num_mounts = getfsstat(NULL, 0, MNT_NOWAIT)) == -1)
		XSI_Log::log()->err(__func__, errno, "getfsstat");

	size = num_mounts * sizeof(struct statfs);
	sfs = new struct statfs[num_mounts];
	if (sfs == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");

	if ((num_mounts = getfsstat(sfs, size, MNT_NOWAIT)) == -1)
		XSI_Log::log()->err(__func__, errno, "getfsstat");

	for (i = 0; i < num_mounts; i++) {
		struct fs fs;

		fs.mountpoint = sfs[i].f_mntonname;
		fs.fs_name = sfs[i].f_mntfromname;
		fs.fs_type = sfs[i].f_fstypename;
		fs.size_total = (uint64_t)sfs[i].f_blocks * (uint64_t)sfs[i].f_bsize;
		fs.size_free = (uint64_t)sfs[i].f_bfree * (uint64_t)sfs[i].f_bsize;
		fs.size_avail = (int64_t)sfs[i].f_bavail * (int64_t)sfs[i].f_bsize;
		fs.nodes_total = sfs[i].f_files;
		fs.nodes_free = sfs[i].f_ffree;

		filesystems.push_back(fs);
	}

	storage.setFilesystems(filesystems);

	delete[] sfs;
}

void
XSI::updateDiskStats(void)
{
	int			mib[2], ndrives, i;
	size_t			size;
	struct diskstats       *dstats;
	vector<XSI_DiskIO>	old_disks = storage.getDisks();
	bool			drop_old_stats = false, disks_changed;

	mib[0] = CTL_HW;
	mib[1] = HW_DISKCOUNT;
	size = sizeof(ndrives);
	if (sysctl(mib, 2, &ndrives, &size, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");

	if (ndrives == 0) {
		old_disks.clear();
		storage.setDisks(old_disks);
		return;
	}

	if (ndrives != (int)old_disks.size())
		drop_old_stats = true;

	mib[0] = CTL_HW;
	mib[1] = HW_DISKSTATS;
	size = ndrives * sizeof(struct diskstats);
	dstats = new struct diskstats[ndrives];
	if (dstats == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	if (sysctl(mib, 2, dstats, &size, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");

	do {
		disks_changed = false;
		if (drop_old_stats) {
			vector<XSI_DiskIO> new_disks;

			for (i = 0; i < ndrives; i++) {
				XSI_DiskIO disk(dstats[i].ds_name);

				disk.setWrittenBytes(dstats[i].ds_wbytes);
				disk.setReadBytes(dstats[i].ds_rbytes);
				disk.setWriteTransfers(dstats[i].ds_wxfer);
				disk.setReadTransfers(dstats[i].ds_rxfer);

				new_disks.push_back(disk);
			}
			storage.setDisks(new_disks);
		} else {
			for (i = 0; i < ndrives; i++) {
				if (strcmp(old_disks[i].getDiskName(), dstats[i].ds_name) != 0) {
					drop_old_stats = true;
					disks_changed = true;
					break;
				}
				old_disks[i].setWrittenBytes(dstats[i].ds_wbytes);
				old_disks[i].setReadBytes(dstats[i].ds_rbytes);
				old_disks[i].setWriteTransfers(dstats[i].ds_wxfer);
				old_disks[i].setReadTransfers(dstats[i].ds_rxfer);
			}
			if (!disks_changed)
				storage.setDisks(old_disks);
		}
	} while (disks_changed);

	delete[] dstats;
}

void
XSI::updateOSUptime(void)
{
	time_t	now;

	time(&now);
	system.setOSUptime(now - priv->boottime);
}

void
XSI::updateLoadAvg(void)
{
	double v[3];

	if (getloadavg(v, 3) == -1)
		XSI_Log::log()->err(__func__, errno, "getloadavg");

	system.setLoadAvg(v);
}

void
XSI::updateProcInfo(void)
{
	int			i, num_procs, tries;
	int			mib[6];
	struct kinfo_proc2     *proc0, *proc;
	size_t			size;
	bool			retry;
	vector<struct process_info> processes;

	mib[0] = CTL_KERN;
	mib[1] = KERN_PROC2;
	mib[2] = KERN_PROC_ALL;
	mib[3] = 0;
	mib[4] = (int)sizeof(struct kinfo_proc2);
	mib[5] = 0;
	proc0 = NULL;
	tries = 0;
	do {
		retry = false;
		if (proc0 != NULL)
			delete[] proc0;

		size = 0;
		if (sysctl(mib, 6, NULL, &size, NULL, 0) == -1)
			XSI_Log::log()->err(__func__, errno, "sysctl (1)");
		if ((proc0 = new struct kinfo_proc2[size]) == NULL)
			XSI_Log::log()->err(__func__, "Out of memory");
		num_procs = (int)(size / sizeof(struct kinfo_proc2));
		mib[5] = num_procs;
		if (sysctl(mib, 6, proc0, &size, NULL, 0) == -1) {
			if (errno == ENOMEM && tries < 100) {
				delete[] proc0;
				proc0 = NULL;
				tries++;
				retry = true;
			} else
				XSI_Log::log()->err(__func__, errno, "sysctl (2)");
		}
	} while (retry);

	num_procs = (int)(size / sizeof(struct kinfo_proc2));
	for (i = 0, proc = proc0; i < num_procs; i++, proc++) {
		struct process_info process;

		if (proc->p__pgid == 0)
			continue;

		process.pid = proc->p_pid;
		process.name = proc->p_comm;
		process.euid = proc->p_uid;
		/* This fails by default inside an unpopulated chroot area. */
		process.username = user_cache->getUsername(process.euid);
		process.login_user = (proc->p_login[0]) ? proc->p_login : "";
		process.running = (proc->p_stat == SRUN || proc->p_stat == SONPROC);
		process.zombie = P_ZOMBIE(proc);
		process.pctcpu = 100.0 * (double)proc->p_pctcpu / (double)FSCALE;
		process.ressize = proc->p_vm_rssize * priv->pagesize;
		process.nice = proc->p_nice - 20;

		processes.push_back(process);
	}

	delete[] proc0;

	proc_info.setProcessVector(processes);
}

void
XSI::updateInterrupts(void)
{
	vector<XSI_Interrupt>	old_ints = system.getInterrupts();
	int			mib[4], nintr, i;
	size_t			size;
	bool			drop_old_stats = false, ints_changed;

	mib[0] = CTL_KERN;
	mib[1] = KERN_INTRCNT;
	mib[2] = KERN_INTRCNT_NUM;
	size = sizeof(nintr);
	if (sysctl(mib, 3, &nintr, &size, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl INTRCNT_NUM");

	if (nintr == 0)
		// huhh??
		return;

	if (nintr != (int)old_ints.size())
		drop_old_stats = true;

	do {
		ints_changed = false;
		if (drop_old_stats) {
			vector<XSI_Interrupt> new_ints;

			for (i = 0; i < nintr; i++) {
				XSI_Interrupt	interrupt;
				char		buf[SMALLBUFSIZ];
				uint64_t 	ints;

				mib[0] = CTL_KERN;
				mib[1] = KERN_INTRCNT;
				mib[2] = KERN_INTRCNT_NAME;
				mib[3] = i;
				size = sizeof(buf);
				if (sysctl(mib, 4, buf, &size, NULL, 0) == -1)
					XSI_Log::log()->err(__func__, errno, "sysctl INTRCNT_NAME");
				interrupt.setSourceName(buf);

				mib[0] = CTL_KERN;
				mib[1] = KERN_INTRCNT;
				mib[2] = KERN_INTRCNT_CNT;
				mib[3] = i;
				size = sizeof(ints);
				if (sysctl(mib, 4, &ints, &size, NULL, 0) == -1)
					XSI_Log::log()->err(__func__, errno, "sysctl INTRCNT_CNT");
				interrupt.setIntsCount(ints);

				new_ints.push_back(interrupt);
			}
			system.setInterrupts(new_ints);
		} else {
			for (i = 0; i < nintr; i++) {
				char		buf[SMALLBUFSIZ];
				uint64_t 	ints;

				mib[0] = CTL_KERN;
				mib[1] = KERN_INTRCNT;
				mib[2] = KERN_INTRCNT_NAME;
				mib[3] = i;
				size = sizeof(buf);
				if (sysctl(mib, 4, buf, &size, NULL, 0) == -1)
					XSI_Log::log()->err(__func__, errno, "sysctl INTRCNT_NAME");
				if (strcmp(old_ints[i].getSourceName(), buf) != 0) {
					ints_changed = true;
					drop_old_stats = true;
					break;
				}

				mib[0] = CTL_KERN;
				mib[1] = KERN_INTRCNT;
				mib[2] = KERN_INTRCNT_CNT;
				mib[3] = i;
				size = sizeof(ints);
				if (sysctl(mib, 4, &ints, &size, NULL, 0) == -1)
					XSI_Log::log()->err(__func__, errno, "sysctl INTRCNT_CNT");
				old_ints[i].setIntsCount(ints);
			}
			if (!ints_changed)
				system.setInterrupts(old_ints);
		}
	} while (ints_changed);
}

void
XSI::updateSensors(void)
{
	int			mib[5];
	int			id = 0;


#ifdef MAXSENSORDEVICES
	/* New (post December 23rd 2006) Sensors API */

	int			dev_num;
	struct sensordev	sdev;
	size_t			sdevlen = sizeof(sdev);
	enum sensor_type	type;

	/* Start over, no gain in updating. */
	sensors.clear();

	mib[0] = CTL_HW;
	mib[1] = HW_SENSORS;
	for (dev_num = 0; dev_num < MAXSENSORDEVICES; dev_num++) {
		mib[2] = dev_num;
		if (sysctl(mib, 3, &sdev, &sdevlen, NULL, 0) == -1)
			continue;

		for ((int)type = 0; type < SENSOR_MAX_TYPES; ((int)type)++) {
			mib[3] = type;
			for (id = 0; id < sdev.maxnumt[type]; id++) {
				XSI_Sensor	sensor;
				struct sensor	s;
				size_t		slen = sizeof(s);

				mib[4] = id;
				if (sysctl(mib, 5, &s, &slen, NULL, 0) == -1)
					continue;

				sensor.setSensorDevice(sdev.xname);

#else
	/* Old (pre December 23rd 2006) Sensors API */

	int			i;
	enum sensor_type	old_type = (enum sensor_type)INT_MAX;

	/* Start over, no gain in updating. */
	sensors.clear();

	mib[0] = CTL_HW;
	mib[1] = HW_SENSORS;
	{
		{
			for (i = 0; i < 256; i++) {
				XSI_Sensor	sensor;
				struct sensor	s;
				size_t		slen = sizeof(s);

				mib[2] = i;
				if (sysctl(mib, 3, &s, &slen, NULL, 0) == -1)
					continue;

				if (old_type != s.type) {
					old_type = s.type;
					id = 0;
				} else
					id++;

				sensor.setSensorDevice(s.device);

#endif /* MAXSENSORDEVICES */

				if (slen == 0 || (s.flags & SENSOR_FINVALID))
					continue;

				sensor.setHasUnknownValue(s.flags & SENSOR_FUNKNOWN);
				sensor.setSensorID(id);
				switch (s.type) {
				case SENSOR_TEMP:
					sensor.setSensorType(XSI_SENSOR_TYPE_TEMP);
					sensor.setSensorValue(s.value);
					break;
				case SENSOR_FANRPM:
					sensor.setSensorType(XSI_SENSOR_TYPE_FANRPM);
					sensor.setSensorValue(s.value);
					break;
				case SENSOR_VOLTS_DC:
					sensor.setSensorType(XSI_SENSOR_TYPE_VOLTS_DC);
					sensor.setSensorValue(s.value);
					break;
				case SENSOR_VOLTS_AC:
					sensor.setSensorType(XSI_SENSOR_TYPE_VOLTS_AC);
					sensor.setSensorValue(s.value);
					break;
				case SENSOR_OHMS:
					sensor.setSensorType(XSI_SENSOR_TYPE_OHMS);
					sensor.setSensorValue(s.value);
					break;
				case SENSOR_WATTS:
					sensor.setSensorType(XSI_SENSOR_TYPE_WATTS);
					sensor.setSensorValue(s.value);
					break;
				case SENSOR_AMPS:
					sensor.setSensorType(XSI_SENSOR_TYPE_AMPS);
					sensor.setSensorValue(s.value);
					break;
				case SENSOR_WATTHOUR:
					sensor.setSensorType(XSI_SENSOR_TYPE_WATTHOUR);
					sensor.setSensorValue(s.value);
					break;
				case SENSOR_AMPHOUR:
					sensor.setSensorType(XSI_SENSOR_TYPE_AMPHOUR);
					sensor.setSensorValue(s.value);
					break;
				case SENSOR_INDICATOR:
					sensor.setSensorType(XSI_SENSOR_TYPE_INDICATOR);
					sensor.setSensorValue(s.value ? 1 : 0);
					break;
				case SENSOR_INTEGER:
					sensor.setSensorType(XSI_SENSOR_TYPE_RAW_INT);
					sensor.setSensorValue(s.value);
					break;
				case SENSOR_PERCENT:
					sensor.setSensorType(XSI_SENSOR_TYPE_PERCENT);
					sensor.setSensorValue(s.value);
					break;
				case SENSOR_LUX:
					sensor.setSensorType(XSI_SENSOR_TYPE_LUX);
					sensor.setSensorValue(s.value);
					break;
				case SENSOR_DRIVE:
					sensor.setSensorType(XSI_SENSOR_TYPE_DRIVE);
					switch (s.value) {
					case SENSOR_DRIVE_EMPTY:
						sensor.setSensorValue(XSI_DRIVE_STATUS_EMPTY);
						break;
					case SENSOR_DRIVE_READY:
						sensor.setSensorValue(XSI_DRIVE_STATUS_READY);
						break;
					case SENSOR_DRIVE_POWERUP:
						sensor.setSensorValue(XSI_DRIVE_STATUS_POWERUP);
						break;
					case SENSOR_DRIVE_ONLINE:
						sensor.setSensorValue(XSI_DRIVE_STATUS_ONLINE);
						break;
					case SENSOR_DRIVE_IDLE:
						sensor.setSensorValue(XSI_DRIVE_STATUS_IDLE);
						break;
					case SENSOR_DRIVE_ACTIVE:
						sensor.setSensorValue(XSI_DRIVE_STATUS_ACTIVE);
						break;
					case SENSOR_DRIVE_REBUILD:
						sensor.setSensorValue(XSI_DRIVE_STATUS_REBUILD);
						break;
					case SENSOR_DRIVE_POWERDOWN:
						sensor.setSensorValue(XSI_DRIVE_STATUS_POWERDOWN);
						break;
					case SENSOR_DRIVE_FAIL:
						sensor.setSensorValue(XSI_DRIVE_STATUS_FAIL);
						break;
					case SENSOR_DRIVE_PFAIL:
						sensor.setSensorValue(XSI_DRIVE_STATUS_PFAIL);
						break;
					default:
						sensor.setSensorValue(XSI_DRIVE_STATUS_OTHER);
						break;
					}
					break;
				case SENSOR_TIMEDELTA:
					sensor.setSensorType(XSI_SENSOR_TYPE_TIMEDELTA);
					sensor.setSensorValue(s.value);
					break;
				default:
#ifdef MAXSENSORDEVICES
					sensor.setSensorType(XSI_SENSOR_TYPE_OTHER);
					sensor.setOtherSensorType(sensor_type_s[type]);
					sensor.setSensorValue(s.value);
					break;
#else
					/* Cannot happen. */
					break;
#endif /* MAXSENSORDEVICES */
				}
				if (s.desc[0] != '\0')
					sensor.setSensorDescription(s.desc);
				switch (s.status) {
				case SENSOR_S_OK:
					sensor.setSensorStatus(XSI_SENSOR_STATUS_OK);
					break;
				case SENSOR_S_WARN:
					sensor.setSensorStatus(XSI_SENSOR_STATUS_WARN);
					break;
				case SENSOR_S_CRIT:
					sensor.setSensorStatus(XSI_SENSOR_STATUS_CRIT);
					break;
				case SENSOR_S_UNKNOWN:
					sensor.setSensorStatus(XSI_SENSOR_STATUS_UNKNOWN);
					break;
				case SENSOR_S_UNSPEC:
				default:
					sensor.setSensorStatus(XSI_SENSOR_STATUS_UNSPEC);
					break;
				}
				if (s.tv.tv_sec > 0 || s.tv.tv_usec > 0)
					sensor.setValueLastChangeTime(&s.tv);

				sensors.push_back(sensor);
			}
		}
	}
}

void
XSI::updateNetStats(void)
{
	struct ifaddrs	       *ifap, *ifa;
	vector<struct ifaddrs *> link_ifs;
	vector<struct ifaddrs *>::iterator it;
	vector<XSI_NetIfStat>	old_nstats = network.getNetIfStats();
	bool			drop_old_stats = false, stats_changed;
	int			i;

	if (getifaddrs(&ifap) == -1)
		XSI_Log::log()->err(__func__, errno, "getifaddrs");

	/* Convenience; make a vector of pointers to link-layer interfaces */
	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
		if (ifa->ifa_addr->sa_family == AF_LINK)
			link_ifs.push_back(ifa);
	}

	/* Loopback should always be there, right ...? */
	if (link_ifs.size() == 0) {
		freeifaddrs(ifap);
		return;
	}

	if (link_ifs.size() != old_nstats.size())
		drop_old_stats = true;

	do {
		stats_changed = false;
		if (drop_old_stats) {
			vector<XSI_NetIfStat> new_nstats;

			for (it = link_ifs.begin(); it != link_ifs.end(); it++) {
				XSI_NetIfStat	netIfStat((*it)->ifa_name,
							  (*it)->ifa_flags & IFF_UP);
				struct if_data *ifd;

				netIfStat.setIsRunning((*it)->ifa_flags & IFF_RUNNING);
				netIfStat.setIsPromiscuous((*it)->ifa_flags & IFF_PROMISC);
				netIfStat.setIsLoopback((*it)->ifa_flags & IFF_LOOPBACK);
				getNetLLAddr(netIfStat, (struct sockaddr_dl *)(*it)->ifa_addr);
				getNetIfInfo(netIfStat, (*it), (struct sockaddr_dl *)(*it)->ifa_addr);
				getNetIfmediaInfo(netIfStat);
				getNetIfAddrs(netIfStat, ifap, (*it)->ifa_name);
				getDefaultRoutes();

				ifd = (struct if_data *)(*it)->ifa_data;

#ifdef LINK_STATE_IS_UP
				if (LINK_STATE_IS_UP(ifd->ifi_link_state))
#else
				if (ifd->ifi_link_state == LINK_STATE_UP)
#endif /* LINK_STATE_IS_UP */
					netIfStat.setLinkStatus(IF_LINK_UP);
				else if (ifd->ifi_link_state == LINK_STATE_DOWN)
					netIfStat.setLinkStatus(IF_LINK_DOWN);
				else
					netIfStat.setLinkStatus(IF_LINK_UNKNOWN);

#ifdef LINK_STATE_FULL_DUPLEX
				if (ifd->ifi_link_state == LINK_STATE_FULL_DUPLEX)
					netIfStat.setDuplexStatus(IF_DUPLEX_FULL);
				else if (ifd->ifi_link_state == LINK_STATE_HALF_DUPLEX)
					netIfStat.setDuplexStatus(IF_DUPLEX_HALF);
				else
					netIfStat.setDuplexStatus(IF_DUPLEX_UNKNOWN);
#endif /* LINK_STATE_FULL_DUPLEX */

				netIfStat.setPacketsIn(ifd->ifi_ipackets);
				netIfStat.setPacketsOut(ifd->ifi_opackets);
				netIfStat.setErrorsIn(ifd->ifi_ierrors);
				netIfStat.setErrorsOut(ifd->ifi_oerrors);
				netIfStat.setBytesIn(ifd->ifi_ibytes);
				netIfStat.setBytesOut(ifd->ifi_obytes);
				netIfStat.setPacketsDropped(ifd->ifi_iqdrops);
				netIfStat.setCollisions(ifd->ifi_collisions);

				new_nstats.push_back(netIfStat);
			}
			network.setNetIfStats(new_nstats);
		} else {
			for (it = link_ifs.begin(), i = 0; it != link_ifs.end(); it++, i++) {
				struct if_data	       *ifd;

				if (strcmp(old_nstats[i].getIfName(), (*it)->ifa_name) != 0) {
					drop_old_stats = true;
					stats_changed = true;
					break;
				}

				old_nstats[i].setIsUp((*it)->ifa_flags & IFF_UP);
				old_nstats[i].setIsRunning((*it)->ifa_flags & IFF_RUNNING);
				old_nstats[i].setIsPromiscuous((*it)->ifa_flags & IFF_PROMISC);
				old_nstats[i].setIsLoopback((*it)->ifa_flags & IFF_LOOPBACK);
				getNetLLAddr(old_nstats[i], (struct sockaddr_dl *)(*it)->ifa_addr);
				getNetIfInfo(old_nstats[i], (*it), (struct sockaddr_dl *)(*it)->ifa_addr);
				getNetIfmediaInfo(old_nstats[i]);
				getNetIfAddrs(old_nstats[i], ifap, old_nstats[i].getIfName());
				getDefaultRoutes();

				ifd = (struct if_data *)(*it)->ifa_data;

#ifdef LINK_STATE_IS_UP
				if (LINK_STATE_IS_UP(ifd->ifi_link_state))
#else
				if (ifd->ifi_link_state == LINK_STATE_UP)
#endif /* LINK_STATE_IS_UP */
					old_nstats[i].setLinkStatus(IF_LINK_UP);
				else if (ifd->ifi_link_state == LINK_STATE_DOWN)
					old_nstats[i].setLinkStatus(IF_LINK_DOWN);
				else
					old_nstats[i].setLinkStatus(IF_LINK_UNKNOWN);

#ifdef LINK_STATE_FULL_DUPLEX
				if (ifd->ifi_link_state == LINK_STATE_FULL_DUPLEX)
					old_nstats[i].setDuplexStatus(IF_DUPLEX_FULL);
				else if (ifd->ifi_link_state == LINK_STATE_HALF_DUPLEX)
					old_nstats[i].setDuplexStatus(IF_DUPLEX_HALF);
				else
					old_nstats[i].setDuplexStatus(IF_DUPLEX_UNKNOWN);
#endif /* LINK_STATE_FULL_DUPLEX */

				old_nstats[i].setPacketsIn(ifd->ifi_ipackets);
				old_nstats[i].setPacketsOut(ifd->ifi_opackets);
				old_nstats[i].setErrorsIn(ifd->ifi_ierrors);
				old_nstats[i].setErrorsOut(ifd->ifi_oerrors);
				old_nstats[i].setBytesIn(ifd->ifi_ibytes);
				old_nstats[i].setBytesOut(ifd->ifi_obytes);
				old_nstats[i].setPacketsDropped(ifd->ifi_iqdrops);
				old_nstats[i].setCollisions(ifd->ifi_collisions);
			}
			if (!stats_changed)
				network.setNetIfStats(old_nstats);
		}
	} while (stats_changed);

	freeifaddrs(ifap);
}

/*
 * updateRaidInfo() reads RAID array information via bio(4), using code
 * inspired by src/sbin/bioctl/bioctl.c.
 */
void
XSI::updateBioRaidInfo(void)
{
	list<string>		controllers;
	list<string>::iterator	i;
	vector<XSI_Raid>	raids_tmp;

	if (bio_fd == -1)
		return;

	controllers = configuration->getRaidControllers();
	for (i = controllers.begin(); i != controllers.end(); i++) {
		XSI_Raid		raid(i->c_str());
		vector<XSI_RVolume>	volumes;
		vector<XSI_RDisk>	rdisks;
		struct bio_locate	bl;
		struct bioc_inq 	bi;
		struct bioc_disk	bd;
		struct bioc_vol 	bv;
		int			j;

		bzero(&bl, sizeof(bl));
		bl.bl_name = const_cast<char *>(raid.getRaidControllerName());
		if (ioctl(bio_fd, BIOCLOCATE, &bl) == -1)
			XSI_Log::log()->err(__func__, "Can't locate controller %s via bio(4)",
					    raid.getRaidControllerName());

		bzero(&bi, sizeof(bi));
		bi.bi_cookie = bl.bl_cookie;
		if (ioctl(bio_fd, BIOCINQ, &bi) == -1)
			XSI_Log::log()->err(__func__, "Can't figure out number of volumes and disks on %s",
					    raid.getRaidControllerName());

		for (j = 0; j < bi.bi_novol; j++) {
			XSI_RVolume		volume;
			vector<XSI_RDisk>	disks;
			int			k;

			bzero(&bv, sizeof(bv));
			bv.bv_cookie = bl.bl_cookie;
			bv.bv_volid = j;
			bv.bv_percent = -1;
			bv.bv_seconds = 0;
			if (ioctl(bio_fd, BIOCVOL, &bv) == -1)
				XSI_Log::log()->err(__func__, "Unable to read RAID %s volume %d information",
						    raid.getRaidControllerName(), j);

			volume.setVolumeDevName(bv.bv_dev);
			volume.setVolumeID(bv.bv_volid);
			if (bv.bv_level >= 0)
				volume.setRaidLevel(bv.bv_level);
			switch (bv.bv_status) {
			case BIOC_SVONLINE:
				volume.setVolumeStatus(XSI_RAID_VOL_ONLINE);
				break;
			case BIOC_SVOFFLINE:
				volume.setVolumeStatus(XSI_RAID_VOL_OFFLINE);
				break;
			case BIOC_SVDEGRADED:
				volume.setVolumeStatus(XSI_RAID_VOL_DEGRADED);
				break;
			case BIOC_SVBUILDING:
				volume.setVolumeStatus(XSI_RAID_VOL_BUILDING);
				break;
			case BIOC_SVSCRUB:
				volume.setVolumeStatus(XSI_RAID_VOL_SCRUBBING);
				break;
			case BIOC_SVREBUILD:
				volume.setVolumeStatus(XSI_RAID_VOL_REBUILD);
				break;
			case BIOC_SVINVALID:
				volume.setVolumeStatus(XSI_RAID_VOL_INVALID);
				break;
			default:
				volume.setVolumeStatus(XSI_RAID_VOL_UNKNOWN);
				break;
			}
			if (bv.bv_percent != -1)
				volume.setVolumeOpProgress(bv.bv_percent);
			else
				// redundant, just for show:
				volume.setVolumeOpProgress(-1);
			if (bv.bv_seconds != 0)
				volume.setVolumeOpElapsed(bv.bv_seconds);
			else
				volume.setVolumeOpElapsed(-1);
			volume.setVolumeSize(bv.bv_size);

			for (k = 0; k < bv.bv_nodisk; k++) {
				XSI_RDisk	disk;

				bzero(&bd, sizeof(bd));
				bd.bd_cookie = bl.bl_cookie;
				bd.bd_diskid = k;
				bd.bd_volid = j;

				if (ioctl(bio_fd, BIOCDISK, &bd) == -1)
					XSI_Log::log()->err(__func__, "Unable to read %s, volume %d, disk %d information",
							    raid.getRaidControllerName(),
							    j, k);

				disk.setDiskID(bd.bd_diskid);
				disk.setDiskLocation(bd.bd_channel, bd.bd_target);
				disk.setDiskLUN(bd.bd_lun);
				switch (bd.bd_status) {
				case BIOC_SDONLINE:
					disk.setDiskStatus(XSI_RAID_DSK_ONLINE);
					break;
				case BIOC_SDOFFLINE:
					disk.setDiskStatus(XSI_RAID_DSK_OFFLINE);
					break;
				case BIOC_SDFAILED:
					disk.setDiskStatus(XSI_RAID_DSK_FAILED);
					break;
				case BIOC_SDREBUILD:
					disk.setDiskStatus(XSI_RAID_DSK_REBUILD);
					break;
				case BIOC_SDHOTSPARE:
					disk.setDiskStatus(XSI_RAID_DSK_HOTSPARE);
					break;
				case BIOC_SDUNUSED:
					disk.setDiskStatus(XSI_RAID_DSK_UNUSED);
					break;
				case BIOC_SDSCRUB:
					disk.setDiskStatus(XSI_RAID_DSK_SCRUBBING);
					break;
				case BIOC_SDINVALID:
					disk.setDiskStatus(XSI_RAID_DSK_INVALID);
					break;
				default:
					disk.setDiskStatus(XSI_RAID_DSK_UNKNOWN);
					break;
				}
				disk.setDiskSize(bd.bd_size);
				if (bd.bd_procdev[0])
					disk.setDiskEnclosureDevName(bd.bd_procdev);
				disk.setDiskVendorName(bd.bd_vendor);
				if (bd.bd_serial[0])
					disk.setDiskSerial(bd.bd_serial);

				if (bv.bv_level >= 0)
					disks.push_back(disk);
				else
					rdisks.push_back(disk);
			}
			volume.setVolumeDisks(disks);

			if (bv.bv_level >= 0)
				volumes.push_back(volume);
		}
		raid.setVolumes(volumes);
		raid.setReserveDisks(rdisks);

		raids_tmp.push_back(raid);
	}

	raids = raids_tmp;
}

void
XSI::getNetLLAddr(XSI_NetIfStat &netIfStat, struct sockaddr_dl *sdl)
{
	if (sdl->sdl_alen) {
		switch (sdl->sdl_type) {
		case IFT_ETHER:
		case IFT_CARP:
			netIfStat.setLladdrString(ether_ntoa((struct ether_addr *)LLADDR(sdl)));
			return;
		default:
			break;
		}
	}
}

void
XSI::getNetIfInfo(XSI_NetIfStat &netIfStat, struct ifaddrs *ifaddr,
		  struct sockaddr_dl *sdl)
{
	struct ifreq	ifr, *ifrp;
	int		sock;
	char		ifdescr[IFDESCRSIZE];
	struct carpreq	carpr;

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
		XSI_Log::log()->err(__func__, errno, "socket");

	bzero(&ifr, sizeof(ifr));
	memcpy(&ifr.ifr_addr, ifaddr->ifa_addr,
	       MIN(sizeof(ifr.ifr_addr), ifaddr->ifa_addr->sa_len));
	ifrp = &ifr;
	strlcpy(ifrp->ifr_name, ifaddr->ifa_name, sizeof(ifrp->ifr_name));

	if (ioctl(sock, SIOCGIFMETRIC, (caddr_t)&ifr) != -1)
		netIfStat.setMetric(ifr.ifr_metric);

	if (ioctl(sock, SIOCGIFMTU, (caddr_t)&ifr) != -1)
		netIfStat.setMTU(ifr.ifr_mtu);

	ifr.ifr_data = (caddr_t)&ifdescr;
	if (ioctl(sock, SIOCGIFDESCR, &ifr) != -1)
		netIfStat.setDescription(ifr.ifr_data);

	if (sdl->sdl_alen && sdl->sdl_type == IFT_CARP) {
		/* It's fish! */
		bzero(&carpr, sizeof(carpr));
		ifr.ifr_data = (caddr_t)&carpr;

		if (ioctl(sock, SIOCGVH, (caddr_t)&ifr) != -1 &&
		    carpr.carpr_vhid > 0) {
			int		preempt, arpbalance;
			int		mib[4];
			size_t		siz = sizeof(int);
			enum xsi_carp_state state;

			if (carpr.carpr_state > CARP_MAXSTATE)
				state = CARP_UNKNOWN;
			else if (strcmp(carp_states[carpr.carpr_state], "INIT") == 0)
				state = CARP_INIT;
			else if (strcmp(carp_states[carpr.carpr_state], "BACKUP") == 0)
				state = CARP_BACKUP;
			else if (strcmp(carp_states[carpr.carpr_state], "MASTER") == 0)
				state = CARP_MASTER;
			else
				state = CARP_UNKNOWN;

			mib[0] = CTL_NET;
			mib[1] = PF_INET;
			mib[2] = IPPROTO_CARP;
			mib[3] = CARPCTL_PREEMPT;
			if (sysctl(mib, 4, &preempt, &siz, NULL, 0) == -1)
				XSI_Log::log()->err(__func__, errno, "sysctl");
			mib[3] = CARPCTL_ARPBALANCE;
			if (sysctl(mib, 4, &arpbalance, &siz, NULL, 0) == -1)
				XSI_Log::log()->err(__func__, errno, "sysctl");

			netIfStat.setCARPStatus(carpr.carpr_carpdev[0] ?
						carpr.carpr_carpdev :
						"none",
						state, carpr.carpr_vhid,
						carpr.carpr_advskew,
						carpr.carpr_advbase,
						preempt == 0 ? 0 : 1,
						arpbalance == 0 ? 0 : 1);
		}
	}

	close(sock);
}

void
XSI::getNetIfmediaInfo(XSI_NetIfStat &netIfStat)
{
	struct ifmediareq	ifmr;
	int			sock, *media_list;
	const char	       *ifname = netIfStat.getIfName();
	static bool		have_warned;

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
		XSI_Log::log()->err(__func__, errno, "socket");

	bzero(&ifmr, sizeof(ifmr));
	strlcpy(ifmr.ifm_name, ifname, sizeof(ifmr.ifm_name));
	if (ioctl(sock, SIOCGIFMEDIA, (caddr_t)&ifmr) == -1) {
		// Interface doesn't support media directly
		close(sock);
		return;
	} else {
		string mediaStr;

		if (ifmr.ifm_count == 0) {
			if (!have_warned) {
				XSI_Log::log()->warn(__func__,
						     "%s: no media types?",
						     ifname);
				have_warned = true;
			}
			close(sock);
			return;
		}

		media_list = new int[ifmr.ifm_count];
		ifmr.ifm_ulist = media_list;
		if (ioctl(sock, SIOCGIFMEDIA, (caddr_t)&ifmr) == -1) {
			close(sock);
			XSI_Log::log()->err(__func__, errno, "SIOCGIFMEDIA");
		}

		mediaStr = getMediaWordString(ifmr.ifm_current, 1);
		if (ifmr.ifm_active != ifmr.ifm_current) {
			mediaStr += " (";
			mediaStr += getMediaWordString(ifmr.ifm_active, 0);
			mediaStr += ")";
		}

		netIfStat.setMediaString(mediaStr.c_str());

		delete[] media_list;
	}

	close(sock);
}

void
XSI::getNetIfAddrs(XSI_NetIfStat &netIfStat, struct ifaddrs *ifap, string ifName)
{
	struct ifaddrs	       *ifa;
	vector<XSI_IfAddress>	addrs;
	static bool		have_warned = false;

	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
		XSI_IfAddress addr;

		if (ifName != ifa->ifa_name
		    || !ifa->ifa_addr
		    || ifa->ifa_addr->sa_family == AF_LINK)
			continue;

		switch (ifa->ifa_addr->sa_family) {
		case AF_INET6:
			xsi_in6_fillscopeid((struct sockaddr_in6 *)ifa->ifa_addr);
			if (ifa->ifa_flags & IFF_POINTOPOINT && ifa->ifa_dstaddr)
				xsi_in6_fillscopeid((struct sockaddr_in6 *)ifa->ifa_dstaddr);
			/* FALLTHROUGH */
		case AF_INET:
			addr.setAddress((struct sockaddr *)ifa->ifa_addr,
					(size_t)ifa->ifa_addr->sa_len,
					(struct sockaddr *)ifa->ifa_netmask,
					ifa->ifa_addr->sa_family);
			if (ifa->ifa_flags & IFF_POINTOPOINT && ifa->ifa_dstaddr)
				netIfStat.setPTPDestAddress((struct sockaddr *)ifa->ifa_dstaddr,
							    ifa->ifa_dstaddr->sa_len);
			break;
		default:
			if (!have_warned) {
				XSI_Log::log()->warn(__func__,
						     "Unhandled address family %d found, please report this!",
						     ifa->ifa_addr->sa_family);
				have_warned = true;
			}
			break;
		}

		addrs.push_back(addr);
	}

	netIfStat.setIfAddresses(addrs);
}

#define ROUNDUP(a) \
	((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
void
XSI::getDefaultRoutes(void)
{
	int			mib[7];
	size_t			siz;
	char		       *buf, *next, *ep = NULL;
	struct rt_msghdr       *rtm;
	struct sockaddr        *sa, *rti_info[RTAX_MAX];
	char			host[MAXHOSTNAMELEN];
	int			error;
	vector<struct default_route> defRoutes;

	mib[0] = CTL_NET;
	mib[1] = PF_ROUTE;
	mib[2] = 0;	/* protocol number */
	mib[3] = 0;	/* AF */
	mib[4] = NET_RT_FLAGS;
	mib[5] = RTF_GATEWAY;
	mib[6] = 0;	/* rtableid */
	if (sysctl(mib, 7, NULL, &siz, NULL, 0) == -1) {
		if (errno != ENOENT)
			XSI_Log::log()->err(__func__, errno, "sysctl");
		/* Try again, with pre-OpenBSD 4.1 MIB. */
		if (sysctl(mib, 6, NULL, &siz, NULL, 0) == -1)
			XSI_Log::log()->err(__func__, errno, "sysctl");
	}

	if (siz == 0)
		return;

	if ((buf = new char[siz]) == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	if (sysctl(mib, 7, buf, &siz, NULL, 0) == -1) {
		if (errno != ENOENT)
			XSI_Log::log()->err(__func__, errno, "sysctl");
		/* Try again, with pre-OpenBSD 4.1 MIB. */
		if (sysctl(mib, 6, buf, &siz, NULL, 0) == -1)
			XSI_Log::log()->err(__func__, errno, "sysctl");
	}

	ep = buf + siz;
	for (next = buf; next < ep; next += rtm->rtm_msglen) {
		unsigned int		i;
		struct default_route	defRt;

		rtm = (struct rt_msghdr *)next;
		sa = (struct sockaddr *)(rtm + 1);
		for (i = 0; i < RTAX_MAX; i++) {
			if (rtm->rtm_addrs & (1 << i)) {
				rti_info[i] = sa;
				sa = (struct sockaddr *)((char *)(sa) +
							 ROUNDUP(sa->sa_len));
			} else
				rti_info[i] = NULL;
		}
		if (rti_info[RTAX_DST] == NULL || rti_info[RTAX_GATEWAY] == NULL)
			continue;
		if (rti_info[RTAX_DST]->sa_family == AF_INET) {
			if (((struct sockaddr_in *)rti_info[RTAX_DST])->sin_addr.s_addr == INADDR_ANY)
				defRt.family = XSI_RTAF_IPV4;
			else
				continue;
		} else if (rti_info[RTAX_DST]->sa_family == AF_INET6) {
			struct sockaddr_in6 sin6;
			bool zero_mask = true;;

			sin6 = *((struct sockaddr_in6 *)rti_info[RTAX_DST]);
			if (rti_info[RTAX_NETMASK]) {
				struct sockaddr_in6 mask;
				unsigned int lim, final = 0, illegal = 0;
				unsigned char *p;

				mask = *((struct sockaddr_in6 *)rti_info[RTAX_NETMASK]);
				lim = mask.sin6_len - offsetof(struct sockaddr_in6, sin6_addr);
				lim = lim < sizeof(struct in6_addr) ? lim : sizeof(struct in6_addr);
				for (p = (unsigned char *)&mask.sin6_addr, i = 0; i < lim; p++) {
					int masks[9] = {0x00, 0x80, 0xc0, 0xe0,
							0xf0, 0xf8, 0xfc, 0xfe,
							0xff};
					int j;

					if (final && *p) {
						illegal++;
						sin6.sin6_addr.s6_addr[i++] = 0x00;
						continue;
					}

					for (j = 0; j < 9; j++) {
						if (*p & 0xff == masks[j])
							break;
					}

					if (j >= 1 && j <= 8) {
						zero_mask = false;
						break;
					}
					if (j == 9)
						illegal++;
					if (j != 8)
						final++;
					if (!illegal)
						sin6.sin6_addr.s6_addr[i++] &= *p;
					else
						sin6.sin6_addr.s6_addr[i++] = 0x00;
				}
				while (i < sizeof(struct in6_addr))
					sin6.sin6_addr.s6_addr[i++] = 0x00;
			} else
				zero_mask = false;

			if (zero_mask && IN6_IS_ADDR_UNSPECIFIED(&sin6.sin6_addr))
				defRt.family = XSI_RTAF_IPV6;
			else
				continue;
		} else
			continue;
		if ((error = getnameinfo(rti_info[RTAX_GATEWAY], rti_info[RTAX_GATEWAY]->sa_len,
					 host, sizeof(host), NULL, 0, NI_NUMERICHOST)) != 0)
			XSI_Log::log()->err(__func__, "getnameinfo: %s",
					    gai_strerror(error));
		defRt.gateway = host;
		defRoutes.push_back(defRt);
	}

	network.setDefaultRoutes(defRoutes);

	delete[] buf;
}

/*
 * Adapted from OpenBSD's src/sbin/ifconfig/ifconfig.c:print_media_word()
 */
string
XSI::getMediaWordString(int ifmw, int print_type)
{
	const struct ifmedia_description *desc;
	int	seen_option = 0;
	char	tbuf[SMALLBUFSIZ];
	string	ret = "";

	tbuf[0] = '\0';

	if (print_type) {
		bool found = false;

		for (desc = ifm_type_descriptions; desc->ifmt_string != NULL;
		     desc++) {
			if (IFM_TYPE(ifmw) == desc->ifmt_word) {
				found = true;
				snprintf(tbuf, sizeof(tbuf), "%s ", desc->ifmt_string);
				break;
			}
		}
		if (!found)
			return ("unknown");
	}
	ret += tbuf;

	{
		bool found = false;

		for (desc = ifm_subtype_descriptions; desc->ifmt_string != NULL;
		     desc++) {
			if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
			    IFM_SUBTYPE(desc->ifmt_word) == IFM_SUBTYPE(ifmw)) {
				found = true;
				snprintf(tbuf, sizeof(tbuf), "%s", desc->ifmt_string);
				break;
			}
		}
		if (!found)
			strlcpy(tbuf, "unknown subtype", sizeof(tbuf));
	}
	ret += tbuf;

	/* Find mode. */
	if (IFM_MODE(ifmw) != 0) {
		bool found = false;

		for (desc = ifm_mode_descriptions; desc->ifmt_string != NULL;
		     desc++) {
			if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
			    IFM_MODE(ifmw) == IFM_MODE(desc->ifmt_word)) {
				found = true;
				snprintf(tbuf, sizeof(tbuf), " mode %s", desc->ifmt_string);
				break;
			}
		}
		if (found)
			ret += tbuf;
	}

	/* Find options. */
	for (desc = ifm_option_descriptions; desc->ifmt_string != NULL;
	     desc++) {
		if (IFM_TYPE_MATCH(desc->ifmt_word, ifmw) &&
		    (IFM_OPTIONS(ifmw) & IFM_OPTIONS(desc->ifmt_word)) != 0 &&
		    (seen_option & IFM_OPTIONS(desc->ifmt_word)) == 0) {
			if (seen_option == 0)
				ret += " ";
			snprintf(tbuf, sizeof(tbuf), "%s%s", seen_option ? "," : "", desc->ifmt_string);
			ret += tbuf;
			seen_option |= IFM_OPTIONS(desc->ifmt_word);
		}
	}
	if (IFM_INST(ifmw) != 0) {
		snprintf(tbuf, sizeof(tbuf), " instance %d", IFM_INST(ifmw));
		ret += tbuf;
	}

	return (ret);
}

/*
 * Adapted from OpenBSD's src/sbin/ifconfig/ifconfig.c:in6_fillscopeid()
 */
static void
xsi_in6_fillscopeid(struct sockaddr_in6 *sin6)
{
	if (IN6_IS_ADDR_LINKLOCAL(&sin6->sin6_addr)) {
		sin6->sin6_scope_id =
			ntohs(*(uint16_t *)&sin6->sin6_addr.s6_addr[2]);
		sin6->sin6_addr.s6_addr[2] = sin6->sin6_addr.s6_addr[3] = 0;
	}
}

string
XSI::getSysctlString(int mib0, int mib1, int mib2, unsigned int len,
		     bool must_exist, const char *caller)
{
	int	mib[] = { mib0, mib1, mib2 };
	size_t	siz;
	string	ret;
	char   *buf;

	if (sysctl(mib, len, NULL, &siz, NULL, 0) == -1) {
		if (errno != EOPNOTSUPP || must_exist)
			XSI_Log::log()->err(caller, errno, "sysctl");
		return ("");
	}

	buf = new char[siz];
	if (buf == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	if (sysctl(mib, len, buf, &siz, NULL, 0) == -1)
		XSI_Log::log()->err(__func__, errno, "sysctl");
	ret = buf;
	delete[] buf;
	return (ret);
}
