/*	$Kolabore: pf.cpp,v 1.20 2007/02/09 13:06:06 maxx Exp $	*/
/*
 * Copyright (c) 2006, 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * pf.cpp:
 * Gather statistics from the OpenBSD packet filter pf(4).
 */

#include "log.h"

extern "C" {
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <netinet/in.h>
#ifdef HAVE_NET_IF_H
#  include <net/if.h>
#endif
#if HAVE_NET_PFVAR_H
#  include <net/pfvar.h>
#elif HAVE_NET_PF_PFVAR_H
#  include <net/pf/pfvar.h>
#endif

#include <fcntl.h>
#include <unistd.h>
}

#include <cerrno>
#include <iostream>
#include <climits>
#include <cstring>

#include <list>
#include <vector>

#include "pf.h"

#ifndef PF_DEVICE
#  define PF_DEVICE	"/dev/pf"
#endif /* !PF_DEVICE */

using namespace std;

PF::PF(void)
{
	pf_fd = -1;
	initialized = false;
}

PF::~PF(void)
{
	if (pf_fd != -1)
		close(pf_fd);
}

/*
 * Initialize the PF class and return 1 on success, 0 on failure.
 */
int
PF::init(void)
{
#ifdef HAVE_PF
	if ((pf_fd = open(PF_DEVICE, O_RDONLY, 0)) == -1) {
		cerr << "Error: Unable to open " PF_DEVICE " for pf(4) stats: "
			<< strerror(errno) << endl;
		return (0);
	}

	initialized = true;
	return (1);
#else
	return (0);
#endif /* HAVE_PF */
}

/*
 * updateStats() for pf(4): reads counters and other information, inspired by
 * OpenBSD's src/sbin/pfctl/pfctl*.c and Daniel Hartmeier's pfstat.
 */
void
PF::updateStats(XSI_PFilter &packetFilter)
{
#ifdef HAVE_PF
	struct pf_status	status;
	struct pfioc_iface	iface;
	struct pfioc_limit	limit;
	struct pfi_kif		ifs[256];	// Eats >50 kB of stack memory
	list<int>		ifs_with_traffic;
	list<int>::iterator	iwt_it;
	int			i;
	vector<FilteredIfStat>	old_fIfStats = packetFilter.getFilteredIfStats();
	bool			drop_old_stats = false, filteredIfs_changed;

	if (!initialized || pf_fd == -1)
		return;

	packetFilter.setPFilterType(PFILTER_PF);
	packetFilter.setAvailable(true);

	memset(&status, 0, sizeof(status));
	if (ioctl(pf_fd, DIOCGETSTATUS, &status) == -1)
		XSI_Log::log()->err(__func__, errno,
				    "Unable to get status from pf(4)");

	if (!status.running) {
		packetFilter.setEnabled(false);
		return;
	} else
		packetFilter.setEnabled(true);
	packetFilter.setRuntime(time(NULL) - status.since);
	packetFilter.setRuleMatches(status.counters[PFRES_MATCH]);
	packetFilter.setStates(status.states);
	packetFilter.setStateSearches(status.fcounters[FCNT_STATE_SEARCH]);
	packetFilter.setStateInserts(status.fcounters[FCNT_STATE_INSERT]);
	packetFilter.setStateRemovals(status.fcounters[FCNT_STATE_REMOVALS]);
#ifdef PFRES_BADSTATE
	packetFilter.setStateMismatches(status.counters[PFRES_BADSTATE]);
#endif

	memset(&limit, 0, sizeof(limit));
	limit.index = PF_LIMIT_STATES;
	if (ioctl(pf_fd, DIOCGETLIMIT, &limit) == -1)
		XSI_Log::log()->err(__func__, errno,
				    "Unable to get current state limit from pf(4)");
	if (limit.limit == UINT_MAX)
		packetFilter.setStatesLimit(-1);
	else
		packetFilter.setStatesLimit((int64_t)limit.limit);

	memset(&iface, 0, sizeof(iface));
	iface.pfiio_buffer = ifs;
	iface.pfiio_esize = sizeof(ifs[0]);
	iface.pfiio_size = sizeof(ifs) / sizeof(ifs[0]);
	if (ioctl(pf_fd, DIOCIGETIFACES, &iface) == -1) {
		/*
		 * The ioctl is always there, even on older implementations of
		 * pf(4), but it can also fail there. On old implementations
		 * where we can't get any further information, so accept this
		 * error and be happy about what we've got so far.
		 */
		if (errno != ENODEV)
			XSI_Log::log()->err(__func__, errno,
					    "Unable to get interface stats from pf(4)");
		return;
	}

	/* Find interfaces that have seen traffic */
	for (i = 0; i < iface.pfiio_size; i++) {
		int j;

		for (j = 0; j < 8; j++) {
			if (ifs[i].pfik_packets[j&1?0:1][j&2?0:1][j&4?0:1] != 0)
				ifs_with_traffic.push_back(i);
		}
	}
	ifs_with_traffic.sort();
	ifs_with_traffic.unique();

	if (ifs_with_traffic.size() != old_fIfStats.size())
		drop_old_stats = true;

	/* Get counters from ifs_with_traffic: bytes/packets[AF][direction][op] */
	do {
		filteredIfs_changed = false;
		if (drop_old_stats) {
			vector<FilteredIfStat>	fIfStats;

			for (iwt_it = ifs_with_traffic.begin();
			     iwt_it != ifs_with_traffic.end();
			     iwt_it++) {
				FilteredIfStat iwt(ifs[*iwt_it].pfik_name);

				iwt.setPktsV4OutBlocked(ifs[*iwt_it].pfik_packets[0][1][PF_DROP]);
				iwt.setPktsV4OutPassed(ifs[*iwt_it].pfik_packets[0][1][PF_PASS]);
				iwt.setPktsV4InBlocked(ifs[*iwt_it].pfik_packets[0][0][PF_DROP]);
				iwt.setPktsV4InPassed(ifs[*iwt_it].pfik_packets[0][0][PF_PASS]);
				iwt.setPktsV6OutBlocked(ifs[*iwt_it].pfik_packets[1][1][PF_DROP]);
				iwt.setPktsV6OutPassed(ifs[*iwt_it].pfik_packets[1][1][PF_PASS]);
				iwt.setPktsV6InBlocked(ifs[*iwt_it].pfik_packets[1][0][PF_DROP]);
				iwt.setPktsV6InPassed(ifs[*iwt_it].pfik_packets[1][0][PF_PASS]);
				iwt.setBytesV4OutBlocked(ifs[*iwt_it].pfik_bytes[0][1][PF_DROP]);
				iwt.setBytesV4OutPassed(ifs[*iwt_it].pfik_bytes[0][1][PF_PASS]);
				iwt.setBytesV4InBlocked(ifs[*iwt_it].pfik_bytes[0][0][PF_DROP]);
				iwt.setBytesV4InPassed(ifs[*iwt_it].pfik_bytes[0][0][PF_PASS]);
				iwt.setBytesV6OutBlocked(ifs[*iwt_it].pfik_bytes[1][1][PF_DROP]);
				iwt.setBytesV6OutPassed(ifs[*iwt_it].pfik_bytes[1][1][PF_PASS]);
				iwt.setBytesV6InBlocked(ifs[*iwt_it].pfik_bytes[1][0][PF_DROP]);
				iwt.setBytesV6InPassed(ifs[*iwt_it].pfik_bytes[1][0][PF_PASS]);

				fIfStats.push_back(iwt);
			}
			packetFilter.setFilteredIfStats(fIfStats);
		} else {
			vector<FilteredIfStat>::iterator it;

			for (it = old_fIfStats.begin(), iwt_it = ifs_with_traffic.begin();
			     it != old_fIfStats.end(); it++, iwt_it++) {
				if (strcmp(ifs[*iwt_it].pfik_name, it->getIfName()) != 0) {
					filteredIfs_changed = true;
					drop_old_stats = true;
					break;
				}
				it->setPktsV4OutBlocked(ifs[*iwt_it].pfik_packets[0][1][PF_DROP]);
				it->setPktsV4OutPassed(ifs[*iwt_it].pfik_packets[0][1][PF_PASS]);
				it->setPktsV4InBlocked(ifs[*iwt_it].pfik_packets[0][0][PF_DROP]);
				it->setPktsV4InPassed(ifs[*iwt_it].pfik_packets[0][0][PF_PASS]);
				it->setPktsV6OutBlocked(ifs[*iwt_it].pfik_packets[1][1][PF_DROP]);
				it->setPktsV6OutPassed(ifs[*iwt_it].pfik_packets[1][1][PF_PASS]);
				it->setPktsV6InBlocked(ifs[*iwt_it].pfik_packets[1][0][PF_DROP]);
				it->setPktsV6InPassed(ifs[*iwt_it].pfik_packets[1][0][PF_PASS]);
				it->setBytesV4OutBlocked(ifs[*iwt_it].pfik_bytes[0][1][PF_DROP]);
				it->setBytesV4OutPassed(ifs[*iwt_it].pfik_bytes[0][1][PF_PASS]);
				it->setBytesV4InBlocked(ifs[*iwt_it].pfik_bytes[0][0][PF_DROP]);
				it->setBytesV4InPassed(ifs[*iwt_it].pfik_bytes[0][0][PF_PASS]);
				it->setBytesV6OutBlocked(ifs[*iwt_it].pfik_bytes[1][1][PF_DROP]);
				it->setBytesV6OutPassed(ifs[*iwt_it].pfik_bytes[1][1][PF_PASS]);
				it->setBytesV6InBlocked(ifs[*iwt_it].pfik_bytes[1][0][PF_DROP]);
				it->setBytesV6InPassed(ifs[*iwt_it].pfik_bytes[1][0][PF_PASS]);
			}
			if (!filteredIfs_changed)
				packetFilter.setFilteredIfStats(old_fIfStats);
		}
	} while (filteredIfs_changed);
#endif /* HAVE_PF */
}
