/*	$Kolabore: raid.cpp,v 1.22 2007/02/09 20:58:49 maxx Exp $	*/
/*
 * Copyright (c) 2006, 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "log.h"

#include <string>
#include <vector>

#include <cstdio>

#include "raid.h"

#define SMALLBUFSIZ	64

using namespace std;

XSI_RDisk::XSI_RDisk(void)
{
	bus = target = lun = -1;
	status = XSI_RAID_DSK_INVALID;
	size = 0;
}

void
XSI_RDisk::setDiskID(int i)
{
	id = i;
}

void
XSI_RDisk::setDiskLocation(int b, int t)
{
	bus = b;
	target = t;
}

void
XSI_RDisk::setDiskLUN(int l)
{
	lun = l;
}

void
XSI_RDisk::setDiskStatus(enum xsi_raid_disk_status s)
{
	status = s;
}

void
XSI_RDisk::setDiskSize(uint64_t siz)
{
	size = siz;
}

void
XSI_RDisk::setDiskEnclosureDevName(const char *encldev)
{
	enclosure = encldev;
}

void
XSI_RDisk::setDiskVendorName(const char *v)
{
	vendor = v;
}

void
XSI_RDisk::setDiskSerial(const char *sn)
{
	serial = sn;
}

const char *
XSI_RDisk::getDiskIDString(void)
{
	return (makeIntStr((int32_t)id));
}

const char *
XSI_RDisk::getDiskBusString(void)
{
	return (makeIntStr((int32_t)bus));
}

const char *
XSI_RDisk::getDiskTargetString(void)
{
	return (makeIntStr((int32_t)target));
}

const char *
XSI_RDisk::getDiskLUNString(void)
{
	if (lun == -1)
		return (NULL);
	else
		return (makeIntStr((int32_t)lun));
}

const char *
XSI_RDisk::getDiskStatusString(void)
{
	switch (status) {
	case XSI_RAID_DSK_ONLINE:
		return ("online");
	case XSI_RAID_DSK_OFFLINE:
		return ("offline");
	case XSI_RAID_DSK_FAILED:
		return ("failed");
	case XSI_RAID_DSK_REBUILD:
		return ("rebuild");
	case XSI_RAID_DSK_HOTSPARE:
		return ("hot_spare");
	case XSI_RAID_DSK_UNUSED:
		return ("unused");
	case XSI_RAID_DSK_SCRUBBING:
		return ("scrubbing");
	case XSI_RAID_DSK_INVALID:
		return ("invalid");
	case XSI_RAID_DSK_UNKNOWN:
	default:
		return ("unknown");
	}
}

const char *
XSI_RDisk::getDiskSizeString(void)
{
	return (makeIntStr(size));
}

const char *
XSI_RDisk::getDiskEnclosureDevName(bool sanitize)
{
	if (enclosure.empty())
		return (NULL);
	else {
		if (sanitize)
			sanitizeString(enclosure);
		return (enclosure.c_str());
	}
}

const char *
XSI_RDisk::getDiskVendorName(bool sanitize)
{
	if (vendor.empty())
		return (NULL);
	else {
		if (sanitize)
			normalizeString(vendor);
		return (vendor.c_str());
	}
}

const char *
XSI_RDisk::getDiskSerial(bool sanitize)
{
	if (serial.empty())
		return (NULL);
	else {
		if (sanitize)
			normalizeString(serial);
		return (serial.c_str());
	}
}


XSI_RVolume::XSI_RVolume(void)
{
	id = -1;
	level = -1;
	status = XSI_RAID_VOL_INVALID;
	progress = -1;
	elapsed = -1;
	size = 0;
}

void
XSI_RVolume::setVolumeDevName(const char *dev)
{
	device = dev;
}

void
XSI_RVolume::setVolumeID(int i)
{
	id = i;
}

void
XSI_RVolume::setRaidLevel(int lvl)
{
	level = lvl;
}

void
XSI_RVolume::setVolumeStatus(enum xsi_raid_vol_status s)
{
	status = s;
}

void
XSI_RVolume::setVolumeOpProgress(int p)
{
	progress = p;
}

void
XSI_RVolume::setVolumeOpElapsed(int seconds)
{
	elapsed = seconds;
}

void
XSI_RVolume::setVolumeSize(uint64_t siz)
{
	size = siz;
}

void
XSI_RVolume::setVolumeDisks(vector<XSI_RDisk> dsks)
{
	disks = dsks;
}

const char *
XSI_RVolume::getVolumeDevName(void)
{
	sanitizeString(device, true);
	return (device.c_str());
}

const char *
XSI_RVolume::getVolumeIDString(void)
{
	return (makeIntStr((int32_t)id));
}

const char *
XSI_RVolume::getRaidLevelString(void)
{
	static string	ret;
	char		buf[64];

	snprintf(buf, sizeof(buf), "RAID%d", level);
	ret = buf;

	return (ret.c_str());
}

const char *
XSI_RVolume::getVolumeStatusString(void)
{
	switch (status) {
	case XSI_RAID_VOL_ONLINE:
		return ("online");
	case XSI_RAID_VOL_OFFLINE:
		return ("offline");
	case XSI_RAID_VOL_DEGRADED:
		return ("degraded");
	case XSI_RAID_VOL_BUILDING:
		return ("building");
	case XSI_RAID_VOL_SCRUBBING:
		return ("scrubbing");
	case XSI_RAID_VOL_REBUILD:
		return ("rebuild");
	case XSI_RAID_VOL_INVALID:
		return ("invalid");
	case XSI_RAID_VOL_UNKNOWN:
	default:
		return ("unknown");
	}
}

const char *
XSI_RVolume::getVolumeOpProgressString(void)
{
	if (progress < 0)
		return (NULL);
	else
		return (makeIntStr((int32_t)progress));
}

const char *
XSI_RVolume::getVolumeOpElapsedString(void)
{
	if (elapsed <= 0)
		return (NULL);
	else
		return (makeIntStr((int32_t)elapsed));
}

const char *
XSI_RVolume::getVolumeSizeString(void)
{
	return (makeIntStr(size));
}

vector<XSI_RDisk>
XSI_RVolume::getVolumeDisks(void)
{
	return (disks);
}

XSI_Raid::XSI_Raid(const char *ctrl)
{
	controller = ctrl;
}

void
XSI_Raid::setVolumes(vector<XSI_RVolume> vols)
{
	volumes = vols;
}

void
XSI_Raid::setReserveDisks(vector<XSI_RDisk> dsks)
{
	disks = dsks;
}

vector<XSI_RVolume>
XSI_Raid::getVolumes(void)
{
	return (volumes);
}

vector<XSI_RDisk>
XSI_Raid::getReserveDisks(void)
{
	return (disks);
}

const char *
XSI_Raid::getRaidControllerName(bool sanitize)
{
	if (sanitize)
		sanitizeString(controller, true);
	return (controller.c_str());
}
