/*	$Kolabore: sensor.cpp,v 1.44 2007/02/09 13:04:35 maxx Exp $	*/
/*
 * Copyright (c) 2006, 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * sensor.cpp:
 * Provides a way for xsi to set and read sensor data, without having to care
 * for where the data originally originated from. It is close to what the
 * OpenBSD sensor framework looks like.
 */

#include "log.h"

extern "C" {
#include <sys/time.h>
#include <sys/types.h>
}

#include <cerrno>
#include <climits>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <string>

#include "sensor.h"

#define SMALLBUFSIZ	64

using namespace std;

XSI_Sensor::XSI_Sensor(void)
{
	sensor_id = -1;
	sensor_type = XSI_SENSOR_TYPE_UNSPEC;
	sensor_value = 0;
	other_divisor = 1;
	other_frac_digits = 0;
	sensor_status = XSI_SENSOR_STATUS_UNSPEC;
	sensor_valid = true;
	value_unknown = -1;
	memset(&value_last_change_time, 0, sizeof(value_last_change_time));
	value_age = -1.0;
	has_value_age = false;
}

void
XSI_Sensor::setIsValid(bool valid)
{
	sensor_valid = valid;
}

void
XSI_Sensor::setHasUnknownValue(bool unknown)
{
	if (unknown)
		value_unknown = 1;
	else
		value_unknown = 0;
}

void
XSI_Sensor::setSensorID(int32_t id)
{
	sensor_id = id;
}

void
XSI_Sensor::setSensorDevice(const char *device)
{
	sensor_device = device;
}

void
XSI_Sensor::setSensorType(enum xsi_sensor_type type)
{
	sensor_type = type;
}

void
XSI_Sensor::setSensorValue(int64_t value, double divisor, uint16_t frac_digits)
{
	sensor_value = value;
	if (divisor > 0)
		other_divisor = divisor;
	else
		XSI_Log::log()->err(__func__, "Internal error: Divisor cannot be 0 or negative");
	other_frac_digits = frac_digits;
	if (value_unknown == -1)
		value_unknown = 0;
}

void
XSI_Sensor::setSensorValue(enum xsi_drive_sensor_status status)
{
	sensor_value = (int64_t)status;
	if (value_unknown == -1)
		value_unknown = 0;
}

void
XSI_Sensor::setSensorDescription(const char *description)
{
	if (description == NULL)
		sensor_description = "";
	else
		sensor_description = description;
}

void
XSI_Sensor::setOtherSensorType(const char *otype)
{
	other_type_name = otype;
}

void
XSI_Sensor::setOtherSensorUnit(const char *ounit)
{
	other_unit_name = ounit;
}

void
XSI_Sensor::setSensorStatus(enum xsi_sensor_status status)
{
	sensor_status = status;
}

void
XSI_Sensor::setValueLastChangeTime(struct timeval *tv)
{
	if (tv == NULL) {
		memset(&value_last_change_time, 0,
		       sizeof(value_last_change_time));
		has_value_age = false;
		return;
	}

	value_last_change_time.tv_sec = tv->tv_sec;
	value_last_change_time.tv_usec = tv->tv_usec;
	has_value_age = true;
}

void
XSI_Sensor::setValueAge(double age)
{
	value_age = age;
	if (value_age < 0)
		has_value_age = false;
	else
		has_value_age = true;
}

bool
XSI_Sensor::isValid(void)
{
	return (sensor_valid);
}

bool
XSI_Sensor::hasUnknownValue(void)
{
	return (value_unknown != 0);
}

bool
XSI_Sensor::hasStatus(void)
{
	return (sensor_status != XSI_SENSOR_STATUS_UNSPEC);
}

bool
XSI_Sensor::hasDescription(void)
{
	return (!sensor_description.empty());
}

bool
XSI_Sensor::hasValueAge(void)
{
	return (has_value_age);
}

bool
XSI_Sensor::isIndicatorSensor(void)
{
	return (sensor_type == XSI_SENSOR_TYPE_INDICATOR);
}

bool
XSI_Sensor::isDriveSensor(void)
{
	return (sensor_type == XSI_SENSOR_TYPE_DRIVE);
}

bool
XSI_Sensor::isUnknownSensor(void)
{
	return (sensor_type == XSI_SENSOR_TYPE_OTHER);
}

const char *
XSI_Sensor::getSensorIDString(void)
{
	return (makeIntStr(sensor_id));
}

const char *
XSI_Sensor::getSensorDeviceString(void)
{
	sanitizeString(sensor_device, true);
	return (sensor_device.c_str());
}

const char *
XSI_Sensor::getSensorTypeString(void)
{
	switch (sensor_type) {
	case XSI_SENSOR_TYPE_UNSPEC:
		XSI_Log::log()->err(__func__, "Internal error: XSI class neglected to set sensor type");
	case XSI_SENSOR_TYPE_OTHER:
		return ("other");
	case XSI_SENSOR_TYPE_TEMP:
		return ("temperature");
	case XSI_SENSOR_TYPE_FANRPM:
		return ("fan_speed");
	case XSI_SENSOR_TYPE_VOLTS_DC:
		return ("voltage_dc");
	case XSI_SENSOR_TYPE_VOLTS_AC:
		return ("voltage_ac");
	case XSI_SENSOR_TYPE_OHMS:
		return ("resistance");
	case XSI_SENSOR_TYPE_WATTS:
		return ("power");
	case XSI_SENSOR_TYPE_AMPS:
		return ("current");
	case XSI_SENSOR_TYPE_WATTHOUR:
		return ("energy");
	case XSI_SENSOR_TYPE_AMPHOUR:
		return ("charge");
	case XSI_SENSOR_TYPE_LUX:
		return ("illuminance");
	case XSI_SENSOR_TYPE_PERCENT:
		return ("percentage");
	case XSI_SENSOR_TYPE_INDICATOR:
		return ("indicator");
	case XSI_SENSOR_TYPE_RAW_INT:
		return ("generic");
	case XSI_SENSOR_TYPE_TIMEDELTA:
		return ("timedelta");
	case XSI_SENSOR_TYPE_DRIVE:
		return ("drive");
	default:
		XSI_Log::log()->err(__func__, "Internal error: Unknown sensor type set");
	}

	/* NOTREACHED */
	/* Squelch warnings: */
	exit(1);
}

const char *
XSI_Sensor::getOtherSensorTypeString(void)
{
	sanitizeString(other_type_name);
	return (other_type_name.c_str());
}

const char *
XSI_Sensor::getSensorDescriptionString(void)
{
	sanitizeString(sensor_description);
	if (sensor_description.empty())
		return (NULL);
	else
		return (sensor_description.c_str());
}

const char *
XSI_Sensor::getSensorValueString(void)
{
	switch (sensor_type) {
	case XSI_SENSOR_TYPE_UNSPEC:
		XSI_Log::log()->err(__func__, "Internal error: XSI class neglected to set sensor type");
	case XSI_SENSOR_TYPE_OTHER:
		if (other_divisor == 1 && other_frac_digits == 0)
			return (makeIntStr(sensor_value));
		else
			return (makeFloatStr(sensor_value,
					     other_divisor,
					     other_frac_digits));
	case XSI_SENSOR_TYPE_TEMP:
		/* Return degree Celsius. */
		return (makeFloatStr(sensor_value - 273150000, 1000000, 6));
	case XSI_SENSOR_TYPE_FANRPM:
	case XSI_SENSOR_TYPE_RAW_INT:
		return (makeIntStr(sensor_value));
	case XSI_SENSOR_TYPE_VOLTS_DC:
	case XSI_SENSOR_TYPE_VOLTS_AC:
	case XSI_SENSOR_TYPE_OHMS:
	case XSI_SENSOR_TYPE_WATTS:
	case XSI_SENSOR_TYPE_AMPS:
	case XSI_SENSOR_TYPE_WATTHOUR:
	case XSI_SENSOR_TYPE_AMPHOUR:
	case XSI_SENSOR_TYPE_LUX:
	case XSI_SENSOR_TYPE_TIMEDELTA:
		return (makeFloatStr(sensor_value, 1000000, 6));
	case XSI_SENSOR_TYPE_PERCENT:
		return (makeFloatStr(sensor_value, 1000, 3));
	case XSI_SENSOR_TYPE_INDICATOR:
		if (sensor_value == 1)
			return ("ON");
		else if (sensor_value == 0)
			return ("OFF");
		else
			return ("invalid");
	case XSI_SENSOR_TYPE_DRIVE:
		return (getDriveStatusString());
	default:
		XSI_Log::log()->err(__func__, "Internal error: Unknown sensor type set");
	}

	/* NOTREACHED */
	/* Squelch warnings: */
	exit(1);
}

const char *
XSI_Sensor::getSensorUnitString(void)
{
	switch (sensor_type) {
	case XSI_SENSOR_TYPE_UNSPEC:
		XSI_Log::log()->err(__func__, "Internal error: XSI class neglected to set sensor type");
	case XSI_SENSOR_TYPE_OTHER:
		if (other_unit_name.empty())
			return (NULL);
		else {
			sanitizeString(other_unit_name, true);
			return (other_unit_name.c_str());
		}
	case XSI_SENSOR_TYPE_TEMP:
		return ("C");
	case XSI_SENSOR_TYPE_FANRPM:
		return ("rpm");
	case XSI_SENSOR_TYPE_VOLTS_DC:
	case XSI_SENSOR_TYPE_VOLTS_AC:
		return ("V");
	case XSI_SENSOR_TYPE_OHMS:
		return ("Ohm");
	case XSI_SENSOR_TYPE_WATTS:
		return ("W");
	case XSI_SENSOR_TYPE_AMPS:
		return ("A");
	case XSI_SENSOR_TYPE_WATTHOUR:
		return ("Wh");
	case XSI_SENSOR_TYPE_AMPHOUR:
		return ("Ah");
	case XSI_SENSOR_TYPE_LUX:
		return ("lx");
	case XSI_SENSOR_TYPE_PERCENT:
		return ("percent");
	case XSI_SENSOR_TYPE_TIMEDELTA:
		return ("ms");
	case XSI_SENSOR_TYPE_INDICATOR:
	case XSI_SENSOR_TYPE_RAW_INT:
	case XSI_SENSOR_TYPE_DRIVE:
		return (NULL);
	default:
		XSI_Log::log()->err(__func__, "Internal error: Unknown sensor type set");
	}

	/* NOTREACHED */
	/* Squelch warnings: */
	exit(1);
}

const char *
XSI_Sensor::getSensorStatusString(void)
{
	switch (sensor_status) {
	case XSI_SENSOR_STATUS_UNSPEC:
		return ("unspecified");
	case XSI_SENSOR_STATUS_OK:
		return ("OK");
	case XSI_SENSOR_STATUS_WARN:
		return ("WARNING");
	case XSI_SENSOR_STATUS_CRIT:
		return ("CRITICAL");
	case XSI_SENSOR_STATUS_UNKNOWN:
		return ("unknown");
	default:
		XSI_Log::log()->err(__func__, "Internal error: Unknown sensor status");
	}

	/* NOTREACHED */
	/* Squelch warnings: */
	exit(1);
}

const char *
XSI_Sensor::getSensorValueAgeString(void)
{
	if (!has_value_age)
		return (NULL);

	if (value_age < 0) {
		struct timeval tv;
		double age;

		if (gettimeofday(&tv, NULL) == -1)
			XSI_Log::log()->err(__func__, errno, "gettimeofday");

		age = (double)tv.tv_sec;
		age += (double)tv.tv_usec / 1000000.0;
		age -= (double)value_last_change_time.tv_sec;
		age -= (double)value_last_change_time.tv_usec / 1000000.0;

		if (age < 0.0)
			age = 0.0;

		return (makeFloatStr(age, 1, 3));
	} else
		return (makeFloatStr(value_age, 1, 3));
}

const char *
XSI_Sensor::getDriveStatusString(void)
{
	switch (sensor_value) {
	case XSI_DRIVE_STATUS_UNSPEC:
		XSI_Log::log()->err(__func__, "Internal error: Drive sensor had no drive status set");
	case XSI_DRIVE_STATUS_EMPTY:
		return ("empty");
	case XSI_DRIVE_STATUS_READY:
		return ("ready");
	case XSI_DRIVE_STATUS_POWERUP:
		return ("powering_up");
	case XSI_DRIVE_STATUS_ONLINE:
		return ("online");
	case XSI_DRIVE_STATUS_IDLE:
		return ("idle");
	case XSI_DRIVE_STATUS_ACTIVE:
		return ("active");
	case XSI_DRIVE_STATUS_REBUILD:
		return ("rebuilding");
	case XSI_DRIVE_STATUS_POWERDOWN:
		return ("powering_down");
	case XSI_DRIVE_STATUS_FAIL:
		return ("failed");
	case XSI_DRIVE_STATUS_PFAIL:
		return ("degraded");
	case XSI_DRIVE_STATUS_OTHER:
		return ("other");
	default:
		XSI_Log::log()->err(__func__, "Internal error: Unknown drive status");
	}

	/* NOTREACHED */
	/* Squelch warnings: */
	exit(1);
}
