/*	$Kolabore: sensor.h,v 1.28 2007/01/03 12:19:33 maxx Exp $	*/
/*
 * Copyright (c) 2006, 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * sensor.h:
 * Interface to XSI sensor data.
 * Many sensor types are known, and will show up in the output in elements with
 * restricted tokens for optimal parseability. Other sensors not known to XSI
 * (who knows, maybe a water-flow sensor measuring liters/hour or something)
 * can still be supplied and will show in less easily parsed <other_* />
 * elements.
 * Care must be taken to supply values in their proper units for known sensor
 * types, or the result in the output will be wrong.
 */

#ifndef __SENSOR_H__
#define __SENSOR_H__

extern "C" {
#include <sys/time.h>
}

#include "util.h"

using namespace std;

/*
 * Sensor types:
 * Leaving the "unspecified" default alone is an error and will cause an error
 * (internal safeguard.) Otherwise, be as specific as possible and set _OTHER
 * as the last resort. It will enable the <other_* /> elements with the
 * sensor type token of the OS' choice and display the sensor's raw value.
 */
enum xsi_sensor_type {
	/* unspecified */
	XSI_SENSOR_TYPE_UNSPEC = 0,
	/* other */
	XSI_SENSOR_TYPE_OTHER,
	/* Temperature, uK (micro-Kelvin) */
	XSI_SENSOR_TYPE_TEMP,
	/* Fan speed, RPM */
	XSI_SENSOR_TYPE_FANRPM,
	/* Voltage, uV DC */
	XSI_SENSOR_TYPE_VOLTS_DC,
	/* Voltage, uV AC */
	XSI_SENSOR_TYPE_VOLTS_AC,
	/* Resistance, Ohm */
	XSI_SENSOR_TYPE_OHMS,
	/* Power, uW */
	XSI_SENSOR_TYPE_WATTS,
	/* Current, uA */
	XSI_SENSOR_TYPE_AMPS,
	/* Energy, uWh */
	XSI_SENSOR_TYPE_WATTHOUR,
	/* Charge, uAh */
	XSI_SENSOR_TYPE_AMPHOUR,
	/* Illuminance, ulx */
	XSI_SENSOR_TYPE_LUX,
	/* Percentage, 1/10th bp (0.001%) */
	XSI_SENSOR_TYPE_PERCENT,
	/* On/Off indicator (1 = ON, 0 = OFF) */
	XSI_SENSOR_TYPE_INDICATOR,
	/* Raw integer value */
	XSI_SENSOR_TYPE_RAW_INT,
	/* System time error, ns (nano-seconds) */
	XSI_SENSOR_TYPE_TIMEDELTA,
	/* Drive status "sensor" */
	XSI_SENSOR_TYPE_DRIVE,
	/* Always the last item: */
	XSI_SENSOR_TYPE_MAX
};

/*
 * General status of the sensor.
 */
enum xsi_sensor_status {
	XSI_SENSOR_STATUS_UNSPEC = 0,
	XSI_SENSOR_STATUS_OK,
	XSI_SENSOR_STATUS_WARN,
	XSI_SENSOR_STATUS_CRIT,
	XSI_SENSOR_STATUS_UNKNOWN,
	XSI_SENSOR_STATUS_MAX
};

/*
 * Drive sensors, e.g. in disk enclosures, attach meaning to certain values.
 * Understood by XSI are:
 */
enum xsi_drive_sensor_status {
	XSI_DRIVE_STATUS_UNSPEC = 0,
	XSI_DRIVE_STATUS_EMPTY,
	XSI_DRIVE_STATUS_READY,
	XSI_DRIVE_STATUS_POWERUP,
	XSI_DRIVE_STATUS_ONLINE,
	XSI_DRIVE_STATUS_IDLE,
	XSI_DRIVE_STATUS_ACTIVE,
	XSI_DRIVE_STATUS_REBUILD,
	XSI_DRIVE_STATUS_POWERDOWN,
	XSI_DRIVE_STATUS_FAIL,
	XSI_DRIVE_STATUS_PFAIL,
	XSI_DRIVE_STATUS_OTHER,
	XSI_DRIVE_STATUS_MAX
};

class XSI_Sensor : private XSI_Util {
public:
	XSI_Sensor(void);

	/*
	 * Newly created sensor objects are "valid".
	 * As soon as the value is set with setSensorValue(), the value is
	 * "known", unless it was previously set otherwise with
	 * setHasUnknownValue(), which takes precedence.
	 */
	void			setIsValid(bool);
	void			setHasUnknownValue(bool);
	void			setSensorID(int32_t);
	void			setSensorDevice(const char *);
	void			setSensorType(enum xsi_sensor_type);
	/*
	 * Divisor and frac_digits parameters are only used with
	 * XSI_SENSOR_TYPE_OTHER.
	 */
	void			setSensorValue(int64_t /* value */,
					       double divisor = 1,
					       uint16_t frac_digits = 0);
	void			setSensorValue(enum xsi_drive_sensor_status);
	void			setSensorDescription(const char *);
	void			setOtherSensorType(const char *);
	void			setOtherSensorUnit(const char *);
	void			setSensorStatus(enum xsi_sensor_status);
	/*
	 * Use EITHER setValueLastChanged() XOR setValueAge(). The former takes
	 * precedence. Use only when this data comes from the sensor itself.
	 */
	void			setValueLastChangeTime(struct timeval * /* time of last change */);
	void			setValueAge(double /* seconds */);

	bool			isValid(void);
	bool			hasUnknownValue(void);
	bool			hasStatus(void);
	bool			hasDescription(void);
	bool			hasValueAge(void);
	bool			isIndicatorSensor(void);
	bool			isDriveSensor(void);
	bool			isUnknownSensor(void);
	const char *		getSensorIDString(void);
	const char *		getSensorDeviceString(void);
	const char *		getSensorTypeString(void);
	const char *		getOtherSensorTypeString(void);
	const char *		getSensorDescriptionString(void);
	const char *		getSensorValueString(void);
	const char *		getSensorUnitString(void);
	const char *		getSensorStatusString(void);
	const char *		getSensorValueAgeString(void);

private:
	const char *		getDriveStatusString(void);

	int32_t			sensor_id;
	string			sensor_device;
	enum xsi_sensor_type	sensor_type;
	int64_t 		sensor_value;
	string			sensor_description;
	string			other_type_name;
	string			other_unit_name;
	double			other_divisor;
	uint16_t		other_frac_digits;
	enum xsi_sensor_status	sensor_status;
	bool			sensor_valid;
	int			value_unknown;
	struct timeval		value_last_change_time;
	double			value_age;
	bool			has_value_age;
};

#endif /* __SENSOR_H__ */
