/*	$Kolabore: server.cpp,v 1.35 2007/02/09 13:02:39 maxx Exp $	*/
/*
 * Copyright (c) 2006, 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * server.cpp:
 * as-simple-as-possible server class.
 */

#include "log.h"

extern "C" {
#include <sys/param.h>
#ifdef HAVE_LINUX_PARAM_H
#  include <linux/param.h>
#endif
#include <sys/socket.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include <netdb.h>
#include <poll.h>
#include <unistd.h>
}

#include <iostream>
#include <list>
#include <string>
#include <vector>

#include <cerrno>
#include <cstdlib>
#include <cstring>

#include "server.h"
#include "compat/xsicompat.h"

extern int errno;

using namespace std;

Server::Server(Config *c)
{
	int i;

	if (c == NULL) {
		cerr << "Internal error in " << __func__
			<< ": Configuration cannot be NULL" << endl;
		exit(1);
	}
	cfg = c;
	may_talk = cfg->dontDetach();
	memset(pfd, 0, MAX_POLLFDS * sizeof(struct pollfd));
	for (i = 0; i < MAX_POLLFDS; i++)
		pfd[i].fd = -1;
	listen();
}

Server::~Server(void)
{
	int i;

	for (i = 0; i < MAX_POLLFDS; i++) {
		if (pfd[i].fd != -1)
		    close(pfd[i].fd);
	}
}

void
Server::listen(void)
{
	struct addrinfo        *res, *res0 = cfg->getListenAddr();
	int			error, i;
	const int		yes = 1;
	char			sbuf[NI_MAXSERV];

	addr_count = 0;
	for (res = res0, i = 0; res != NULL && i < MAX_POLLFDS;
	     res = res->ai_next) {
		char	hbuf[MAXHOSTNAMELEN];

		if ((pfd[i].fd = socket(res->ai_family, res->ai_socktype,
					res->ai_protocol)) == -1) {
			if (errno == EAFNOSUPPORT) {
				/*
				 * This happens on Frankenstein'ed systems
				 * where the kernel doesn't support IPv6
				 * and the resolver returns IPv6 addresses
				 * anyways.
				 */
				cerr << "Warning: socket: "
					<< strerror(errno) << endl;
				continue;
			} else {
				cerr << "Error: socket: "
					<< strerror(errno) << endl;
				exit(1);
			}
		}

		if (setsockopt(pfd[i].fd, SOL_SOCKET, SO_REUSEADDR, &yes,
			       sizeof(yes)) == -1) {
			cerr << "Error: setsockopt(SO_REUSEADDR): "
				<< strerror(errno) << endl;
			exit(1);
		}

		if (res->ai_family == AF_INET6
		    && setsockopt(pfd[i].fd, IPPROTO_IPV6, IPV6_V6ONLY, &yes,
				  sizeof(yes)) == -1) {
			cerr << "Error: setsockopt(IPV6_V6ONLY): "
				<< strerror(errno) << endl;
			exit(1);
		}

		if (bind(pfd[i].fd, res->ai_addr, res->ai_addrlen) == -1) {
			cerr << "Error: bind: " << strerror(errno) << endl;
			exit(1);
		}

		if (::listen(pfd[i].fd, 10) == -1) {
			cerr << "Error: listen: " << strerror(errno) << endl;
			exit(1);
		}

		pfd[i].events = POLLIN;
		i++;
		addr_count++;

		error = getnameinfo(res->ai_addr, res->ai_addrlen,
				    hbuf, sizeof(hbuf), NULL, 0,
				    NI_NUMERICHOST);
		if (error) {
			cerr << "Error: getnameinfo (1): "
				<< gai_strerror(error) << endl;
			exit(1);
		}

		listenAddrs.push_back(hbuf);
	}

	error = getnameinfo(res0->ai_addr, res0->ai_addrlen, NULL, 0,
			    sbuf, sizeof(sbuf), NI_NUMERICSERV);
	if (error) {
		cerr << "Error: getnameinfo (2): " << gai_strerror(error)
			<< endl;
		exit(1);
	}

	listenService = sbuf;
}

vector<struct client>
Server::poll(void)
{
	vector<struct client> clients;
	int pret, i;

	if ((pret = ::poll(pfd, addr_count, -1)) == -1) {
		if (errno != EINTR)
			XSI_Log::log()->err(__func__, errno, "poll");
		/*
		 * A signal interrupted us, probably SIGALRM. Return empty
		 * clients vector to tell the caller that it should do some
		 * maintenance.
		 */
		return (clients);
	}

	/*
	if (pret == 0)
		return (clients);
	*/

	for (i = 0; i < addr_count; i++) {
		struct sockaddr_storage c_addr;
		socklen_t c_addr_len = sizeof(c_addr);

		if (pfd[i].revents & (POLLERR | POLLNVAL))
			XSI_Log::log()->err(__func__, "poll error");
		if (pfd[i].revents & POLLIN) {
			struct client	client;

			if ((client.fd = accept(pfd[i].fd,
						(struct sockaddr *)&c_addr,
						&c_addr_len)) == -1)
				XSI_Log::log()->err(__func__, errno, "accept");
			memcpy(&client.c_addr, &c_addr, c_addr_len);
			client.c_addr_len = c_addr_len;

			clients.push_back(client);
		} else
			continue;
	}

	if (pret != (int)clients.size())
		XSI_Log::log()->err(__func__, "Internal error: Assertion (pret == clients.size()) failed");

	return (clients);
}

vector<string>
Server::getListenAddrStrings(void)
{
	return (listenAddrs);
}

string
Server::getListenServiceString(void)
{
	return (listenService);
}
