/*	$Kolabore: solaris.cpp,v 1.63 2007/02/09 20:58:49 maxx Exp $	*/
/*
 * Copyright (c) 2006, 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * solaris.cpp:
 * Gathering information and stats on Solaris 10 or newer.
 */

#include "log.h"

extern "C" {
#include <stdio.h>	 // <sys/mnttab.h> is screwed up and needs this, here.
#include <sys/param.h>
#include <sys/loadavg.h>
#include <sys/mnttab.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sys/stat.h>
#include <sys/statvfs.h>
#include <sys/swap.h>
#include <sys/sysinfo.h>
#include <sys/systeminfo.h>

#include <net/if.h>

#include <dirent.h>
#include <fcntl.h>
#include <kstat.h>
#include <netdb.h>
#include <picl.h>
#include <procfs.h>
#include <strings.h>
#include <unistd.h>
}

#include <cctype>
#include <cerrno>
#include <climits>
#include <cstdlib>
#include <cstring>
#include <ctime>

#include <string>
#include <vector>

#include "solaris.h"

#ifndef SA_LEN
#  ifdef HAVE_STRUCT_SOCKADDR_SA_LEN
#    define SA_LEN(x)	((x)->sa_len)
#  else
#    define SA_LEN(x)	((x)->sa_family == AF_INET6 ? \
			sizeof(struct sockaddr_in6) : \
			sizeof(struct sockaddr_in))
#  endif /* HAVE_STRUCT_SOCKADDR_SA_LEN */
#endif /* SA_LEN */

#define SMALLBUFSIZ	128
#define FILESYSTEM_DESCRIPTION_FILE	"/etc/mnttab"
#ifndef PROC_PATH
#  define PROC_PATH			"/proc"
#endif /* PROC_PATH */

using namespace std;

static bool xsi_in6_fillscopeid(struct sockaddr_in6 *sin6);
extern "C" {
	static int addSensor_callback(picl_nodehdl_t, void *);
}
static int solaris_sensor_id;

struct solaris_private {
	kstat_ctl_t *kc;
	time_t boottime;
};

struct solaris_disk_io {
	string name;
	uint64_t rbytes;
	uint64_t wbytes;
	uint64_t rxfers;
	uint64_t wxfers;
};

struct solaris_interrupt {
	string source;
	uint64_t count;
};

enum solaris_sensor_type {
	SENSOR_T_UNSPEC,
	SENSOR_T_TEMP,
	SENSOR_T_FAN
};

struct solaris_interface {
	string name;
	bool up;
	bool running;
	bool promiscuous;
	bool loopback;
	bool is_ptp;
	bool ptp_is_in6_linklocal;
	unsigned int mtu;
	int metric;
	struct sockaddr_storage ptpdest;
	size_t ptpsiz;
	vector<XSI_IfAddress> addrs;
};

struct solaris_sensor {
	enum solaris_sensor_type type;
	int id;
	string name;
	string device;
	string descr;
	int64_t value;
	enum xsi_sensor_status status;
	bool sensor_valid;
};

/*
 * XSI Constructor
 *
 * !! WARNING !!
 * !! Runs before dropping privileges when running as root
 * !! WARNING !!
 */
XSI::XSI(void)
{
	priv = NULL;
	num_cpu = -1;
}

XSI::~XSI(void)
{
	kstat_close(priv->kc);
	if (priv != NULL)
		delete priv;
}

/*
 * Information read once on startup
 */
void
XSI::collectSysInfo(void)
{
	solarisInit();

	readSysinfo();
	readOSInfo();
}

/*
 * Stats to be updated regularly
 */
void
XSI::updateStats(void)
{
	updateKstat();

	updateCPUFreq();
	updateCPUStats();
	updateVMStats();
	updateFSStats();
	updateDiskStats();
	updateOSUptime();
	updateLoadAvg();
	updateProcInfo();
	updateInterrupts();
	updateSensors();
	updateNetStats();
}

void
XSI::solarisInit(void)
{
	static bool	first = true;
	int		i, j;

	if (first)
		first = false;
	else
		return;

	if ((priv = new struct solaris_private) == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");

	if ((priv->kc = kstat_open()) == NULL)
		XSI_Log::log()->err(__func__, errno, "kstat_open");

	updateKstat();
	priv->boottime = getNamedKstatUValue("unix", -1, "system_misc",
					     "boot_time");
	if ((num_cpu = (int32_t)getSysconfValue(_SC_NPROCESSORS_CONF)) < 1)
		XSI_Log::log()->err(__func__, "Error: No CPUs found");
	system.setNumCPUs(num_cpu);
	cur_cp_time = new uint64_t*[num_cpu];
	last_cp_time = new uint64_t*[num_cpu];
	if (cur_cp_time == NULL || last_cp_time == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	for (i = 0; i < num_cpu; i++) {
		cur_cp_time[i] = new uint64_t[CPU_STATES];
		last_cp_time[i] = new uint64_t[CPU_STATES];
		if (cur_cp_time[i] == NULL || last_cp_time[i] == NULL)
			XSI_Log::log()->err(__func__, "Out of memory");
		for (j = 0; j < CPU_STATES; j++) {
			cur_cp_time[i][j] = 0;
			last_cp_time[i][j] = 0;
		}
	}
}

void
XSI::updateKstat(void)
{
	if (kstat_chain_update(priv->kc) == -1)
		XSI_Log::log()->err(__func__, errno, "kstat_chain_update");
}

void
XSI::readSysinfo(void)
{
	system.setModel(getNamedKstatString("cpu_info", -1, NULL,
					    "brand").c_str());
	system.setVendor(getSysinfoString(SI_HW_PROVIDER).c_str());
	system.setSerialNo(getSysinfoString(SI_HW_SERIAL).c_str());
}

void
XSI::readOSInfo(void)
{
	char buf[BUFSIZ];
	char *p, *p2, *build;
	string os_type = getSysinfoString(SI_SYSNAME);
	string os_release = getSysinfoString(SI_RELEASE);
	string os_kernel_string = getSysinfoString(SI_VERSION);
	string os_kernel, os_version_string, patchlevel, kernel_name, buildstr;

	if (os_kernel_string.find("-", 0) < os_kernel_string.size()) {
		/* Looks like the Solaris way of things ... */
		build = NULL;
		strlcpy(buf, os_kernel_string.c_str(), sizeof(buf));
		if ((p = strrchr(buf, '-')) != NULL)
			build = p + 1;
		if ((p = strrchr(buf, '_')) != NULL) {
			p2 = p;
			p++;
			while (isdigit(*p))
				patchlevel += *p++;
			p = buf;
			while (p != p2)
				kernel_name += *p++;
		}

		if (build == NULL ||
		    patchlevel.empty() ||
		    kernel_name.empty())
			XSI_Log::log()->err(__func__, "Internal error: Kernel version parser #1 broken. Please report this!");

		os_kernel = kernel_name;
		os_kernel += "#";
		os_kernel += build;
		os_kernel += " p";
		os_kernel += patchlevel;
	} else {
		/*
		 * We're probably OpenSolaris or so, which does stuff like
		 * "snv_51".
		 */
		strlcpy(buf, os_kernel_string.c_str(), sizeof(buf));
		if ((p = strrchr(buf, '_')) != NULL) {
			p2 = p;
			p++;
			while (isdigit(*p))
				buildstr += *p++;
			p = buf;
			while (p != p2)
				kernel_name += *p++;
		}
		patchlevel = "0";

		if (buildstr.empty() ||
		    patchlevel.empty() ||
		    kernel_name.empty())
			XSI_Log::log()->err(__func__, "Internal error: Kernel version parser #2 broken. Please report this!");

		os_kernel = kernel_name;
		os_kernel += "#";
		os_kernel += buildstr;
		os_kernel += " p";
		os_kernel += patchlevel;
	}

	os_version_string = os_type + " ";
	os_version_string += os_release + " ";
	os_version_string += os_kernel_string + " ";
	os_version_string += getSysinfoString(SI_MACHINE) + " ";
	os_version_string += getSysinfoString(SI_ARCHITECTURE) + " ";
	os_version_string += getSysinfoString(SI_PLATFORM);

	system.setOSInfo(NULL, os_kernel.c_str(), os_version_string.c_str());
}

void
XSI::updateCPUFreq(void)
{
	system.setCPUFreq((int32_t)getNamedKstatValue("cpu_info", -1, NULL,
						      "clock_MHz"));
}

void
XSI::updateCPUStats(void)
{
	uint64_t	       *total;
	uint64_t	      **used;
	double		       *percent;
	vector<struct cpu>	cpus;
	int			i, j;

	for (i = 0; i < num_cpu; i++) {
		cpu_stat_t     *cpu;
		kstat_t        *ksp;
		char		name[20];

		snprintf(name, sizeof(name), "cpu_stat%d", i);
		if ((ksp = kstat_lookup(priv->kc, const_cast<char *>("cpu_stat"),
					-1, const_cast<char *>(name))) == NULL)
			XSI_Log::log()->err(__func__, errno, "kstat_lookup");
		if (kstat_read(priv->kc, ksp, NULL) == -1)
			XSI_Log::log()->err(__func__, errno, "kstat_read");
		cpu = (cpu_stat_t *)ksp->ks_data;

		for (j = 0; j < CPU_STATES; j++)
			cur_cp_time[i][j] = cpu->cpu_sysinfo.cpu[j];
	}

	total = new uint64_t[num_cpu];
	used = new uint64_t*[num_cpu];
	percent = new double[num_cpu];
	if (total == NULL || used == NULL || percent == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");
	for (i = 0; i < num_cpu; i++) {
		used[i] = new uint64_t[CPU_STATES];
		if (used[i] == NULL)
			XSI_Log::log()->err(__func__, "Out of memory");
	}

	for (i = 0; i < num_cpu; i++) {
		struct cpu	cpu;

		total[i] = 0;
		for (j = 0; j < CPU_STATES; j++) {
			used[i][j] = cur_cp_time[i][j] - last_cp_time[i][j];
			last_cp_time[i][j] = cur_cp_time[i][j];
			total[i] += used[i][j];
		}
		if (total[i] > 0)
			percent[i] = 100 / (double)total[i];
		else
			percent[i] = 0;
		cpu.num = i;
		cpu.system = used[i][CPU_KERNEL] * percent[i];
		cpu.user = used[i][CPU_USER] * percent[i];
#ifdef CPU_NICE
		/* I can't believe they don't get this sooner or later, and I
		 * doubt it will be called any other than "CPU_NICE". */
		cpu.nice = used[i][CPU_NICE] * percent[i];
#else
		cpu.nice = 0;
#endif
		cpu.interrupt = used[i][CPU_WAIT] * percent[i];
		cpu.idle = used[i][CPU_IDLE] * percent[i];
		cpus.push_back(cpu);
	}

	system.setCPUs(cpus);

	for (i = 0; i < num_cpu; i++)
		delete[] used[i];
	delete[] used;
	delete[] total;
	delete[] percent;
}

void
XSI::updateVMStats(void)
{
	struct vm_info	vm_info;
	long		pages, pagesize, apages;
	int		num_swapdevs, i, n;
	swaptbl_t      *swt = NULL;
	char	       *strtab;
	bool		try_again;

	if ((pages = getSysconfValue(_SC_PHYS_PAGES)) == -1)
		XSI_Log::log()->err(__func__, "Cannot get number of pages");
	if ((apages = getSysconfValue(_SC_AVPHYS_PAGES)) == -1)
		XSI_Log::log()->err(__func__, "Cannot get number of free pages");
	if ((pagesize = getSysconfValue(_SC_PAGESIZE)) == -1)
		XSI_Log::log()->err(__func__, "Cannot get pagesize");

	vm_info.ram_total = (uint64_t)pages * (uint64_t)pagesize;
	vm_info.ram_free = (uint64_t)apages * (uint64_t)pagesize;
	vm_info.swap_total = vm_info.swap_free = 0;

	do {
		try_again = false;

		if ((num_swapdevs = swapctl(SC_GETNSWP, NULL)) == -1)
			XSI_Log::log()->err(__func__, errno, "swapctl: SC_GETNSWP");

		if (num_swapdevs == 0) {
			system.setVMInfo(vm_info);
			return;
		}

		/* Allocate space on the heap for num_swapdevs + 1 entries.
		 * The initial sizeof(swaptbl_t) is the `+ 1' */
		if ((swt = (swaptbl_t *)malloc(sizeof(swaptbl_t)
				  + num_swapdevs * sizeof(swapent_t))) == NULL)
			XSI_Log::log()->err(__func__, errno, "malloc");

		if ((strtab = (char *)malloc((num_swapdevs + 1) * PATH_MAX))
		    == NULL)
			XSI_Log::log()->err(__func__, errno, "malloc");

		for (i = 0; i < (num_swapdevs + 1); i++)
			swt->swt_ent[i].ste_path = strtab + (i * PATH_MAX);

		swt->swt_n = num_swapdevs + 1;

		if ((n = swapctl(SC_LIST, swt)) == -1)
			XSI_Log::log()->err(__func__, errno, "swapctl: SC_LIST");

		if (n > num_swapdevs) {
			free(swt);
			free(strtab);
			try_again = true;
		}
	} while (try_again);

	for (i = 0; i < n; i++) {
		vm_info.swap_total += (uint64_t)swt->swt_ent[i].ste_pages *
			(uint64_t)pagesize;
		vm_info.swap_free += (uint64_t)swt->swt_ent[i].ste_free *
			(uint64_t)pagesize;
	}

	free(swt);
	free(strtab);

	system.setVMInfo(vm_info);
}

void
XSI::updateFSStats(void)
{
	vector<struct fs>	filesystems;
	struct mnttab	        mnt;
	FILE		       *fp = NULL;
	int			ret;

	if ((fp = fopen(FILESYSTEM_DESCRIPTION_FILE, "r")) == NULL)
		XSI_Log::log()->err(__func__, errno, "fopen: " FILESYSTEM_DESCRIPTION_FILE);

	while ((ret = getmntent(fp, &mnt)) == 0) {
		struct fs	fs;
		struct statvfs64 f;
		string		mntopts = mnt.mnt_mntopts;
		uint64_t	blocksize;

		fs.mountpoint = mnt.mnt_mountp;
		fs.fs_name = mnt.mnt_special;
		fs.fs_type = mnt.mnt_fstype;

		if (statvfs64(mnt.mnt_mountp, &f) == -1) {
			if (errno == EACCES || errno == EIO)
				/*
				 * If the OS insists on not letting us look at
				 * something, or if a media we're trying to
				 * look at has an error, ignore and move on.
				 */
				continue;
			else
				XSI_Log::log()->err(__func__, errno, "statvfs64");
		}

		if ((blocksize = f.f_frsize) == 0)
			blocksize = f.f_bsize;

		fs.size_total = (uint64_t)f.f_blocks * blocksize;
		if (mntopts.find("ignore") < mntopts.size() ||
		    fs.size_total == 0 || fs.fs_type == "lofs" ||
		    fs.fs_type == "tmpfs" || fs.fs_type == "zfs")
			/*
			 * Ignore uninteresting filesystems and ZFS, which
			 * does not work with XSI's output in the <filesystems/>
			 * element. It would have to be handled separately.
			 */
			continue;

		fs.size_free = (uint64_t)f.f_bfree * blocksize;
		fs.size_avail = (int64_t)f.f_bavail * (int64_t)blocksize;
		if (f.f_files == UINT64_MAX)
			fs.nodes_total = 0;
		else
			fs.nodes_total = f.f_files;
		fs.nodes_free = f.f_ffree;

		filesystems.push_back(fs);
	}

	if (ret != -1)
		XSI_Log::log()->err(__func__, "%s is broken",
				    FILESYSTEM_DESCRIPTION_FILE);

	fclose(fp);

	storage.setFilesystems(filesystems);
}

void
XSI::updateDiskStats(void)
{
	vector<XSI_DiskIO>	old_disks = storage.getDisks();
	bool			drop_old_stats = false, disks_changed;
	kstat_t                *ksp;
	vector<struct solaris_disk_io> disks;
	vector<struct solaris_disk_io>::iterator ii;

	for (ksp = priv->kc->kc_chain; ksp != NULL; ksp = ksp->ks_next) {
		kstat_io_t kio;

		if (ksp->ks_type == KSTAT_TYPE_IO) {
			struct solaris_disk_io disk;

			if (kstat_read(priv->kc, ksp, &kio) == -1)
				XSI_Log::log()->err(__func__, errno, "kstat_read");

			disk.name = ksp->ks_name;
			/* Skip NFS mounts and USB root HUBs. :-P */
			if (disk.name.find("nfs", 0) == 0 ||
			    disk.name.find("ehci", 0) == 0 ||
			    disk.name.find("uhci", 0) == 0)
				continue;
			disk.rbytes = kio.nread;
			disk.wbytes = kio.nwritten;
			disk.rxfers = kio.reads;
			disk.wxfers = kio.writes;

			disks.push_back(disk);
		}
	}

	if (disks.size() != old_disks.size())
		drop_old_stats = true;

	do {
		disks_changed = false;
		if (drop_old_stats) {
			vector<XSI_DiskIO> new_disks;

			for (ii = disks.begin(); ii != disks.end(); ii++) {
				XSI_DiskIO disk(ii->name.c_str());

				disk.setWrittenBytes(ii->wbytes);
				disk.setReadBytes(ii->rbytes);
				disk.setWriteTransfers(ii->wxfers);
				disk.setReadTransfers(ii->rxfers);

				new_disks.push_back(disk);
			}
			storage.setDisks(new_disks);
		} else {
			for (unsigned int i = 0; i < disks.size(); i++) {
				if (disks[i].name != old_disks[i].getDiskName()) {
					disks_changed = true;
					drop_old_stats = true;
					break;
				}
				old_disks[i].setWrittenBytes(disks[i].wbytes);
				old_disks[i].setReadBytes(disks[i].rbytes);
				old_disks[i].setWriteTransfers(disks[i].wxfers);
				old_disks[i].setReadTransfers(disks[i].rxfers);
			}
			if (!disks_changed)
				storage.setDisks(old_disks);
		}
	} while (disks_changed);
}

void
XSI::updateOSUptime(void)
{
	time_t	now;

	time(&now);
	system.setOSUptime(now - priv->boottime);
}

void
XSI::updateLoadAvg(void)
{
	double v[3];

	if (getloadavg(v, 3) == -1)
		XSI_Log::log()->err(__func__, errno, "getloadavg");

	system.setLoadAvg(v);
}

void
XSI::updateProcInfo(void)
{
	vector<psinfo_t>	psinfovect;
	vector<psinfo_t>::iterator ps_it;
	vector<struct process_info> processes;
	static bool		have_warned = false;

	{
		const char     *procdir = PROC_PATH;
		DIR	       *dirp;
		struct dirent  *d;

		if ((dirp = opendir(procdir)) == NULL)
			XSI_Log::log()->err(__func__, errno, "%s", procdir);

		while ((d = readdir(dirp)) != NULL) {
			int             fd, pid;
			char            buf[SMALLBUFSIZ];
			psinfo_t        ps;
			ssize_t         rret;
			bool            retry;

			if (sscanf(d->d_name, "%d", &pid) != 1)
				continue;

			snprintf(buf, sizeof(buf), "%s/%d/psinfo",
				 procdir, pid);
			if ((fd = open(buf, O_RDONLY)) == -1) {
				if (errno == ENOENT)
					continue;
				else
					XSI_Log::log()->err(__func__, errno, "%s", buf);
			}

			do {
				retry = false;
				if ((rret = read(fd, &ps, sizeof(ps))) == -1) {
					if (errno == EAGAIN)
						retry = true;
					else
						XSI_Log::log()->err(__func__, errno, "read");
				}
			} while (retry);

			close (fd);

			if (rret != (ssize_t)sizeof(ps))
				continue;

			psinfovect.push_back(ps);
		}

		closedir(dirp);
	}

	if (psinfovect.size() == 0) {
		proc_info.setProcessVector(processes);
		return;
	}

	for (ps_it = psinfovect.begin(); ps_it != psinfovect.end(); ps_it++) {
		struct process_info	process;

		process.pid = ps_it->pr_pid;
		process.name = ps_it->pr_fname;
		process.euid = ps_it->pr_uid;
		process.username = user_cache->getUsername(process.euid);
		process.login_user = "";
		/* XXX:
		 * I'm having serious trouble including sys/proc.h to get the
		 * definitions of SRUN, SONPROC and SZOMB. They're defined as
		 * 2, 6 and 3 respectively, so just use those numbers here for
		 * now.
		 */
		process.running = ((int)ps_it->pr_lwp.pr_state == 2 /* SRUN */ ||
				   (int)ps_it->pr_lwp.pr_state == 6 /* SONPROC */);
		process.zombie = ((int)ps_it->pr_lwp.pr_state == 3 /* SZOMB */);
		/*
		 * From sys/procfs.h:
		 * "The following percent number is a 16-bit binary
		 * fraction [0 .. 1] with the binary point to the
		 * right of the high-order bit (1.0 == 0x8000)"
		 */
		if ((unsigned int)ps_it->pr_pctcpu > 0x8000) {
			if (!have_warned) {
				XSI_Log::log()->warn(NULL, "updateProcInfo() assertion failed: 0x%04x > 0x8000! Please report this.",
						     (unsigned int)ps_it->pr_pctcpu);
				have_warned = true;
			}
			process.pctcpu = 100.0;
		} else
			process.pctcpu = 100.0 *
				((double)((ps_it->pr_pctcpu & 0x8000) != 0) +
				 (double)(ps_it->pr_pctcpu & 0x7fff) / (double)0x7fff);
		process.ressize = ps_it->pr_rssize * 1024; /* Comes in kB. */
		process.nice = (int)ps_it->pr_lwp.pr_nice - 20;

		processes.push_back(process);
	}

	proc_info.setProcessVector(processes);
}

void
XSI::updateInterrupts(void)
{
	vector<XSI_Interrupt>	old_ints = system.getInterrupts();
	bool			drop_old_stats = false, ints_changed;
	kstat_t 	       *ksp;
	vector<struct solaris_interrupt> ints;
	vector<struct solaris_interrupt>::iterator ii;

	{
		struct solaris_interrupt clock;

		/* Only Sun knows why this is separate and "special". */
		clock.source = "clock";
		clock.count = getNamedKstatUValue("unix", 0, "system_misc",
						  "clk_intr");
		ints.push_back(clock);
	}

	for (ksp = priv->kc->kc_chain; ksp != NULL; ksp = ksp->ks_next) {
		if (ksp->ks_type == KSTAT_TYPE_INTR) {
			struct solaris_interrupt intr;
			int		i;
			kstat_intr_t   *ki;

			if (kstat_read(priv->kc, ksp, NULL) == -1)
				XSI_Log::log()->err(__func__, errno, "kstat_read");
			ki = KSTAT_INTR_PTR(ksp);

			intr.source = ksp->ks_name;
			intr.count = 0;
			for (i = 0; i < KSTAT_NUM_INTRS; i++)
				intr.count += ki->intrs[i];

			ints.push_back(intr);
		}
	}

	if (ints.size() != old_ints.size())
		drop_old_stats = true;

	do {
		ints_changed = false;
		if (drop_old_stats) {
			vector<XSI_Interrupt> new_ints;

			for (ii = ints.begin(); ii != ints.end(); ii++) {
				XSI_Interrupt intr(ii->source.c_str());

				intr.setIntsCount(ii->count);

				new_ints.push_back(intr);
			}
			system.setInterrupts(new_ints);
		} else {
			for (unsigned int i = 0; i < ints.size(); i++) {
				if (ints[i].source !=
				    old_ints[i].getSourceName()) {
					ints_changed = true;
					drop_old_stats = true;
					break;
				}
				old_ints[i].setIntsCount(ints[i].count);
			}
			if (!ints_changed)
				system.setInterrupts(old_ints);
		}
	} while (ints_changed);
}

void
XSI::updateSensors(void)
{
	picl_nodehdl_t root;
	int error;

	sensors.clear();

	if ((error = picl_initialize()) > 0)
		/*
		 * Don't care about the error, it's probably just picld not
		 * running (any more.)
		 */
		return;

	if ((error = picl_get_root(&root)) > 0)
		XSI_Log::log()->err(__func__, "picl_get_root: %s",
				    picl_strerror(error));

	solaris_sensor_id = 0;
	if ((error = picl_walk_tree_by_class(root, "temperature-sensor",
					     this, addSensor_callback)) > 0)
		XSI_Log::log()->err(__func__, "picl_walk_tree_by_class (temperature-sensor): %s",
				    picl_strerror(error));

	solaris_sensor_id = 0;
	if ((error = picl_walk_tree_by_class(root, "fan",
					     this, addSensor_callback)) > 0)
		XSI_Log::log()->err(__func__, "picl_walk_tree_by_class (fan): %s",
				    picl_strerror(error));

	picl_shutdown();
}

static int
addSensor_callback(picl_nodehdl_t node, void *caller)
{
	int			error;
	picl_propinfo_t 	pinfo;
	picl_prophdl_t		phdlr, *phdlr_prev;
	string			propname;
	struct solaris_sensor	sol_sensor;
	char			default_name[16];
	bool			is_percentage_sensor = false;
	XSI_Sensor		s;
	union {
		uint64_t ui;
		int64_t si;
		double d;
		uint8_t *str;
		char *cstr;
	} propvalue;

	sol_sensor.type = SENSOR_T_UNSPEC;
	sol_sensor.status = XSI_SENSOR_STATUS_UNSPEC;
	sol_sensor.device = "picl";
	sol_sensor.value = INT64_MIN;
	sol_sensor.sensor_valid = true;

	if ((error = picl_get_first_prop(node, &phdlr)) > 0)
		XSI_Log::log()->err(__func__, "picl_get_first_prop: %s",
				    picl_strerror(error));

	error = 0;
	do {
		phdlr_prev = &phdlr;

		if (error > 0)
			XSI_Log::log()->err(__func__, "picl_get_next_prop: %s",
					    picl_strerror(error));

		if ((error = picl_get_propinfo(phdlr, &pinfo)) > 0)
			XSI_Log::log()->err(__func__, "picl_get_propinfo: %s",
					    picl_strerror(error));

		if (pinfo.size == 0)
			XSI_Log::log()->err(__func__, "Internal error: pinfo.size is 0");
		propname = pinfo.name;

		switch (pinfo.type) {
		case PICL_PTYPE_BYTEARRAY:
			if ((propvalue.str = new uint8_t[pinfo.size]) == NULL)
				XSI_Log::log()->err(__func__, "Out of memory");
			if ((error = picl_get_propval(phdlr, (void *)propvalue.str,
						      pinfo.size)) > 0)
				XSI_Log::log()->err(__func__, "picl_get_propval: %s",
						    picl_strerror(error));
			break;
		case PICL_PTYPE_CHARSTRING:
			if ((propvalue.cstr = new char[pinfo.size]) == NULL)
				XSI_Log::log()->err(__func__, "Out of memory");
			if ((error = picl_get_propval(phdlr, (void *)propvalue.cstr,
						      pinfo.size)) > 0)
				XSI_Log::log()->err(__func__, "picl_get_propval: %s",
						    picl_strerror(error));
			break;
		case PICL_PTYPE_INT:
			if ((error = picl_get_propval(phdlr, (void *)&propvalue.si,
						      sizeof(propvalue.si))) > 0)
				XSI_Log::log()->err(__func__, "picl_get_propval: %s",
						    picl_strerror(error));
			break;
		case PICL_PTYPE_UNSIGNED_INT:
			if ((error = picl_get_propval(phdlr, (void *)&propvalue.ui,
						      sizeof(propvalue.ui))) > 0)
				XSI_Log::log()->err(__func__, "picl_get_propval: %s",
						    picl_strerror(error));
			break;
		case PICL_PTYPE_FLOAT:
			if ((error = picl_get_propval(phdlr, (void *)&propvalue.d,
						      sizeof(propvalue.d))) > 0)
				XSI_Log::log()->err(__func__, "picl_get_propval: %s",
						    picl_strerror(error));
			break;
		default:
			/* Read currently not supported */
			continue;
		}

		if (pinfo.type == PICL_PTYPE_CHARSTRING) {
			string tmp = propvalue.cstr;

			for (unsigned int i = 0; i < (tmp.size() - 1); i++)
				tmp[i] = tolower(tmp[i]);

			if (propname == "_class") {
				if (tmp == "temperature-sensor") {
					sol_sensor.type = SENSOR_T_TEMP;
					snprintf(default_name, sizeof(default_name),
						 "temp%d", solaris_sensor_id);
				} else if (tmp == "fan") {
					sol_sensor.type = SENSOR_T_FAN;
					snprintf(default_name, sizeof(default_name),
						 "fan%d", solaris_sensor_id);
				} else
					XSI_Log::log()->err(__func__, "Internal error: Unsupported type `%s' received",
							    tmp.c_str());
			}

			if (propname == "name")
				sol_sensor.name = propvalue.cstr;

			if (propname == "Label")
				sol_sensor.descr = propvalue.cstr;

			if (propname == "State") {
				if (tmp == "ok")
					sol_sensor.status = XSI_SENSOR_STATUS_OK;
				else if (tmp == "warning")
					sol_sensor.status = XSI_SENSOR_STATUS_WARN;
				else if (tmp == "failed")
					sol_sensor.status = XSI_SENSOR_STATUS_CRIT;
				else if (tmp == "unknown")
					sol_sensor.status = XSI_SENSOR_STATUS_UNKNOWN;
			}

			if (propname == "Speedunit") {
				if (tmp == "%")
					is_percentage_sensor = true;
			}
		}

		if (pinfo.type == PICL_PTYPE_INT) {
			if (propname == "Temperature")
				sol_sensor.value = propvalue.si * 1000000
					+ 273150000;

			if (propname == "Speed")
				sol_sensor.value = propvalue.si;
		}

		if (pinfo.type == PICL_PTYPE_BYTEARRAY)
			delete[] propvalue.str;
		else if (pinfo.type == PICL_PTYPE_CHARSTRING)
			delete[] propvalue.cstr;
	} while ((error = picl_get_next_prop(*phdlr_prev, &phdlr)) != PICL_ENDOFLIST);

	if (sol_sensor.name.empty())
		sol_sensor.name = default_name;
	sol_sensor.id = solaris_sensor_id++;

	if (sol_sensor.type == SENSOR_T_UNSPEC)
		XSI_Log::log()->err(__func__, "Internal error: Sensor type is still unspecified after reading properties");
	if (sol_sensor.value == INT64_MIN &&
	    (sol_sensor.status != XSI_SENSOR_STATUS_UNSPEC &&
	     sol_sensor.status != XSI_SENSOR_STATUS_UNKNOWN))
		XSI_Log::log()->err(__func__, "Internal error: Sensor `%s' has a status but no value?",
				    sol_sensor.name.c_str());

	if (sol_sensor.value == INT64_MIN) {
		sol_sensor.status = XSI_SENSOR_STATUS_UNKNOWN;
		sol_sensor.sensor_valid = false;
	}

	s.setSensorID(sol_sensor.id);
	s.setSensorDevice(sol_sensor.device.c_str());
	s.setSensorValue(sol_sensor.value);
	s.setSensorStatus(sol_sensor.status);
	s.setHasUnknownValue(sol_sensor.status == XSI_SENSOR_STATUS_UNKNOWN);
	s.setIsValid(sol_sensor.sensor_valid);

	switch (sol_sensor.type) {
	case SENSOR_T_TEMP:
		s.setSensorType(XSI_SENSOR_TYPE_TEMP);
		((XSI *)caller)->sensors.push_back(s);
		break;
	case SENSOR_T_FAN:
		if (is_percentage_sensor)
			s.setSensorType(XSI_SENSOR_TYPE_PERCENT);
		else
			s.setSensorType(XSI_SENSOR_TYPE_FANRPM);
		((XSI *)caller)->sensors.push_back(s);
		break;
	default:
		XSI_Log::log()->err(__func__, "Internal error: Unreachable code reached");
	}

	return (PICL_WALK_CONTINUE);
}

void
XSI::updateNetStats(void)
{
	struct lifreq	       *lifrp;
	struct lifnum		lifn;
	struct lifconf		lifc;
	int			sock;
	int			num_ifs, i;
	uint8_t 	       *buf;
	long			bufsiz;
	vector<XSI_NetIfStat>	old_nstats = network.getNetIfStats();
	bool			drop_old_stats = false, stats_changed;
	vector<struct solaris_interface> solifs;
	vector<struct solaris_interface>::iterator ii;

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
		XSI_Log::log()->err(__func__, errno, "socket");

	bzero(&lifn, sizeof(lifn));
	lifn.lifn_family = AF_UNSPEC;
	lifn.lifn_flags = LIFC_NOXMIT | LIFC_TEMPORARY | LIFC_ALLZONES;
	if (ioctl(sock, SIOCGLIFNUM, &lifn) == -1)
		XSI_Log::log()->err(__func__, errno, "SIOCGLIFNUM");
	num_ifs = lifn.lifn_count;

	if ((bufsiz = num_ifs * sizeof(struct lifreq)) <= 0)
		XSI_Log::log()->err(__func__, "Internal error: Negative number of interfaces");
	if ((buf = new uint8_t[(size_t)bufsiz]) == NULL)
		XSI_Log::log()->err(__func__, "Out of memory");

	bzero(&lifc, sizeof(lifc));
	lifc.lifc_family = AF_UNSPEC;
	lifc.lifc_flags = lifn.lifn_flags;
	lifc.lifc_len = bufsiz;
	lifc.lifc_buf = (caddr_t)buf;
	if (ioctl(sock, SIOCGLIFCONF, &lifc) == -1)
		XSI_Log::log()->err(__func__, errno, "SIOCGLIFCONF");

	for (i = 0, lifrp = lifc.lifc_req; i < num_ifs; i++, lifrp++) {
		bool found_if;
		int ifsock;
		struct lifreq lifr2;

		/*
		 * Solaris has one (logical) interface per address family,
		 * which doesn't fit in with XSI, nor is it anything like the
		 * other ports. Therefore, traverse our own list of interfaces,
		 * find out whether we already know about an interface or not,
		 * and then just add the additional information via the same
		 * iterator.
		 */
		found_if = false;
		for (ii = solifs.begin(); ii != solifs.end(); ii++) {
			if (ii->name == lifrp->lifr_name) {
				found_if = true;
				break;
			}
		}
		if (!found_if) {
			struct solaris_interface sif;

			solifs.push_back(sif);
			ii = solifs.end();
			ii--;
			ii->name = lifrp->lifr_name;
		}

		if ((ifsock = socket((lifrp->lifr_addr.ss_family == AF_UNSPEC)
				     ? AF_INET : lifrp->lifr_addr.ss_family,
				     SOCK_DGRAM, 0)) == -1)
			XSI_Log::log()->err(__func__, errno, "socket");

		bzero(&lifr2, sizeof(lifr2));
		strlcpy(lifr2.lifr_name, ii->name.c_str(),
			sizeof(lifr2.lifr_name));
		if (ioctl(ifsock, SIOCGLIFFLAGS, &lifr2) == -1)
			XSI_Log::log()->err(__func__, errno, "SIOCGLIFFLAGS");
		ii->up = (lifr2.lifr_flags & IFF_UP) ? true : false;
		ii->running = (lifr2.lifr_flags & IFF_RUNNING) ? true : false;
		ii->promiscuous = (lifr2.lifr_flags & IFF_PROMISC) ? true : false;
		ii->loopback = (lifr2.lifr_flags & IFF_LOOPBACK) ? true : false;
		ii->is_ptp = (lifr2.lifr_flags & IFF_POINTOPOINT) ? true : false;

		if (ii->is_ptp) {
			struct sockaddr *daddr;

			bzero(&lifr2, sizeof(lifr2));
			strlcpy(lifr2.lifr_name, ii->name.c_str(),
				sizeof(lifr2.lifr_name));
			if (ioctl(ifsock, SIOCGLIFDSTADDR, &lifr2) == -1)
				XSI_Log::log()->err(__func__, errno, "SIOCGLIFDSTADDR");

			daddr = (struct sockaddr *)&lifr2.lifr_dstaddr;
			ii->ptpsiz = SA_LEN(daddr);
			if (daddr->sa_family == AF_INET6)
				ii->ptp_is_in6_linklocal =
					xsi_in6_fillscopeid((struct sockaddr_in6 *)daddr);
			else
				ii->ptp_is_in6_linklocal = false;
			bzero(&ii->ptpdest, sizeof(ii->ptpdest));
			memcpy(&ii->ptpdest, daddr, ii->ptpsiz);
		}

		{
			struct sockaddr addr, *ap, *nmask;
			XSI_IfAddress xsiaddr;
			size_t addrsiz;
			int af;
			bool linklocal = false;

			bzero(&addr, sizeof(addr));
			bzero(&lifr2, sizeof(lifr2));
			strlcpy(lifr2.lifr_name, ii->name.c_str(),
				sizeof(lifr2.lifr_name));
			if (ioctl(ifsock, SIOCGLIFADDR, &lifr2) == -1)
				XSI_Log::log()->err(__func__, errno, "SIOCGLIFADDR");
			ap = (struct sockaddr *)&lifr2.lifr_addr;
			addrsiz = SA_LEN(ap);
			memcpy(&addr, ap, addrsiz);
			if ((af = ap->sa_family) == AF_INET6)
				linklocal = xsi_in6_fillscopeid((struct sockaddr_in6 *)&addr);

			bzero(&lifr2, sizeof(lifr2));
			strlcpy(lifr2.lifr_name, ii->name.c_str(),
				sizeof(lifr2.lifr_name));
			if (ioctl(ifsock, SIOCGLIFNETMASK, &lifr2) == -1)
				XSI_Log::log()->err(__func__, errno, "SIOCGLIFNETMASK");
			nmask = (struct sockaddr *)&lifr2.lifr_addr;

			xsiaddr.setAddress(&addr, addrsiz, nmask, af,
					   linklocal ? ii->name.c_str() : NULL);
			ii->addrs.push_back(xsiaddr);
		}

		bzero(&lifr2, sizeof(lifr2));
		strlcpy(lifr2.lifr_name, ii->name.c_str(),
			sizeof(lifr2.lifr_name));
		if (ioctl(ifsock, SIOCGLIFMTU, &lifr2) == -1)
			XSI_Log::log()->err(__func__, errno, "SIOCLIFMTU");
		ii->mtu = lifr2.lifr_mtu;

		bzero(&lifr2, sizeof(lifr2));
		strlcpy(lifr2.lifr_name, ii->name.c_str(),
			sizeof(lifr2.lifr_name));
		if (ioctl(ifsock, SIOCGLIFMETRIC, &lifr2) == -1)
			XSI_Log::log()->err(__func__, errno, "SIOCGLIFMETRIC");
		ii->metric = lifr2.lifr_metric;

		close(ifsock);
	}
	lifrp = NULL;   // Make it point to NULL instead of Nirvana

	delete[] buf;
	close(sock);

	if (solifs.size() != old_nstats.size())
		drop_old_stats = true;

	do {
		stats_changed = false;
		if (drop_old_stats) {
			vector<XSI_NetIfStat> new_nstats;

			for (ii = solifs.begin(); ii != solifs.end(); ii++) {
				XSI_NetIfStat	netIfStat(ii->name.c_str(),
							  ii->up);
				bool		error;
				uint64_t	ipkts;
				string		kmodule;
				int		kinstance = 0;

				{
					char	buf[SMALLBUFSIZ], *bp;

					strlcpy(buf, ii->name.c_str(),
						sizeof(buf));
					for (bp = buf; *bp != '\0'; bp++) {
						if (isalpha(*bp))
							kmodule += *bp;
						else {
							const char *errstr;
							kinstance = strtonum(bp, 0,
									     INT_MAX,
									     &errstr);
							if (errstr)
								XSI_Log::log()->err(__func__, "strtonum: %s is %s",
										    bp, errstr);
							break;
						}
					}
				}

				netIfStat.setIsRunning(ii->running);
				netIfStat.setIsPromiscuous(ii->promiscuous);
				netIfStat.setIsLoopback(ii->loopback);
				netIfStat.setMetric(ii->metric);
				netIfStat.setMTU((int)ii->mtu);
				netIfStat.setIfAddresses(ii->addrs);
				if (ii->is_ptp)
					netIfStat.setPTPDestAddress((struct sockaddr *)&ii->ptpdest,
								    ii->ptpsiz,
								    ii->ptp_is_in6_linklocal ?
								    ii->name.c_str()
								    : NULL);

				error = false;
				ipkts = getNamedKstatUValue(kmodule.c_str(), kinstance,
							    ii->name.c_str(), "ipackets64",
							    &error);
				if (!error) {
					netIfStat.setPacketsIn(ipkts);
					netIfStat.setPacketsOut(getNamedKstatUValue(kmodule.c_str(), kinstance,
										    ii->name.c_str(), "opackets64"));
					netIfStat.setBytesIn(getNamedKstatUValue(kmodule.c_str(), kinstance,
										 ii->name.c_str(), "rbytes64"));
					netIfStat.setBytesOut(getNamedKstatUValue(kmodule.c_str(), kinstance,
										  ii->name.c_str(), "obytes64"));
					netIfStat.setErrorsIn(getNamedKstatUValue(kmodule.c_str(), kinstance,
										  ii->name.c_str(), "ierrors"));
					netIfStat.setErrorsOut(getNamedKstatUValue(kmodule.c_str(), kinstance,
										   ii->name.c_str(), "oerrors"));
					netIfStat.setCollisions(getNamedKstatUValue(kmodule.c_str(), kinstance,
										    ii->name.c_str(), "collisions"));
				} else {
					netIfStat.setPacketsIn(0);
					netIfStat.setPacketsOut(0);
					netIfStat.setBytesIn(0);
					netIfStat.setBytesOut(0);
					netIfStat.setErrorsIn(0);
					netIfStat.setErrorsOut(0);
					netIfStat.setCollisions(0);
				}

				setLinkStatus(netIfStat, kmodule.c_str(),
					      kinstance, ii->name.c_str());
				setMediaString(netIfStat, kmodule.c_str(),
					       kinstance, ii->name.c_str());

				getDefaultRoutes();

				new_nstats.push_back(netIfStat);
			}
			network.setNetIfStats(new_nstats);
		} else {
			for (ii = solifs.begin(), i = 0;
			     ii != solifs.end();
			     ii++, i++) {
				bool		error;
				uint64_t	ipkts;
				string		kmodule;
				int		kinstance = 0;

				if (ii->name != old_nstats[i].getIfName()) {
					drop_old_stats = true;
					stats_changed = true;
					break;
				}

				{
					char	buf[SMALLBUFSIZ], *bp;

					strlcpy(buf, ii->name.c_str(),
						sizeof(buf));
					for (bp = buf; *bp != '\0'; bp++) {
						if (isalpha(*bp))
							kmodule += *bp;
						else {
							const char *errstr;
							kinstance = strtonum(bp, 0,
									     INT_MAX,
									     &errstr);
							if (errstr)
								XSI_Log::log()->err(__func__, "strtonum: %s is %s",
										    bp, errstr);
							break;
						}
					}
				}

				old_nstats[i].setIsRunning(ii->running);
				old_nstats[i].setIsPromiscuous(ii->promiscuous);
				old_nstats[i].setIsLoopback(ii->loopback);
				old_nstats[i].setMetric(ii->metric);
				old_nstats[i].setMTU((int)ii->mtu);
				old_nstats[i].setIfAddresses(ii->addrs);
				if (ii->is_ptp)
					old_nstats[i].setPTPDestAddress((struct sockaddr *)&ii->ptpdest,
									ii->ptpsiz,
									ii->ptp_is_in6_linklocal ?
									ii->name.c_str()
									: NULL);

				error = false;
				ipkts = getNamedKstatUValue(kmodule.c_str(), kinstance,
							    ii->name.c_str(), "ipackets64",
							    &error);
				if (!error) {
					old_nstats[i].setPacketsIn(ipkts);
					old_nstats[i].setPacketsOut(getNamedKstatUValue(kmodule.c_str(), kinstance,
											ii->name.c_str(), "opackets64"));
					old_nstats[i].setBytesIn(getNamedKstatUValue(kmodule.c_str(), kinstance,
										     ii->name.c_str(), "rbytes64"));
					old_nstats[i].setBytesOut(getNamedKstatUValue(kmodule.c_str(), kinstance,
										      ii->name.c_str(), "obytes64"));
					old_nstats[i].setErrorsIn(getNamedKstatUValue(kmodule.c_str(), kinstance,
										      ii->name.c_str(), "ierrors"));
					old_nstats[i].setErrorsOut(getNamedKstatUValue(kmodule.c_str(), kinstance,
										       ii->name.c_str(), "oerrors"));
					old_nstats[i].setCollisions(getNamedKstatUValue(kmodule.c_str(), kinstance,
											ii->name.c_str(), "collisions"));
				} else {
					old_nstats[i].setPacketsIn(0);
					old_nstats[i].setPacketsOut(0);
					old_nstats[i].setBytesIn(0);
					old_nstats[i].setBytesOut(0);
					old_nstats[i].setErrorsIn(0);
					old_nstats[i].setErrorsOut(0);
					old_nstats[i].setCollisions(0);
				}

				setLinkStatus(old_nstats[i], kmodule.c_str(),
					      kinstance, ii->name.c_str());
				setMediaString(old_nstats[i], kmodule.c_str(),
					       kinstance, ii->name.c_str());

				getDefaultRoutes();
			}
			if (!stats_changed)
				network.setNetIfStats(old_nstats);
		}
	} while (stats_changed);
}

void
XSI::setLinkStatus(XSI_NetIfStat &netIfStat, const char *kmodule,
		   int kinstance, const char *kname)
{
	bool		error;
	unsigned int	link_up;

	error = false;
	link_up = getNamedKstatUValue(kmodule, kinstance,
				      kname, "link_up",
				      &error);
	if (!error) {
		if (link_up > 0)
			netIfStat.setLinkStatus(IF_LINK_UP);
		else
			netIfStat.setLinkStatus(IF_LINK_DOWN);
	} else
		netIfStat.setLinkStatus(IF_LINK_UNKNOWN);
}

void
XSI::setMediaString(XSI_NetIfStat &netIfStat, const char *kmodule,
		    int kinstance, const char *kname)
{
	string		media, tmp;
	uint64_t	ifspeed;
	bool		error, have_duplex;
	unsigned int	link_duplex;

	media = "";
	error = false;
	tmp = getNamedKstatCharString(kmodule, kinstance, kname,
				      "media", &error);
	if (!error && tmp != "unknown")
		media = tmp;
	error = false;
	ifspeed = getNamedKstatUValue(kmodule, kinstance, kname,
				      "ifspeed", &error);
	if (!error) {
		char buf[SMALLBUFSIZ];

		buf[0] = '\0';
		if (ifspeed >= 1000000000) {
			if ((ifspeed % 1000000000) != 0)
				snprintf(buf, sizeof(buf), "%.2f GBit/s",
					 (double)ifspeed / 1000000000.0);
			else
				snprintf(buf, sizeof(buf), "%" PRIu64 " GBit/s",
					 ifspeed / 1000000000);
		} else if (ifspeed >= 1000000) {
			if ((ifspeed % 1000000) != 0)
				snprintf(buf, sizeof(buf), "%.2f MBit/s",
					 (double)ifspeed / 1000000.0);
			else
				snprintf(buf, sizeof(buf), "%" PRIu64 " MBit/s",
					 ifspeed / 1000000);
		} else if (ifspeed >= 1000) {
			if ((ifspeed % 1000) != 0)
				snprintf(buf, sizeof(buf), "%.2f KBit/s",
					 (double)ifspeed / 1000.0);
			else
				snprintf(buf, sizeof(buf), "%" PRIu64 " KBit/s",
					 ifspeed / 1000);
		} else if (ifspeed > 0)
			snprintf(buf, sizeof(buf), "%" PRIu64 " Bit/s",
				 ifspeed);

		if (buf[0] != '\0') {
			if (!media.empty())
				media += " ";
			media += buf;
		}
	}

	error = false;
	have_duplex = false;
	link_duplex = getNamedKstatUValue(kmodule, kinstance, kname,
					  "link_duplex", &error);
	if (!error) {
		bool parens = false;

		if (!media.empty() && link_duplex > 0 && link_duplex <= 2) {
			media += " (";
			parens = true;
		}

		have_duplex = true;
		switch (link_duplex) {
		case 2:
			netIfStat.setDuplexStatus(IF_DUPLEX_FULL);
			media += "full-duplex";
			break;
		case 1:
			netIfStat.setDuplexStatus(IF_DUPLEX_HALF);
			media += "half-duplex";
			break;
		default:
			netIfStat.setDuplexStatus(IF_DUPLEX_UNKNOWN);
			break;
		}

		if (parens)
			media += ")";
	} else
		netIfStat.setDuplexStatus(IF_DUPLEX_UNKNOWN);

	if (!have_duplex) {
		error = false;
		tmp = getNamedKstatCharString(kmodule, kinstance,
					      kname, "duplex",
					      &error);
		if (!error) {
			bool parens = false;

			if (!media.empty() && !tmp.empty() &&
			    tmp != "unknown") {
				media += " (";
				parens = true;
			}

			if (tmp == "full") {
				netIfStat.setDuplexStatus(IF_DUPLEX_FULL);
				media += "full-duplex";
			} else if (tmp == "half") {
				netIfStat.setDuplexStatus(IF_DUPLEX_HALF);
				media += "half-duplex";
			}

			if (parens)
				media += ")";
		}
	}

	netIfStat.setMediaString(media.c_str());
}

void
XSI::getDefaultRoutes(void)
{
	// XXX: missing
}

/*
 * Adapted from OpenBSD's src/sbin/ifconfig/ifconfig.c:in6_fillscopeid()
 */
static bool
xsi_in6_fillscopeid(struct sockaddr_in6 *sin6)
{
	if (IN6_IS_ADDR_LINKLOCAL(&sin6->sin6_addr)) {
		sin6->sin6_scope_id =
			ntohs(*(uint16_t *)&sin6->sin6_addr.s6_addr[2]);
		sin6->sin6_addr.s6_addr[2] = sin6->sin6_addr.s6_addr[3] = 0;
		return (true);
	}
	return (false);
}

string
XSI::getSysinfoString(int cmd)
{
	char	staticbuf[2];
	char   *buf;
	string	ret;
	size_t	siz;

	if ((siz = sysinfo(cmd, staticbuf, sizeof(staticbuf))) < 0)
		XSI_Log::log()->err(__func__, errno, "sysinfo");
	if (siz == 0)
		return (ret);

	buf = new char[siz];
	if (sysinfo(cmd, buf, siz) == -1)
		XSI_Log::log()->err(__func__, errno, "sysinfo");

	ret = buf;

	delete[] buf;
	return (ret);
}

long
XSI::getSysconfValue(int name)
{
	long ret;
	int old_errno = errno;

	if ((ret = sysconf(name)) == -1) {
		if (old_errno != errno)
			XSI_Log::log()->err(__func__, errno, "sysconf (%d)", name);
	}

	return (ret);
}

int64_t
XSI::getNamedKstatValue(const char *module, int instance, const char *ks_name,
			const char *name, bool *must_exist)
{
	kstat_t        *ksp;
	kstat_named_t  *knp;

	if ((ksp = kstat_lookup(priv->kc, const_cast<char *>(module), instance,
				const_cast<char *>(ks_name))) == NULL)
		XSI_Log::log()->err(__func__, errno, "kstat_lookup(%s, %d, %s)",
				    module, instance, ks_name);
	if (kstat_read(priv->kc, ksp, NULL) == -1)
		XSI_Log::log()->err(__func__, errno, "kstat_read");
	if ((knp = (kstat_named_t *)kstat_data_lookup(ksp, const_cast<char *>(name)))
	    == NULL) {
		if (must_exist && *must_exist == false && errno == ENOENT) {
			*must_exist = true;
			return (0);
		} else
			XSI_Log::log()->err(__func__, errno,
					    "kstat_data_lookup(%s)", name);
	}
	switch(knp->data_type) {
	case KSTAT_DATA_INT32:
		return (knp->value.i32);
	case KSTAT_DATA_INT64:
		return (knp->value.i64);
	default:
		XSI_Log::log()->err(__func__, "Internal error: Named kstat has wrong type. Expected INT, got %d",
				    (int)knp->data_type);
	}
}

uint64_t
XSI::getNamedKstatUValue(const char *module, int instance, const char *ks_name,
			 const char *name, bool *must_exist)
{
	kstat_t        *ksp;
	kstat_named_t  *knp;

	if ((ksp = kstat_lookup(priv->kc, const_cast<char *>(module), instance,
				const_cast<char *>(ks_name))) == NULL)
		XSI_Log::log()->err(__func__, errno, "kstat_lookup(%s, %d, %s)",
				    module, instance, ks_name);
	if (kstat_read(priv->kc, ksp, NULL) == -1)
		XSI_Log::log()->err(__func__, errno, "kstat_read");
	if ((knp = (kstat_named_t *)kstat_data_lookup(ksp, const_cast<char *>(name)))
	    == NULL) {
		if (must_exist && *must_exist == false && errno == ENOENT) {
			*must_exist = true;
			return (0);
		} else
			XSI_Log::log()->err(__func__, errno,
					    "kstat_data_lookup(%s)", name);
	}
	switch(knp->data_type) {
	case KSTAT_DATA_UINT32:
		return (knp->value.ui32);
	case KSTAT_DATA_UINT64:
		return (knp->value.ui64);
	default:
		XSI_Log::log()->err(__func__, "Internal error: Named kstat has wrong type. Expected UINT, got %d",
				    (int)knp->data_type);
	}
}

string
XSI::getNamedKstatString(const char *module, int instance, const char *ks_name,
			 const char *name, bool *must_exist)
{
	kstat_t        *ksp;
	kstat_named_t  *knp;
	string		ret;

	if ((ksp = kstat_lookup(priv->kc, const_cast<char *>(module), instance,
				const_cast<char *>(ks_name))) == NULL)
		XSI_Log::log()->err(__func__, errno, "kstat_lookup(%s, %d, %s)",
				    module, instance, ks_name);
	if (kstat_read(priv->kc, ksp, NULL) == -1)
		XSI_Log::log()->err(__func__, errno, "kstat_read");
	if ((knp = (kstat_named_t *)kstat_data_lookup(ksp, const_cast<char *>(name)))
	    == NULL) {
		if (must_exist && *must_exist == false && errno == ENOENT) {
			*must_exist = true;
			return (ret);
		} else
			XSI_Log::log()->err(__func__, errno,
					    "kstat_data_lookup(%s)", name);
	}
	switch(knp->data_type) {
	case KSTAT_DATA_STRING:
		/* XXX - kstat type marked "evolving" */
#ifdef HAVE_STRUCT_KSTAT_NAMED_VALUE_STRING
		ret = knp->value.string.addr.ptr;
#elif HAVE_STRUCT_KSTAT_NAMED_VALUE_STR
		ret = knp->value.str.addr.ptr;
#else
#  error "I don't know your struct kstat_named, yet!"
#endif
		return (ret);
	default:
		XSI_Log::log()->err(__func__, "Internal error: Named kstat has wrong type. Expected STRING, got %d",
				    (int)knp->data_type);
	}
}

string
XSI::getNamedKstatCharString(const char *module, int instance, const char *ks_name,
			     const char *name, bool *must_exist)
{
	kstat_t        *ksp;
	kstat_named_t  *knp;
	string		ret;
	unsigned int	i;

	if ((ksp = kstat_lookup(priv->kc, const_cast<char *>(module), instance,
				const_cast<char *>(ks_name))) == NULL)
		XSI_Log::log()->err(__func__, errno, "kstat_lookup(%s, %d, %s)",
				    module, instance, ks_name);
	if (kstat_read(priv->kc, ksp, NULL) == -1)
		XSI_Log::log()->err(__func__, errno, "kstat_read");
	if ((knp = (kstat_named_t *)kstat_data_lookup(ksp, const_cast<char *>(name)))
	    == NULL) {
		if (must_exist && *must_exist == false && errno == ENOENT) {
			*must_exist = true;
			return (ret);
		} else
			XSI_Log::log()->err(__func__, errno,
					    "kstat_data_lookup(%s)", name);
	}
	switch(knp->data_type) {
	case KSTAT_DATA_CHAR:
		for (i = 0; i < sizeof(knp->value.c); i++)
			if ((int)knp->value.c[i] >= 32)
				ret.push_back(knp->value.c[i]);
			else
				break;
		return (ret);
	default:
		XSI_Log::log()->err(__func__, "Internal error: Named kstat has wrong type. Expected CHAR, got %d",
				    (int)knp->data_type);
	}
}
