/*	$Kolabore: storage.cpp,v 1.18 2007/01/05 21:27:45 maxx Exp $	*/
/*
 * Copyright (c) 2006, 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "log.h"

#include <string>
#include <vector>

#include "storage.h"
#include "timekeeper.h"
#include "xmlsysinfo.h"

using namespace std;

XSI_DiskIO::XSI_DiskIO(void)
{
	bytes_written = last_bytes_written = 0;
	bytes_read = last_bytes_read = 0;
	transfers_write = last_transfers_write = 0;
	transfers_read = last_transfers_read = 0;
}

XSI_DiskIO::XSI_DiskIO(const char *name)
{
	disk_name = name;
	bytes_written = last_bytes_written = 0;
	bytes_read = last_bytes_read = 0;
	transfers_write = last_transfers_write = 0;
	transfers_read = last_transfers_read = 0;
}

void
XSI_DiskIO::setDiskName(const char *name)
{
	disk_name = name;
}

void
XSI_DiskIO::setWrittenBytes(uint64_t writtenb)
{
	last_bytes_written = bytes_written;
	bytes_written = writtenb;
}

void
XSI_DiskIO::setReadBytes(uint64_t readb)
{
	last_bytes_read = bytes_read;
	bytes_read = readb;
}

void
XSI_DiskIO::setWriteTransfers(uint64_t writet)
{
	last_transfers_write = transfers_write;
	transfers_write = writet;
}

void
XSI_DiskIO::setReadTransfers(uint64_t readt)
{
	last_transfers_read = transfers_read;
	transfers_read = readt;
}

bool
XSI_DiskIO::isIdle(void)
{
	return (getWriteBPS() == 0 &&
		getReadBPS() == 0 &&
		getWriteTPS() == 0 &&
		getReadTPS() == 0);
}

bool
XSI_DiskIO::hasBPS(void)
{
	return (bytes_written > 0 || bytes_read > 0);
}

bool
XSI_DiskIO::hasTPS(void)
{
	return (transfers_write > 0 || transfers_read > 0);
}

const char *
XSI_DiskIO::getDiskName(bool sanitize)
{
	if (sanitize)
		sanitizeString(disk_name, true);
	return (disk_name.c_str());
}

const char *
XSI_DiskIO::getWriteBPSString(void)
{
	return (makeFloatStr(getWriteBPS(), 1));
}

const char *
XSI_DiskIO::getReadBPSString(void)
{
	return (makeFloatStr(getReadBPS(), 1));
}

const char *
XSI_DiskIO::getWriteTPSString(void)
{
	return (makeFloatStr(getWriteTPS(), 1));
}

const char *
XSI_DiskIO::getReadTPSString(void)
{
	return (makeFloatStr(getReadTPS(), 1));
}

double
XSI_DiskIO::getWriteBPS(void)
{
	uint64_t value;

	if (!bytes_written || XSI_TimeKeeper::get()->timeDelta() < 0.01 ||
	    !last_bytes_written)
		return (0);

	value = bytes_written - last_bytes_written;

	if (value == 0)
		return (0);
	else
		return ((double)value / XSI_TimeKeeper::get()->timeDelta());
}

double
XSI_DiskIO::getReadBPS(void)
{
	uint64_t value;

	if (!bytes_read || XSI_TimeKeeper::get()->timeDelta() < 0.01 ||
	    !last_bytes_read)
		return (0);

	value = bytes_read - last_bytes_read;

	if (value == 0)
		return (0);
	else
		return ((double)value / XSI_TimeKeeper::get()->timeDelta());
}

double
XSI_DiskIO::getWriteTPS(void)
{
	uint64_t value;

	if (!transfers_write || XSI_TimeKeeper::get()->timeDelta() < 0.01 ||
	    !last_transfers_write)
		return (0);

	value = transfers_write - last_transfers_write;

	if (value == 0)
		return (0);
	else
		return ((double)value / XSI_TimeKeeper::get()->timeDelta());
}

double
XSI_DiskIO::getReadTPS(void)
{
	uint64_t value;

	if (!transfers_read || XSI_TimeKeeper::get()->timeDelta() < 0.01 ||
	    !last_transfers_read)
		return (0);

	value = transfers_read - last_transfers_read;

	if (value == 0)
		return (0);
	else
		return ((double)value / XSI_TimeKeeper::get()->timeDelta());
}


XSI_Storage::XSI_Storage(void)
{
	read_bps_total = write_bps_total = 0;
	up2date = false;
}

void
XSI_Storage::setFilesystems(vector<struct fs> f)
{
	filesystems = f;
}

void
XSI_Storage::updateIOTotals(void)
{
	vector<XSI_DiskIO>::iterator it;

	if (up2date)
		return;

	read_bps_total = write_bps_total = 0;
	read_tps_total = write_tps_total = 0;
	for (it = disks.begin(); it != disks.end(); it++) {
		write_bps_total += it->getWriteBPS();
		read_bps_total += it->getReadBPS();
		write_tps_total += it->getWriteTPS();
		read_tps_total += it->getReadTPS();
	}

	up2date = true;
}

bool
XSI_Storage::hasFilesystemInfo(void)
{
	return (filesystems.size() > 0);
}

bool
XSI_Storage::hasDiskInfo(void)
{
	return (disks.size() > 0);
}

vector<struct fs>
XSI_Storage::getFilesystems(void)
{
	return (filesystems);
}

void
XSI_Storage::setDisks(vector<XSI_DiskIO> d)
{
	disks = d;
	up2date = false;
}

vector<XSI_DiskIO>
XSI_Storage::getDisks(void)
{
	return (disks);
}

const char *
XSI_Storage::getReadBPSTotalString(void)
{
	if (!up2date)
		XSI_Log::log()->err(__func__,
				    "Internal error: I/O totals not up to date");
	return (makeFloatStr(read_bps_total, 1));
}

const char *
XSI_Storage::getWriteBPSTotalString(void)
{
	if (!up2date)
		XSI_Log::log()->err(__func__,
				    "Internal error: I/O totals not up to date");
	return (makeFloatStr(write_bps_total, 1));
}

const char *
XSI_Storage::getReadTPSTotalString(void)
{
	if (!up2date)
		XSI_Log::log()->err(__func__,
				    "Internal error: I/O totals not up to date");
	return (makeFloatStr(read_tps_total, 1));
}

const char *
XSI_Storage::getWriteTPSTotalString(void)
{
	if (!up2date)
		XSI_Log::log()->err(__func__,
				    "Internal error: I/O totals not up to date");
	return (makeFloatStr(write_tps_total, 1));
}
