/*	$Kolabore: system.cpp,v 1.31 2007/01/05 15:09:44 maxx Exp $	*/
/*
 * Copyright (c) 2006, 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "log.h"

extern "C" {
#include <sys/types.h>
}

#include <cstdio>

#include <string>
#include <vector>

#include "system.h"
#include "timekeeper.h"
#include "xmlsysinfo.h"

#define SMALLBUFSIZ	64

using namespace std;

XSI_Top::XSI_Top(void)
{
	has_cpu_hog = has_memory_hog = has_big_hog = false;
	cpu_top = processes.begin();
	mem_top = processes.begin();
	processed = false;
}

void
XSI_Top::setProcessVector(vector<struct process_info> newprocs)
{
	processes = newprocs;
	has_cpu_hog = has_memory_hog = has_big_hog = false;
	cpu_top = processes.begin();
	mem_top = processes.begin();
	processed = false;
}

void
XSI_Top::updateProcessStats(void)
{
	vector<struct process_info>::iterator it;

	running_procs = zombie_procs = 0;
	for (it = processes.begin(); it != processes.end(); it++) {
		if (it->running)
			running_procs++;
		if (it->zombie)
			zombie_procs++;
		if (it->pctcpu >= 0.005 &&
		    (cpu_top == it || it->pctcpu > cpu_top->pctcpu)) {
			cpu_top = it;
			has_cpu_hog = true;
		}
		if (it->ressize > 1024 &&
		    (mem_top == it || it->ressize > mem_top->ressize)) {
			mem_top = it;
			has_memory_hog = true;
		}
	}

	if (cpu_top == mem_top && has_cpu_hog && has_memory_hog)
		has_big_hog = true;

	if (has_memory_hog) {
		sanitizeString(mem_top->name);
		sanitizeString(mem_top->login_user);
	}
	if (has_cpu_hog && !has_big_hog) {
		sanitizeString(cpu_top->name);
		sanitizeString(cpu_top->login_user);
	}

	processed = true;
}

bool
XSI_Top::hasProcesses(void)
{
	return (processed && !processes.empty());
}

bool
XSI_Top::hasCPUHog(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (has_cpu_hog);
}

bool
XSI_Top::hasMemoryHog(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (has_memory_hog);
}

bool
XSI_Top::hasOneBigHog(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (has_big_hog);
}

bool
XSI_Top::hasUsername(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (!cpu_top->username.empty() &&
		!mem_top->username.empty());
}

bool
XSI_Top::hasLoginUsername(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (!cpu_top->login_user.empty() &&
		!mem_top->login_user.empty());
}

bool
XSI_Top::hasZombies(void)
{
	return (zombie_procs > 0);
}

const char *
XSI_Top::getNumProcString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return(makeIntStr((uint32_t)processes.size()));
}

const char *
XSI_Top::getNumRunningString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (makeIntStr(running_procs));
}

const char *
XSI_Top::getNumZombiesString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (makeIntStr(zombie_procs));
}

bool
XSI_Top::topCPUHogIsRunning(void)
{
	return (cpu_top->running);
}

const char *
XSI_Top::getTopCPUHogPIDString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (makeIntStr(cpu_top->pid));
}

const char *
XSI_Top::getTopCPUHogNameString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (cpu_top->name.c_str());
}

const char *
XSI_Top::getTopCPUHogEUIDString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (makeIntStr(cpu_top->euid));
}

const char *
XSI_Top::getTopCPUHogUsernameString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (cpu_top->username.c_str());
}

const char *
XSI_Top::getTopCPUHogLoginUsernameString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (cpu_top->login_user.c_str());
}

const char *
XSI_Top::getTopCPUHogIsRunningBoolString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return ((cpu_top->running) ? "true" : "false");
}

const char *
XSI_Top::getTopCPUHogPctCPUString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (makeFloatStr(cpu_top->pctcpu, 1, 2));
}

const char *
XSI_Top::getTopCPUHogResSizeString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (makeIntStr(cpu_top->ressize));
}

const char *
XSI_Top::getTopCPUHogNiceString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	if (cpu_top->nice == 0)
		return (NULL);
	else
		return (makeIntStr(cpu_top->nice));
}

bool
XSI_Top::topMemHogIsRunning(void)
{
	return (mem_top->running);
}

const char *
XSI_Top::getTopMemHogPIDString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (makeIntStr(mem_top->pid));
}

const char *
XSI_Top::getTopMemHogNameString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (mem_top->name.c_str());
}

const char *
XSI_Top::getTopMemHogEUIDString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (makeIntStr(mem_top->euid));
}

const char *
XSI_Top::getTopMemHogUsernameString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (mem_top->username.c_str());
}

const char *
XSI_Top::getTopMemHogLoginUsernameString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (mem_top->login_user.c_str());
}

const char *
XSI_Top::getTopMemHogIsRunningBoolString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return ((mem_top->running) ? "true" : "false");
}

const char *
XSI_Top::getTopMemHogPctCPUString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (makeFloatStr(mem_top->pctcpu, 1, 2));
}

const char *
XSI_Top::getTopMemHogResSizeString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	return (makeIntStr(mem_top->ressize));
}

const char *
XSI_Top::getTopMemHogNiceString(void)
{
	if (!processed)
		XSI_Log::log()->err(__func__, "Internal error: Process list not processed, yet");

	if (mem_top->nice == 0)
		return (NULL);
	else
		return (makeIntStr(mem_top->nice));
}

XSI_Interrupt::XSI_Interrupt(void)
{
	count = last_count = 0;
}

XSI_Interrupt::XSI_Interrupt(const char *src)
{
	name = src;
	count = last_count = 0;
}

void
XSI_Interrupt::setSourceName(const char *src)
{
	name = src;
}

void
XSI_Interrupt::setIntsCount(uint64_t n)
{
	last_count = count;
	count = n;
}

bool
XSI_Interrupt::hasInterrupts(void)
{
	return (count > 0);
}

const char *
XSI_Interrupt::getSourceName(bool sanitize)
{
	if (sanitize)
		sanitizeString(name, true);
	return (name.c_str());
}

double
XSI_Interrupt::getIntsPerSec(void)
{
	uint64_t value;

	if (!count || XSI_TimeKeeper::get()->timeDelta() < 0.01 || !last_count)
		return (0);

	value = count - last_count;

	if (value == 0)
		return (0);
	else
		return ((double)value / XSI_TimeKeeper::get()->timeDelta());
}

const char *
XSI_Interrupt::getIntsPerSecString(void)
{
	return (makeFloatStr(getIntsPerSec(), 1));
}


XSI_System::XSI_System(void)
{
	has_loadavg = has_vminfo = false;
	vm_info.ram_total = vm_info.ram_free = 0;
	vm_info.swap_total = vm_info.swap_free = 0;
	uptime_seconds = -1;
	load_averages[0] = load_averages[1] = load_averages[2] = 0;
	num_cpus = cpu_freq = -1;
}

void
XSI_System::setArch(const char *str)
{
	arch = str;
	sanitizeString(arch);
}

void
XSI_System::setModel(const char *str)
{
	model = str;
}

void
XSI_System::setVendor(const char *str)
{
	vendor = str;
}

void
XSI_System::setProduct(const char *str)
{
	product = str;
}

void
XSI_System::setVersion(const char *str)
{
	version = str;
}

void
XSI_System::setSerialNo(const char *str)
{
	serial_no = str;
}

void
XSI_System::setUUID(const char *str)
{
	uuid = str;
}

void
XSI_System::setNumCPUs(int32_t ncpu)
{
	num_cpus = ncpu;
}

void
XSI_System::setCPUFreq(int32_t speed)
{
	cpu_freq = speed;
}

void
XSI_System::setCPUs(vector<struct cpu> ncpus)
{
	cpus = ncpus;
}

void
XSI_System::setInterrupts(vector<XSI_Interrupt> ints)
{
	interrupts = ints;
}

void
XSI_System::setVMInfo(struct vm_info mem)
{
	vm_info = mem;
	has_vminfo = true;
}

void
XSI_System::setLoadAvg(double load[3])
{
	load_averages[0] = load[0];
	load_averages[1] = load[1];
	load_averages[2] = load[2];
	has_loadavg = true;
}

void
XSI_System::setOSType(const char *type)
{
	os_type = type;
	sanitizeString(os_type);
}

void
XSI_System::setOSRelease(const char *release)
{
	os_release = release;
	sanitizeString(os_release);
}

void
XSI_System::setOSInfo(const char *status, const char *kernel,
		      const char *version_string)
{
	if (status)
		os_status = status;
	os_kernel = kernel;
	os_version_string = version_string;
}

void
XSI_System::setOSUptime(time_t uptime)
{
	uptime_seconds = (int64_t)uptime;
}

bool
XSI_System::hasCPUInfo(void)
{
	return (num_cpus > 0);
}

bool
XSI_System::hasVMInfo(void)
{
	return (has_vminfo);
}

bool
XSI_System::hasUptime(void)
{
	return (uptime_seconds >= 0);
}

bool
XSI_System::hasLoadAvg(void)
{
	return (has_loadavg);
}

const char *
XSI_System::getArch(void)
{
	return (arch.c_str());
}

const char *
XSI_System::getModel(bool sanitize)
{
	if (model.empty())
		return (NULL);
	else {
		if (sanitize)
			sanitizeString(model);
		return (model.c_str());
	}
}

const char *
XSI_System::getCPUFreqString(void)
{
	if (cpu_freq == -1)
		return (NULL);
	else
		return (makeIntStr(cpu_freq));
}

const char *
XSI_System::getVendor(bool sanitize)
{
	if (vendor.empty())
		return (NULL);
	else {
		if (sanitize)
			sanitizeString(vendor);
		return (vendor.c_str());
	}
}

const char *
XSI_System::getProduct(bool sanitize)
{
	if (product.empty())
		return (NULL);
	else {
		if (sanitize)
			normalizeString(product);
		return (product.c_str());
	}
}

const char *
XSI_System::getVersion(bool sanitize)
{
	if (version.empty())
		return (NULL);
	else {
		if (sanitize)
			sanitizeString(version);
		return (version.c_str());
	}
}

const char *
XSI_System::getSerialNo(bool sanitize)
{
	if (serial_no.empty())
		return (NULL);
	else {
		if (sanitize)
			normalizeString(serial_no);
		return (serial_no.c_str());
	}
}

const char *
XSI_System::getUUID(bool sanitize)
{
	if (uuid.empty())
		return (NULL);
	else {
		if (sanitize)
			sanitizeString(uuid);
		return (uuid.c_str());
	}
}

const char *
XSI_System::getNumCPUString(void)
{
	return (makeIntStr(num_cpus));
}

vector<struct cpu>
XSI_System::getCPUs(void)
{
	return (cpus);
}

vector<XSI_Interrupt>
XSI_System::getInterrupts(void)
{
	return (interrupts);
}

struct vm_info
XSI_System::getVMInfo(void)
{
	return (vm_info);
}

const char *
XSI_System::getLoadAvg(int sample)
{
	char buf[SMALLBUFSIZ];
	static string ret;

	if (sample < 0 || sample > 2)
		XSI_Log::log()->err(__func__, "Internal error: Bad sample number requested");
	snprintf(buf, sizeof(buf), "%.2f", load_averages[sample]);
	ret = buf;
	return (ret.c_str());
}

const char *
XSI_System::getOSType(void)
{
	return (os_type.c_str());
}

const char *
XSI_System::getOSRelease(void)
{
	return (os_release.c_str());
}

const char *
XSI_System::getOSStatus(bool sanitize)
{
	if (os_status.empty())
		return (NULL);
	else {
		if (sanitize)
			sanitizeString(os_status);
		return (os_status.c_str());
	}
}

const char *
XSI_System::getOSKernel(void)
{
	sanitizeString(os_kernel);
	return (os_kernel.c_str());
}

const char *
XSI_System::getOSVersionString(void)
{
	sanitizeString(os_version_string);
	return (os_version_string.c_str());
}

const char *
XSI_System::getOSUptime(void)
{
	if (uptime_seconds < 0)
		return (NULL);
	else
		return (makeIntStr(uptime_seconds));
}
