/*	$Kolabore: xmlsysinfo.cpp,v 1.159 2007/01/10 14:34:37 maxx Exp $	*/
/*
 * Copyright (c) 2006, 2007 Moritz Grimm <gtgbr@gmx.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "log.h"

extern "C" {
#include <sys/types.h>
#include <sys/socket.h>

#include <unistd.h>

#include <libxml/tree.h>
}

#include <cstring>

#include <list>
#include <string>
#include <vector>

#include "xmlsysinfo.h"
#include "timekeeper.h"

#define SMALLBUFSIZ		64
#ifndef USER_CACHE_TIMEOUT
#  define USER_CACHE_TIMEOUT	900
#endif

using namespace std;


/***************************************************************************/

XSIBase::XSIBase(void)
{
	sysInfoDocument = NULL;
	configuration = NULL;
	user_cache = NULL;
}

XSIBase::~XSIBase(void)
{
	if (sysInfoDocument != NULL)
		xmlFreeDoc(sysInfoDocument);
	xmlCleanupParser();
	if (user_cache != NULL)
		delete user_cache;
}

void
XSIBase::XSIBase_init(Config *cfg)
{
	configuration = cfg;
	user_cache = new XSI_UserCache(configuration->noNames(),
				       USER_CACHE_TIMEOUT);
}

void
XSIBase::setListenAddrs(vector<string> laddrs)
{
	listenAddrs = laddrs;
}

void
XSIBase::setListenService(string service)
{
	listenService = service;
}

void
XSIBase::readSysInfo(void)
{
	if (user_cache == NULL)
		XSI_Log::log()->err(__func__, "Internal error: XMLSysInfo is not properly initialized");
	collectCommonInfo();
	collectSysInfo();
	updateSysInfo();
}

void
XSIBase::updateSysInfo(void)
{
	updateTimeKeeper();
	updateStats();
	updatePacketfilterStats();
}

void
XSIBase::buildSysInfoDoc(void)
{
	xmlNodePtr	root_node = NULL, node1 = NULL, node2 = NULL,
			node3 = NULL, node4 = NULL, node5 = NULL;
	xmlNsPtr	ns = NULL;
	char		buffer[BUFSIZ];
	unsigned int	num_valid_sensors;

	if (sysInfoDocument != NULL)
		xmlFreeDoc(sysInfoDocument);

	/*
	 * XXX:
	 * The whole namespace nonsense that follows is a stupid hack, simply
	 * to coerce libxml2 into putting out what I want. I am certain that
	 * this is not how it's supposed to be done, but it serves its purpose
	 * until either the secret, proper way of doing this is discovered,
	 * libxml actually supports XML Schemas, ...
	 */
	sysInfoDocument = xmlNewDoc(BAD_CAST "1.0");
	root_node = xmlNewNode(NULL, BAD_CAST "xmlsysinfo");
	ns = xmlNewNs(root_node, BAD_CAST XSI_XSD_NAMESPACE, NULL);
	xmlNewNsProp(root_node, ns, BAD_CAST "xmlns:xsi",
		     BAD_CAST "http://www.w3.org/2001/XMLSchema-instance");
	xmlNewNsProp(root_node, ns, BAD_CAST "xsi:schemaLocation",
		     BAD_CAST XSI_XSD_NS_URI_STRING);
	xmlDocSetRootElement(sysInfoDocument, root_node);

	/* XSI CONFIGURATION */

	node1 = xmlNewNode(NULL, BAD_CAST "xsi_config");
	xmlAddChild(root_node, node1);
	xmlNewProp(node1, BAD_CAST "schema_version", BAD_CAST XSI_XSD_VERSION);
	xmlNewProp(node1, BAD_CAST "xsi_version", BAD_CAST PACKAGE_VERSION);

	xmlNewChild(node1, NULL, BAD_CAST "update_timestamp", BAD_CAST XSI_TimeKeeper::get()->updateTimestampString());
	{
		list<string> opts = configuration->getXSIOptionsList();
		list<string>::iterator it;

		for (it = opts.begin(); it != opts.end(); it++) {
			xmlNewChild(node1, NULL, BAD_CAST (*it).c_str(), NULL);
		}
	}
	if (!configuration->oneShot()) {
		vector<string>::iterator it;

		if (XSI_TimeKeeper::get()->haveTimeDelta())
			xmlNewChild(node1, NULL, BAD_CAST "update_interval", BAD_CAST makeFloatStr(XSI_TimeKeeper::get()->timeDelta(), 1, 4));

		for (it = listenAddrs.begin(); it != listenAddrs.end(); it++) {
			xmlNewChild(node1, NULL, BAD_CAST "listen_addr", BAD_CAST (*it).c_str());
		}
		xmlNewChild(node1, NULL, BAD_CAST "listen_port", BAD_CAST listenService.c_str());
	}

	/* SYSTEM INFO */

	node1 = xmlNewNode(NULL, BAD_CAST "system");
	xmlAddChild(root_node, node1);

	node2 = xmlNewNode(NULL, BAD_CAST "machine");
	xmlAddChild(node1, node2);
	xmlNewChild(node2, NULL, BAD_CAST "arch",
		    BAD_CAST system.getArch());
	if (system.getModel())
		xmlNewChild(node2, NULL, BAD_CAST "model",
			    BAD_CAST system.getModel(true));
	if (system.getCPUFreqString())
		xmlNewChild(node2, NULL, BAD_CAST "cpu_freq",
			    BAD_CAST system.getCPUFreqString());
	if (system.getVendor())
		xmlNewChild(node2, NULL, BAD_CAST "vendor",
			    BAD_CAST system.getVendor(true));
	if (system.getProduct())
		xmlNewChild(node2, NULL, BAD_CAST "product",
			    BAD_CAST system.getProduct(true));
	if (system.getVersion())
		xmlNewChild(node2, NULL, BAD_CAST "version",
			    BAD_CAST system.getVersion(true));
	if (system.getSerialNo())
		xmlNewChild(node2, NULL, BAD_CAST "serial_no",
			    BAD_CAST system.getSerialNo(true));
	if (system.getUUID())
		xmlNewChild(node2, NULL, BAD_CAST "uuid",
			    BAD_CAST system.getUUID(true));

	if (system.hasCPUInfo()) {
		node2 = xmlNewNode(NULL, BAD_CAST "cpus");
		xmlAddChild(node1, node2);
		xmlNewChild(node2, NULL, BAD_CAST "count",
			    BAD_CAST system.getNumCPUString());
		{
			vector<struct cpu> cpuvect = system.getCPUs();
			vector<struct cpu>::iterator it;

			for (it = cpuvect.begin(); it != cpuvect.end(); it++) {
				node3 = xmlNewNode(NULL, BAD_CAST (configuration->oneShot()
								   ? "cpu_load_total"
								   : "cpu_load"));
				/*
				 * OSs that can only report overall CPU loads, even on
				 * SMP systems, set only one struct cpu and set num to
				 * -1.
				 */
				if (cpuvect.size() != 1 || it->num != -1) {
					snprintf(buffer, sizeof(buffer), "%d", it->num);
					xmlNewProp(node3, BAD_CAST "num", BAD_CAST buffer);
				}
				xmlAddChild(node2, node3);
				snprintf(buffer, sizeof(buffer), "%.2f",
					 it->system);
				xmlNewChild(node3, NULL, BAD_CAST "system",
					    BAD_CAST buffer);
				snprintf(buffer, sizeof(buffer), "%.2f",
					 it->user);
				xmlNewChild(node3, NULL, BAD_CAST "user",
					    BAD_CAST buffer);
				snprintf(buffer, sizeof(buffer), "%.2f",
					 it->nice);
				xmlNewChild(node3, NULL, BAD_CAST "nice",
					    BAD_CAST buffer);
				snprintf(buffer, sizeof(buffer), "%.2f",
					 it->interrupt);
				xmlNewChild(node3, NULL, BAD_CAST "interrupt",
					    BAD_CAST buffer);
				snprintf(buffer, sizeof(buffer), "%.2f",
					 it->idle);
				xmlNewChild(node3, NULL, BAD_CAST "idle",
					    BAD_CAST buffer);
			}
		}
	}

	if (!configuration->oneShot() && !system.getInterrupts().empty()) {
		node2 = xmlNewNode(NULL, BAD_CAST "interrupts");
		xmlAddChild(node1, node2);
		{
			vector<XSI_Interrupt> intrvect = system.getInterrupts();
			vector<XSI_Interrupt>::iterator it;
			double	totalIntrs = 0;

			for (it = intrvect.begin(); it != intrvect.end(); it++) {
				if (!it->hasInterrupts())
					continue;
				node3 = xmlNewChild(node2, NULL, BAD_CAST "ints_persec", BAD_CAST it->getIntsPerSecString());
				xmlNewProp(node3, BAD_CAST "source", BAD_CAST it->getSourceName(true));
				totalIntrs += it->getIntsPerSec();
			}

			xmlNewChild(node2, NULL, BAD_CAST "ints_persec_total", BAD_CAST makeFloatStr(totalIntrs, 1));
		}
	}

	if (system.hasVMInfo()) {
		node2 = xmlNewNode(NULL, BAD_CAST "memory");
		xmlAddChild(node1, node2);

		node3 = xmlNewNode(NULL, BAD_CAST "ram");
		xmlAddChild(node2, node3);
		snprintf(buffer, sizeof(buffer), "%" PRIu64,
			 system.getVMInfo().ram_total);
		xmlNewChild(node3, NULL, BAD_CAST "total", BAD_CAST buffer);
		snprintf(buffer, sizeof(buffer), "%" PRIu64,
			 system.getVMInfo().ram_free);
		xmlNewChild(node3, NULL, BAD_CAST "free", BAD_CAST buffer);

		node3 = xmlNewNode(NULL, BAD_CAST "swap");
		xmlAddChild(node2, node3);
		snprintf(buffer, sizeof(buffer), "%" PRIu64,
			 system.getVMInfo().swap_total);
		xmlNewChild(node3, NULL, BAD_CAST "total", BAD_CAST buffer);
		snprintf(buffer, sizeof(buffer), "%" PRIu64,
			 system.getVMInfo().swap_free);
		xmlNewChild(node3, NULL, BAD_CAST "free", BAD_CAST buffer);
	}

	node2 = xmlNewNode(NULL, BAD_CAST "os");
	xmlAddChild(node1, node2);
	xmlNewChild(node2, NULL, BAD_CAST "type",
		    BAD_CAST system.getOSType());
	xmlNewChild(node2, NULL, BAD_CAST "release",
		    BAD_CAST system.getOSRelease());
	if (system.getOSStatus() != NULL)
		xmlNewChild(node2, NULL, BAD_CAST "os_status",
			    BAD_CAST system.getOSStatus(true));
	xmlNewChild(node2, NULL, BAD_CAST "kernel",
		    BAD_CAST system.getOSKernel());
	xmlNewChild(node2, NULL, BAD_CAST "version_string",
		    BAD_CAST system.getOSVersionString());

	if (system.hasUptime())
		xmlNewChild(node2, NULL, BAD_CAST "uptime", BAD_CAST system.getOSUptime());

	if (system.hasLoadAvg()) {
		node3 = xmlNewNode(NULL, BAD_CAST "load_averages");
		xmlAddChild(node2, node3);
		xmlNewChild(node3, NULL, BAD_CAST "avg1min", BAD_CAST system.getLoadAvg(0));
		xmlNewChild(node3, NULL, BAD_CAST "avg5min", BAD_CAST system.getLoadAvg(1));
		xmlNewChild(node3, NULL, BAD_CAST "avg15min", BAD_CAST system.getLoadAvg(2));
	}

	proc_info.updateProcessStats();
	if (proc_info.hasProcesses()) {
		node3 = xmlNewNode(NULL, BAD_CAST "processes");
		xmlAddChild(node2, node3);
		xmlNewChild(node3, NULL, BAD_CAST "count", BAD_CAST proc_info.getNumProcString());
		xmlNewChild(node3, NULL, BAD_CAST "running", BAD_CAST proc_info.getNumRunningString());
		if (proc_info.hasZombies())
			xmlNewChild(node3, NULL, BAD_CAST "zombies", BAD_CAST proc_info.getNumZombiesString());
		if (proc_info.hasOneBigHog()) {
			node4 = xmlNewNode(NULL, BAD_CAST "top");
			xmlAddChild(node3, node4);
			xmlNewChild(node4, NULL, BAD_CAST "name", BAD_CAST proc_info.getTopCPUHogNameString());
			xmlNewChild(node4, NULL, BAD_CAST "pid", BAD_CAST proc_info.getTopCPUHogPIDString());
			node5 = xmlNewChild(node4, NULL, BAD_CAST "euid", BAD_CAST proc_info.getTopCPUHogEUIDString());
			if (proc_info.hasUsername() && !configuration->noNames())
				xmlNewProp(node5, BAD_CAST "username", BAD_CAST proc_info.getTopCPUHogUsernameString());
			if (proc_info.hasLoginUsername() && !configuration->noNames())
				xmlNewChild(node4, NULL, BAD_CAST "login_user", BAD_CAST proc_info.getTopCPUHogLoginUsernameString());
			xmlNewChild(node4, NULL, BAD_CAST "cpu_percent", BAD_CAST proc_info.getTopCPUHogPctCPUString());
			xmlNewChild(node4, NULL, BAD_CAST "res_size", BAD_CAST proc_info.getTopCPUHogResSizeString());
			if (proc_info.getTopCPUHogNiceString())
				xmlNewChild(node4, NULL, BAD_CAST "nice", BAD_CAST proc_info.getTopCPUHogNiceString());
			if (proc_info.topCPUHogIsRunning())
				xmlNewChild(node4, NULL, BAD_CAST "running", NULL);
		} else {
			if (proc_info.hasCPUHog()) {
				node4 = xmlNewNode(NULL, BAD_CAST "cpu_top");
				xmlAddChild(node3, node4);
				xmlNewChild(node4, NULL, BAD_CAST "name", BAD_CAST proc_info.getTopCPUHogNameString());
				xmlNewChild(node4, NULL, BAD_CAST "pid", BAD_CAST proc_info.getTopCPUHogPIDString());
				node5 = xmlNewChild(node4, NULL, BAD_CAST "euid", BAD_CAST proc_info.getTopCPUHogEUIDString());
				if (proc_info.hasUsername() && !configuration->noNames())
					xmlNewProp(node5, BAD_CAST "username", BAD_CAST proc_info.getTopCPUHogUsernameString());
				if (proc_info.hasLoginUsername() && !configuration->noNames())
					xmlNewChild(node4, NULL, BAD_CAST "login_user", BAD_CAST proc_info.getTopCPUHogLoginUsernameString());
				xmlNewChild(node4, NULL, BAD_CAST "cpu_percent", BAD_CAST proc_info.getTopCPUHogPctCPUString());
				xmlNewChild(node4, NULL, BAD_CAST "res_size", BAD_CAST proc_info.getTopCPUHogResSizeString());
				if (proc_info.getTopCPUHogNiceString())
					xmlNewChild(node4, NULL, BAD_CAST "nice", BAD_CAST proc_info.getTopCPUHogNiceString());
				if (proc_info.topCPUHogIsRunning())
					xmlNewChild(node4, NULL, BAD_CAST "running", NULL);
			}
			if (proc_info.hasMemoryHog()) {
				node4 = xmlNewNode(NULL, BAD_CAST "mem_top");
				xmlAddChild(node3, node4);
				xmlNewChild(node4, NULL, BAD_CAST "name", BAD_CAST proc_info.getTopMemHogNameString());
				xmlNewChild(node4, NULL, BAD_CAST "pid", BAD_CAST proc_info.getTopMemHogPIDString());
				node5 = xmlNewChild(node4, NULL, BAD_CAST "euid", BAD_CAST proc_info.getTopMemHogEUIDString());
				if (proc_info.hasUsername() && !configuration->noNames())
					xmlNewProp(node5, BAD_CAST "username", BAD_CAST proc_info.getTopMemHogUsernameString());
				if (proc_info.hasLoginUsername() && !configuration->noNames())
					xmlNewChild(node4, NULL, BAD_CAST "login_user", BAD_CAST proc_info.getTopMemHogLoginUsernameString());
				xmlNewChild(node4, NULL, BAD_CAST "cpu_percent", BAD_CAST proc_info.getTopMemHogPctCPUString());
				xmlNewChild(node4, NULL, BAD_CAST "res_size", BAD_CAST proc_info.getTopMemHogResSizeString());
				if (proc_info.getTopMemHogNiceString())
					xmlNewChild(node4, NULL, BAD_CAST "nice", BAD_CAST proc_info.getTopMemHogNiceString());
				if (proc_info.topMemHogIsRunning())
					xmlNewChild(node4, NULL, BAD_CAST "running", NULL);
			}
		}
	}

	/* SENSORS INFO */

	{
		vector<XSI_Sensor>::iterator it;

		num_valid_sensors = 0;
		for (it = sensors.begin(); it != sensors.end(); it++) {
			if (it->isValid())
				num_valid_sensors++;
		}
	}

	if (num_valid_sensors > 0) {
		vector<XSI_Sensor>::iterator it;

		node1 = xmlNewNode(NULL, BAD_CAST "sensors");
		xmlAddChild(root_node, node1);
		snprintf(buffer, sizeof(buffer), "%u", num_valid_sensors);
		xmlNewChild(node1, NULL, BAD_CAST "count", BAD_CAST buffer);

		for (it = sensors.begin(); it != sensors.end(); it++) {
			if (!it->isValid())
				continue;
			node2 = xmlNewNode(NULL, BAD_CAST "sensor");
			xmlAddChild(node1, node2);
			xmlNewProp(node2, BAD_CAST "id", BAD_CAST it->getSensorIDString());
			xmlNewProp(node2, BAD_CAST "dev", BAD_CAST it->getSensorDeviceString());
			xmlNewProp(node2, BAD_CAST "type", BAD_CAST it->getSensorTypeString());

			if (it->isUnknownSensor())
				xmlNewChild(node2, NULL, BAD_CAST "other_type", BAD_CAST it->getOtherSensorTypeString());
			if (it->hasStatus())
				xmlNewChild(node2, NULL, BAD_CAST "status", BAD_CAST it->getSensorStatusString());
			if (it->hasDescription())
				xmlNewChild(node2, NULL, BAD_CAST "sensor_descr", BAD_CAST it->getSensorDescriptionString());
			if (it->hasUnknownValue()) {
				xmlNewChild(node2, NULL, BAD_CAST "value_unknown", NULL);
				continue;
			}

			if (it->isIndicatorSensor()) {
				xmlNewChild(node2, NULL, BAD_CAST "indicator", BAD_CAST it->getSensorValueString());
			} else if (it->isDriveSensor()) {
				xmlNewChild(node2, NULL, BAD_CAST "drive_status", BAD_CAST it->getSensorValueString());
			} else {
				if (it->isUnknownSensor())
					node3 = xmlNewChild(node2, NULL, BAD_CAST "other_value", BAD_CAST it->getSensorValueString());
				else
					node3 = xmlNewChild(node2, NULL, BAD_CAST "value", BAD_CAST it->getSensorValueString());
				if (it->getSensorUnitString())
					xmlNewProp(node3, BAD_CAST "unit", BAD_CAST it->getSensorUnitString());
			}

			if (it->hasValueAge())
				xmlNewChild(node2, NULL, BAD_CAST "value_age", BAD_CAST it->getSensorValueAgeString());
		}
	}

	/* STORAGE INFO */

	if (storage.hasFilesystemInfo() || storage.hasDiskInfo() ||
	    raids.size() > 0) {
		vector<struct fs> fsvect = storage.getFilesystems();
		vector<struct fs>::iterator it;

		node1 = xmlNewNode(NULL, BAD_CAST "storage");
		xmlAddChild(root_node, node1);

		node2 = xmlNewNode(NULL, BAD_CAST "filesystems");
		xmlAddChild(node1, node2);
		snprintf(buffer, sizeof(buffer), "%" PRIu64, (uint64_t)fsvect.size());
		xmlNewChild(node2, NULL, BAD_CAST "count", BAD_CAST buffer);

		for (it = fsvect.begin(); it != fsvect.end(); it++) {
			node3 = xmlNewNode(NULL, BAD_CAST "mount");
			xmlAddChild(node2, node3);
			sanitizeString(it->fs_type, true);
			sanitizeString(it->fs_name, true);
			sanitizeString(it->mountpoint, true);
			xmlNewProp(node3, BAD_CAST "type", BAD_CAST it->fs_type.c_str());
			xmlNewProp(node3, BAD_CAST "from", BAD_CAST it->fs_name.c_str());
			xmlNewProp(node3, BAD_CAST "to", BAD_CAST it->mountpoint.c_str());

			node4 = xmlNewNode(NULL, BAD_CAST "size");
			xmlAddChild(node3, node4);
			snprintf(buffer, sizeof(buffer), "%" PRIu64, it->size_total);
			xmlNewChild(node4, NULL, BAD_CAST "total", BAD_CAST buffer);
			snprintf(buffer, sizeof(buffer), "%" PRIu64, it->size_free);
			xmlNewChild(node4, NULL, BAD_CAST "free", BAD_CAST buffer);
			snprintf(buffer, sizeof(buffer), "%" PRIi64, it->size_avail);
			xmlNewChild(node4, NULL, BAD_CAST "available", BAD_CAST buffer);

			node4 = xmlNewNode(NULL, BAD_CAST "nodes");
			xmlAddChild(node3, node4);
			snprintf(buffer, sizeof(buffer), "%" PRIu64, it->nodes_total);
			xmlNewChild(node4, NULL, BAD_CAST "total", BAD_CAST buffer);
			snprintf(buffer, sizeof(buffer), "%" PRIu64, it->nodes_free);
			xmlNewChild(node4, NULL, BAD_CAST "free", BAD_CAST buffer);
		}

		if (!configuration->oneShot() && storage.hasDiskInfo()) {
			bool	have_tps = false, have_bps = false;

			storage.updateIOTotals();
			node2 = xmlNewNode(NULL, BAD_CAST "disks");
			xmlAddChild(node1, node2);
			snprintf(buffer, sizeof(buffer), "%" PRIi64, (int64_t)storage.getDisks().size());
			xmlNewChild(node2, NULL, BAD_CAST "count", BAD_CAST buffer);
			{
				vector<XSI_DiskIO> dIO = storage.getDisks();
				vector<XSI_DiskIO>::iterator it;

				for (it = dIO.begin(); it != dIO.end(); it++) {
					node3 = xmlNewNode(NULL, BAD_CAST "disk");
					xmlAddChild(node2, node3);
					xmlNewProp(node3, BAD_CAST "name", BAD_CAST it->getDiskName(true));
					if (it->isIdle()) {
						xmlNewProp(node3, BAD_CAST "idle", BAD_CAST "true");
						continue;
					}

					if (it->hasTPS()) {
						have_tps = true;

						node4 = xmlNewNode(NULL, BAD_CAST "transfers");
						xmlAddChild(node3, node4);

						xmlNewChild(node4, NULL, BAD_CAST "read_persec", BAD_CAST it->getReadTPSString());
						xmlNewChild(node4, NULL, BAD_CAST "write_persec", BAD_CAST it->getWriteTPSString());
					}

					if (it->hasBPS()) {
						have_bps = true;

						node4 = xmlNewNode(NULL, BAD_CAST "bytes");
						xmlAddChild(node3, node4);

						xmlNewChild(node4, NULL, BAD_CAST "read_persec", BAD_CAST it->getReadBPSString());
						xmlNewChild(node4, NULL, BAD_CAST "write_persec", BAD_CAST it->getWriteBPSString());
					}
				}
			}
			node3 = xmlNewNode(NULL, BAD_CAST "disk_summary");
			xmlAddChild(node2, node3);

			if (!have_tps && !have_bps)
				xmlNewProp(node3, BAD_CAST "idle", BAD_CAST "true");

			if (have_tps) {
				node4 = xmlNewNode(NULL, BAD_CAST "transfers");
				xmlAddChild(node3, node4);

				xmlNewChild(node4, NULL, BAD_CAST "read_persec", BAD_CAST storage.getReadTPSTotalString());
				xmlNewChild(node4, NULL, BAD_CAST "write_persec", BAD_CAST storage.getWriteTPSTotalString());
			}

			if (have_bps) {
				node4 = xmlNewNode(NULL, BAD_CAST "bytes");
				xmlAddChild(node3, node4);

				xmlNewChild(node4, NULL, BAD_CAST "read_persec", BAD_CAST storage.getReadBPSTotalString());
				xmlNewChild(node4, NULL, BAD_CAST "write_persec", BAD_CAST storage.getWriteBPSTotalString());
			}
		}

		if (configuration->showRaid()) {
			vector<XSI_Raid>::iterator it;

			for (it = raids.begin(); it != raids.end(); it++) {
				vector<XSI_RVolume>		rvol = it->getVolumes();
				vector<XSI_RDisk>		rdisk = it->getReserveDisks();
				vector<XSI_RVolume>::iterator	vit;
				vector<XSI_RDisk>::iterator	rit;

				node2 = xmlNewNode(NULL, BAD_CAST "raid");
				xmlAddChild(node1, node2);
				xmlNewProp(node2, BAD_CAST "controller", BAD_CAST it->getRaidControllerName(true));
				if (rvol.size() > 0) {
					for (vit = rvol.begin(); vit != rvol.end(); vit++) {
						vector<XSI_RDisk> vdisk = vit->getVolumeDisks();
						vector<XSI_RDisk>::iterator dit;

						if (vdisk.size() == 0)
							XSI_Log::log()->err(__func__, "Internal error: Can't have RAID volumes without disks");

						node3 = xmlNewNode(NULL, BAD_CAST "volume");
						xmlAddChild(node2, node3);
						xmlNewProp(node3, BAD_CAST "dev", BAD_CAST vit->getVolumeDevName());
						xmlNewProp(node3, BAD_CAST "id", BAD_CAST vit->getVolumeIDString());
						xmlNewChild(node3, NULL, BAD_CAST "status", BAD_CAST vit->getVolumeStatusString());
						xmlNewChild(node3, NULL, BAD_CAST "size", BAD_CAST vit->getVolumeSizeString());
						xmlNewChild(node3, NULL, BAD_CAST "level", BAD_CAST vit->getRaidLevelString());
						if (vit->getVolumeOpProgressString() != NULL ||
						    vit->getVolumeOpElapsedString() != NULL) {
							node4 = xmlNewNode(NULL, BAD_CAST "task");
							xmlAddChild(node3, node4);
							if (vit->getVolumeOpProgressString() != NULL)
								xmlNewChild(node4, NULL, BAD_CAST "progress_percent", BAD_CAST vit->getVolumeOpProgressString());
							if (vit->getVolumeOpElapsedString() != NULL)
								xmlNewChild(node4, NULL, BAD_CAST "elapsed_time", BAD_CAST vit->getVolumeOpElapsedString());
						}
						for (dit = vdisk.begin(); dit != vdisk.end(); dit++) {
							node4 = xmlNewNode(NULL, BAD_CAST "volume_disk");
							xmlAddChild(node3, node4);
							xmlNewProp(node4, BAD_CAST "id", BAD_CAST dit->getDiskIDString());
							xmlNewProp(node4, BAD_CAST "bus", BAD_CAST dit->getDiskBusString());
							xmlNewProp(node4, BAD_CAST "target", BAD_CAST dit->getDiskTargetString());
							if (dit->getDiskLUNString() != NULL)
								xmlNewProp(node4, BAD_CAST "lun", BAD_CAST dit->getDiskLUNString());
							xmlNewChild(node4, NULL, BAD_CAST "status", BAD_CAST dit->getDiskStatusString());
							xmlNewChild(node4, NULL, BAD_CAST "size", BAD_CAST dit->getDiskSizeString());
							if (dit->getDiskEnclosureDevName() != NULL)
								xmlNewChild(node4, NULL, BAD_CAST "enclosure", BAD_CAST dit->getDiskEnclosureDevName(true));
							if (dit->getDiskVendorName() != NULL)
								xmlNewChild(node4, NULL, BAD_CAST "vendor", BAD_CAST dit->getDiskVendorName(true));
							if (dit->getDiskSerial() != NULL)
								xmlNewChild(node4, NULL, BAD_CAST "serial_no", BAD_CAST dit->getDiskSerial(true));
						}
					}
				}
				if (rdisk.size() > 0) {
					for (rit = rdisk.begin(); rit != rdisk.end(); rit++) {
						node3 = xmlNewNode(NULL, BAD_CAST "reserve_disk");
						xmlAddChild(node2, node3);
						xmlNewProp(node3, BAD_CAST "id", BAD_CAST rit->getDiskIDString());
						xmlNewProp(node3, BAD_CAST "bus", BAD_CAST rit->getDiskBusString());
						xmlNewProp(node3, BAD_CAST "target", BAD_CAST rit->getDiskTargetString());
						if (rit->getDiskLUNString() != NULL)
							xmlNewProp(node3, BAD_CAST "lun", BAD_CAST rit->getDiskLUNString());
						xmlNewChild(node3, NULL, BAD_CAST "status", BAD_CAST rit->getDiskStatusString());
						xmlNewChild(node3, NULL, BAD_CAST "size", BAD_CAST rit->getDiskSizeString());
						if (rit->getDiskEnclosureDevName() != NULL)
							xmlNewChild(node3, NULL, BAD_CAST "enclosure", BAD_CAST rit->getDiskEnclosureDevName(true));
						if (rit->getDiskVendorName() != NULL)
							xmlNewChild(node3, NULL, BAD_CAST "vendor", BAD_CAST rit->getDiskVendorName(true));
						if (rit->getDiskSerial() != NULL)
							xmlNewChild(node3, NULL, BAD_CAST "serial_no", BAD_CAST rit->getDiskSerial(true));
					}
				}
			}
		}
	}

	/* NETWORK INFO */

	node1 = xmlNewNode(NULL, BAD_CAST "networking");
	xmlAddChild(root_node, node1);
	xmlNewChild(node1, NULL, BAD_CAST "hostname", BAD_CAST network.getHostName());

	if (network.getDefaultRoutes().size() > 0) {
		vector<struct default_route> routes = network.getDefaultRoutes();
		vector<struct default_route>::iterator it;

		node2 = xmlNewNode(NULL, BAD_CAST "ip_routes");
		xmlAddChild(node1, node2);

		for (it = routes.begin(); it != routes.end(); it++) {
			node3 = xmlNewChild(node2, NULL, BAD_CAST "default", BAD_CAST it->gateway.c_str());
			if (it->family == XSI_RTAF_IPV4)
				xmlNewProp(node3, BAD_CAST "family", BAD_CAST "IPv4");
			else
				xmlNewProp(node3, BAD_CAST "family", BAD_CAST "IPv6");
		}
	}

	if (network.getNetIfStats().size() > 0) {
		vector<XSI_NetIfStat> nsvect = network.getNetIfStats();
		vector<XSI_NetIfStat>::iterator it;

		node2 = xmlNewNode(NULL, BAD_CAST "interfaces");
		xmlAddChild(node1, node2);
		snprintf(buffer, sizeof(buffer), "%" PRIi64, (int64_t)nsvect.size());
		xmlNewChild(node2, NULL, BAD_CAST "count", BAD_CAST buffer);

		for (it = nsvect.begin(); it != nsvect.end(); it++) {
			node3 = xmlNewNode(NULL, BAD_CAST "interface");
			xmlAddChild(node2, node3);
			xmlNewProp(node3, BAD_CAST "name", BAD_CAST it->getIfName(true));
			if (!it->isUp()) {
				xmlNewProp(node3, BAD_CAST "up", BAD_CAST "false");
				continue;
			}
			if (it->hasMetric()) {
				xmlNewChild(node3, NULL, BAD_CAST "metric", BAD_CAST it->getMetricString());
			}
			if (it->hasMTU()) {
				xmlNewChild(node3, NULL, BAD_CAST "mtu", BAD_CAST it->getMTUString());
			}
			if (it->isRunning() || it->isLoopback()
			    || it->isPointToPoint() || it->isPromiscuous()) {
				node4 = xmlNewNode(NULL, BAD_CAST "properties");
				xmlAddChild(node3, node4);
				if (it->isRunning())
					xmlNewChild(node4, NULL, BAD_CAST "running", NULL);
				if (it->isPromiscuous())
					xmlNewChild(node4, NULL, BAD_CAST "promiscuous", NULL);
				if (it->isLoopback())
					xmlNewChild(node4, NULL, BAD_CAST "loopback", NULL);
				if (it->isPointToPoint()) {
					node5 = xmlNewNode(NULL, BAD_CAST "point-to-point");
					xmlAddChild(node4, node5);
					xmlNewProp(node5, BAD_CAST "dest_addr", BAD_CAST it->getPTPDestAddressString());
				}
			}
			if (it->getDescriptionString())
				xmlNewChild(node3, NULL, BAD_CAST "if_descr", BAD_CAST it->getDescriptionString(true));
			if (it->getMediaString())
				xmlNewChild(node3, NULL, BAD_CAST "media", BAD_CAST it->getMediaString(true));
			if (it->hasLinkStatus())
				xmlNewChild(node3, NULL, BAD_CAST "link_state", BAD_CAST it->getLinkStatusString());
			if (it->hasDuplexStatus())
				xmlNewChild(node3, NULL, BAD_CAST "duplex", BAD_CAST it->getDuplexStatusString());
			if (it->isCARPDev()) {
				node4 = xmlNewChild(node3, NULL, BAD_CAST "carp", BAD_CAST it->getCARPStatusString());
				if (it->getCARPDeviceString())
					xmlNewProp(node4, BAD_CAST "carpdev", BAD_CAST it->getCARPDeviceString(true));
				xmlNewProp(node4, BAD_CAST "vhid", BAD_CAST it->getCARPVHIDString());
				xmlNewProp(node4, BAD_CAST "advskew", BAD_CAST it->getCARPAdvskewString());
				xmlNewProp(node4, BAD_CAST "advbase", BAD_CAST it->getCARPAdvbaseString());
				if (it->getCARPPreemptString())
					xmlNewProp(node4, BAD_CAST "preempt", BAD_CAST it->getCARPPreemptString());
				if (it->getCARPArpbalanceString())
					xmlNewProp(node4, BAD_CAST "arpbalance", BAD_CAST it->getCARPArpbalanceString());
			}
			if (it->hasAddresses()) {
				vector<XSI_IfAddress> addrs = it->getIfAddresses();
				vector<XSI_IfAddress>::iterator addrIt;

				node4 = xmlNewNode(NULL, BAD_CAST "addresses");
				xmlAddChild(node3, node4);
				if (it->getLladdrString())
					xmlNewChild(node4, NULL, BAD_CAST "ll_addr", BAD_CAST it->getLladdrString());
				for (addrIt = addrs.begin(); addrIt != addrs.end(); addrIt++) {
					node5 = xmlNewChild(node4, NULL, BAD_CAST "net_addr", BAD_CAST addrIt->getAddressString());
					xmlNewProp(node5, BAD_CAST "family", BAD_CAST addrIt->getAddressFamilyString());
					if (addrIt->hasPrefixLen()) {
						xmlNewProp(node5, BAD_CAST "prefixlen", BAD_CAST addrIt->getPrefixLenString());
					}
				}
			}
			if (!it->isLoopback()) {
				node4 = xmlNewNode(NULL, BAD_CAST "statistics");
				xmlAddChild(node3, node4);

				if (!it->hasTraffic())
					xmlNewProp(node4, BAD_CAST "has_traffic", BAD_CAST "false");
				else {
					node5 = xmlNewNode(NULL, BAD_CAST "packets");
					xmlAddChild(node4, node5);
					xmlNewChild(node5, NULL, BAD_CAST "in", BAD_CAST it->getPacketsInString());
					if (it->getPacketsInPerSecString() && !configuration->oneShot())
						xmlNewChild(node5, NULL, BAD_CAST "in_persec", BAD_CAST it->getPacketsInPerSecString());
					xmlNewChild(node5, NULL, BAD_CAST "out", BAD_CAST it->getPacketsOutString());
					if (it->getPacketsOutPerSecString() && !configuration->oneShot())
						xmlNewChild(node5, NULL, BAD_CAST "out_persec", BAD_CAST it->getPacketsOutPerSecString());
					if (it->hasDroppedPkgs()) {
						xmlNewChild(node5, NULL, BAD_CAST "dropped", BAD_CAST it->getPacketsDroppedString());
						if (it->getPacketsDroppedPerSecString() && !configuration->oneShot())
							xmlNewChild(node5, NULL, BAD_CAST "dropped_persec", BAD_CAST it->getPacketsDroppedPerSecString());
					}

					node5 = xmlNewNode(NULL, BAD_CAST "bytes");
					xmlAddChild(node4, node5);
					xmlNewChild(node5, NULL, BAD_CAST "in", BAD_CAST it->getBytesInString());
					if (it->getBytesInPerSecString() && !configuration->oneShot())
						xmlNewChild(node5, NULL, BAD_CAST "in_persec", BAD_CAST it->getBytesInPerSecString());
					xmlNewChild(node5, NULL, BAD_CAST "out", BAD_CAST it->getBytesOutString());
					if (it->getBytesOutPerSecString() && !configuration->oneShot())
						xmlNewChild(node5, NULL, BAD_CAST "out_persec", BAD_CAST it->getBytesOutPerSecString());
				}

				if (it->hasErrors()) {
					xmlNewChild(node4, NULL, BAD_CAST "errors_in", BAD_CAST it->getErrorsInString());
					if (it->getErrorsInPerSecString())
						xmlNewChild(node4, NULL, BAD_CAST "errors_in_persec", BAD_CAST it->getErrorsInPerSecString());
					xmlNewChild(node4, NULL, BAD_CAST "errors_out", BAD_CAST it->getErrorsOutString());
					if (it->getErrorsOutPerSecString())
						xmlNewChild(node4, NULL, BAD_CAST "errors_out_persec", BAD_CAST it->getErrorsOutPerSecString());
				}
				if (it->hasCollisions()) {
					xmlNewChild(node4, NULL, BAD_CAST "collisions", BAD_CAST it->getCollisionsString());
					if (it->getCollisionsPerSecString())
						xmlNewChild(node4, NULL, BAD_CAST "collisions_persec", BAD_CAST it->getCollisionsPerSecString());
				}
			}
		}
	}

	if (configuration->showPacketFilter()) {
		/*
		 * This is getting quite out of hand down here with
		 * the if-levels ...
		 */
		vector<XSI_PFilter>::iterator pfi;

		for (pfi = packetFilters.begin(); pfi != packetFilters.end(); pfi++) {
			vector<FilteredIfStat> fIfStats;
			vector<FilteredIfStat>::iterator ii;

			if (!pfi->isAvailable())
				continue;

			node2 = xmlNewNode(NULL, BAD_CAST "packet_filter");
			xmlAddChild(node1, node2);
			xmlNewProp(node2, BAD_CAST "type", BAD_CAST pfi->getPFilterTypeString());
			if (!pfi->isEnabled()) {
				xmlNewProp(node2, BAD_CAST "enabled", BAD_CAST "false");
				continue;
			}

			if (pfi->getRuntimeSecString())
				xmlNewChild(node2, NULL, BAD_CAST "runtime", BAD_CAST pfi->getRuntimeSecString());
			if (pfi->getRuleMatchesString())
				xmlNewChild(node2, NULL, BAD_CAST "rule_matches", BAD_CAST pfi->getRuleMatchesString());
			if (pfi->getRuleMatchesAverageString())
				xmlNewChild(node2, NULL, BAD_CAST "rule_matches_avg", BAD_CAST pfi->getRuleMatchesAverageString());
			if (pfi->getRuleMatchesPerSecString() && !configuration->oneShot())
				xmlNewChild(node2, NULL, BAD_CAST "rule_matches_persec", BAD_CAST pfi->getRuleMatchesPerSecString());
			if (pfi->hasStates()) {
				node3 = xmlNewNode(NULL, BAD_CAST "states");
				xmlAddChild(node2, node3);
				xmlNewChild(node3, NULL, BAD_CAST "count", BAD_CAST pfi->getStatesString());
				if (pfi->getStatesLimitString())
					xmlNewChild(node3, NULL, BAD_CAST "limit", BAD_CAST pfi->getStatesLimitString());
				if (pfi->getStateSearchesString()) {
					xmlNewChild(node3, NULL, BAD_CAST "searches", BAD_CAST pfi->getStateSearchesString());
					if (pfi->getStateSearchesAverageString())
						xmlNewChild(node3, NULL, BAD_CAST "searches_avg", BAD_CAST pfi->getStateSearchesAverageString());
					if (pfi->getStateSearchesPerSecString() && !configuration->oneShot())
						xmlNewChild(node3, NULL, BAD_CAST "searches_persec", BAD_CAST pfi->getStateSearchesPerSecString());
				}
				if (pfi->getStateInsertsString()) {
					xmlNewChild(node3, NULL, BAD_CAST "inserts", BAD_CAST pfi->getStateInsertsString());
					if (pfi->getStateInsertsAverageString())
						xmlNewChild(node3, NULL, BAD_CAST "inserts_avg", BAD_CAST pfi->getStateInsertsAverageString());
					if (pfi->getStateInsertsPerSecString() && !configuration->oneShot())
						xmlNewChild(node3, NULL, BAD_CAST "inserts_persec", BAD_CAST pfi->getStateInsertsPerSecString());
				}
				if (pfi->getStateRemovalsString()) {
					xmlNewChild(node3, NULL, BAD_CAST "removals", BAD_CAST pfi->getStateRemovalsString());
					if (pfi->getStateRemovalsAverageString())
						xmlNewChild(node3, NULL, BAD_CAST "removals_avg", BAD_CAST pfi->getStateRemovalsAverageString());
					if (pfi->getStateRemovalsPerSecString() && !configuration->oneShot())
						xmlNewChild(node3, NULL, BAD_CAST "removals_persec", BAD_CAST pfi->getStateRemovalsPerSecString());
				}
				if (pfi->getStateMismatchesString()) {
					xmlNewChild(node3, NULL, BAD_CAST "mismatches", BAD_CAST pfi->getStateMismatchesString());
					if (pfi->getStateMismatchesAverageString())
						xmlNewChild(node3, NULL, BAD_CAST "mismatches_avg", BAD_CAST pfi->getStateMismatchesAverageString());
					if (pfi->getStateMismatchesPerSecString() && !configuration->oneShot())
						xmlNewChild(node3, NULL, BAD_CAST "mismatches_persec", BAD_CAST pfi->getStateMismatchesPerSecString());
				}
			}

			fIfStats = pfi->getFilteredIfStats();
			if (fIfStats.size() == 1 && fIfStats[0].isSummary()) {
				node3 = xmlNewNode(NULL, BAD_CAST "filter_summary");
				xmlAddChild(node2, node3);

				if (!fIfStats[0].hasTraffic()) {
					xmlNewProp(node3, BAD_CAST "has_traffic", BAD_CAST "false");
					continue;
				}

				if (fIfStats[0].v4CountersReturnSummaryTraffic()) {
					node4 = xmlNewNode(NULL, BAD_CAST "ip");
					xmlAddChild(node3, node4);

					node5 = xmlNewNode(NULL, BAD_CAST "packets");
					xmlAddChild(node4, node5);
					xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST fIfStats[0].getPktsV4InBlockedString());
					if (fIfStats[0].getPktsV4InBlockedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST fIfStats[0].getPktsV4InBlockedPerSecString());
					xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST fIfStats[0].getPktsV4InPassedString());
					if (fIfStats[0].getPktsV4InPassedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST fIfStats[0].getPktsV4InPassedPerSecString());
					xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST fIfStats[0].getPktsV4OutBlockedString());
					if (fIfStats[0].getPktsV4OutBlockedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST fIfStats[0].getPktsV4OutBlockedPerSecString());
					xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST fIfStats[0].getPktsV4OutPassedString());
					if (fIfStats[0].getPktsV4OutPassedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST fIfStats[0].getPktsV4OutPassedPerSecString());

					if (fIfStats[0].hasByteCounts()) {
						node5 = xmlNewNode(NULL, BAD_CAST "bytes");
						xmlAddChild(node4, node5);
						xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST fIfStats[0].getBytesV4InBlockedString());
						if (fIfStats[0].getBytesV4InBlockedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST fIfStats[0].getBytesV4InBlockedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST fIfStats[0].getBytesV4InPassedString());
						if (fIfStats[0].getBytesV4InPassedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST fIfStats[0].getBytesV4InPassedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST fIfStats[0].getBytesV4OutBlockedString());
						if (fIfStats[0].getBytesV4OutBlockedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST fIfStats[0].getBytesV4OutBlockedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST fIfStats[0].getBytesV4OutPassedString());
						if (fIfStats[0].getBytesV4OutPassedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST fIfStats[0].getBytesV4OutPassedPerSecString());
					}
				} else {
					if (fIfStats[0].hasIPv4Traffic()) {
						node4 = xmlNewNode(NULL, BAD_CAST "ipv4");
						xmlAddChild(node3, node4);

						node5 = xmlNewNode(NULL, BAD_CAST "packets");
						xmlAddChild(node4, node5);
						xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST fIfStats[0].getPktsV4InBlockedString());
						if (fIfStats[0].getPktsV4InBlockedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST fIfStats[0].getPktsV4InBlockedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST fIfStats[0].getPktsV4InPassedString());
						if (fIfStats[0].getPktsV4InPassedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST fIfStats[0].getPktsV4InPassedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST fIfStats[0].getPktsV4OutBlockedString());
						if (fIfStats[0].getPktsV4OutBlockedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST fIfStats[0].getPktsV4OutBlockedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST fIfStats[0].getPktsV4OutPassedString());
						if (fIfStats[0].getPktsV4OutPassedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST fIfStats[0].getPktsV4OutPassedPerSecString());

						if (fIfStats[0].hasByteCounts()) {
							node5 = xmlNewNode(NULL, BAD_CAST "bytes");
							xmlAddChild(node4, node5);
							xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST fIfStats[0].getBytesV4InBlockedString());
							if (fIfStats[0].getBytesV4InBlockedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST fIfStats[0].getBytesV4InBlockedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST fIfStats[0].getBytesV4InPassedString());
							if (fIfStats[0].getBytesV4InPassedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST fIfStats[0].getBytesV4InPassedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST fIfStats[0].getBytesV4OutBlockedString());
							if (fIfStats[0].getBytesV4OutBlockedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST fIfStats[0].getBytesV4OutBlockedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST fIfStats[0].getBytesV4OutPassedString());
							if (fIfStats[0].getBytesV4OutPassedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST fIfStats[0].getBytesV4OutPassedPerSecString());
						}
					}

					if (fIfStats[0].hasIPv6Traffic()) {
						node4 = xmlNewNode(NULL, BAD_CAST "ipv6");
						xmlAddChild(node3, node4);

						node5 = xmlNewNode(NULL, BAD_CAST "packets");
						xmlAddChild(node4, node5);
						xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST fIfStats[0].getPktsV6InBlockedString());
						if (fIfStats[0].getPktsV6InBlockedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST fIfStats[0].getPktsV6InBlockedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST fIfStats[0].getPktsV6InPassedString());
						if (fIfStats[0].getPktsV6InPassedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST fIfStats[0].getPktsV6InPassedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST fIfStats[0].getPktsV6OutBlockedString());
						if (fIfStats[0].getPktsV6OutBlockedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST fIfStats[0].getPktsV6OutBlockedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST fIfStats[0].getPktsV6OutPassedString());
						if (fIfStats[0].getPktsV6OutPassedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST fIfStats[0].getPktsV6OutPassedPerSecString());

						if (fIfStats[0].hasByteCounts()) {
							node5 = xmlNewNode(NULL, BAD_CAST "bytes");
							xmlAddChild(node4, node5);
							xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST fIfStats[0].getBytesV6InBlockedString());
							if (fIfStats[0].getBytesV6InBlockedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST fIfStats[0].getBytesV6InBlockedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST fIfStats[0].getBytesV6InPassedString());
							if (fIfStats[0].getBytesV6InPassedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST fIfStats[0].getBytesV6InPassedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST fIfStats[0].getBytesV6OutBlockedString());
							if (fIfStats[0].getBytesV6OutBlockedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST fIfStats[0].getBytesV6OutBlockedPerSecString());
							xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST fIfStats[0].getBytesV6OutPassedString());
							if (fIfStats[0].getBytesV6OutPassedPerSecString())
								xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST fIfStats[0].getBytesV6OutPassedPerSecString());
						}
					}
				}

				continue;
			}

			for (ii = fIfStats.begin() ; ii != fIfStats.end(); ii++) {
				node3 = xmlNewNode(NULL, BAD_CAST "filtered_if");
				xmlAddChild(node2, node3);
				xmlNewProp(node3, BAD_CAST "name", BAD_CAST ii->getIfName(true));

				if (!ii->hasTraffic()) {
					xmlNewProp(node3, BAD_CAST "has_traffic", BAD_CAST "false");
					continue;
				}

				if (ii->v4CountersReturnSummaryTraffic()) {
					node4 = xmlNewNode(NULL, BAD_CAST "ip");
					xmlAddChild(node3, node4);

					node5 = xmlNewNode(NULL, BAD_CAST "packets");
					xmlAddChild(node4, node5);
					xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST ii->getPktsV4InBlockedString());
					if (ii->getPktsV4InBlockedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST ii->getPktsV4InBlockedPerSecString());
					xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST ii->getPktsV4InPassedString());
					if (ii->getPktsV4InPassedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST ii->getPktsV4InPassedPerSecString());
					xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST ii->getPktsV4OutBlockedString());
					if (ii->getPktsV4OutBlockedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST ii->getPktsV4OutBlockedPerSecString());
					xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST ii->getPktsV4OutPassedString());
					if (ii->getPktsV4OutPassedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST ii->getPktsV4OutPassedPerSecString());

					if (ii->hasByteCounts()) {
						node5 = xmlNewNode(NULL, BAD_CAST "bytes");
						xmlAddChild(node4, node5);
						xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST ii->getBytesV4InBlockedString());
						if (ii->getBytesV4InBlockedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST ii->getBytesV4InBlockedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST ii->getBytesV4InPassedString());
						if (ii->getBytesV4InPassedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST ii->getBytesV4InPassedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST ii->getBytesV4OutBlockedString());
						if (ii->getBytesV4OutBlockedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST ii->getBytesV4OutBlockedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST ii->getBytesV4OutPassedString());
						if (ii->getBytesV4OutPassedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST ii->getBytesV4OutPassedPerSecString());
					}

					continue;
				}

				if (ii->hasIPv4Traffic()) {
					node4 = xmlNewNode(NULL, BAD_CAST "ipv4");
					xmlAddChild(node3, node4);

					node5 = xmlNewNode(NULL, BAD_CAST "packets");
					xmlAddChild(node4, node5);
					xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST ii->getPktsV4InBlockedString());
					if (ii->getPktsV4InBlockedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST ii->getPktsV4InBlockedPerSecString());
					xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST ii->getPktsV4InPassedString());
					if (ii->getPktsV4InPassedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST ii->getPktsV4InPassedPerSecString());
					xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST ii->getPktsV4OutBlockedString());
					if (ii->getPktsV4OutBlockedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST ii->getPktsV4OutBlockedPerSecString());
					xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST ii->getPktsV4OutPassedString());
					if (ii->getPktsV4OutPassedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST ii->getPktsV4OutPassedPerSecString());

					if (ii->hasByteCounts()) {
						node5 = xmlNewNode(NULL, BAD_CAST "bytes");
						xmlAddChild(node4, node5);
						xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST ii->getBytesV4InBlockedString());
						if (ii->getBytesV4InBlockedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST ii->getBytesV4InBlockedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST ii->getBytesV4InPassedString());
						if (ii->getBytesV4InPassedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST ii->getBytesV4InPassedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST ii->getBytesV4OutBlockedString());
						if (ii->getBytesV4OutBlockedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST ii->getBytesV4OutBlockedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST ii->getBytesV4OutPassedString());
						if (ii->getBytesV4OutPassedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST ii->getBytesV4OutPassedPerSecString());
					}
				}

				if (ii->hasIPv6Traffic()) {
					node4 = xmlNewNode(NULL, BAD_CAST "ipv6");
					xmlAddChild(node3, node4);

					node5 = xmlNewNode(NULL, BAD_CAST "packets");
					xmlAddChild(node4, node5);
					xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST ii->getPktsV6InBlockedString());
					if (ii->getPktsV6InBlockedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST ii->getPktsV6InBlockedPerSecString());
					xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST ii->getPktsV6InPassedString());
					if (ii->getPktsV6InPassedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST ii->getPktsV6InPassedPerSecString());
					xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST ii->getPktsV6OutBlockedString());
					if (ii->getPktsV6OutBlockedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST ii->getPktsV6OutBlockedPerSecString());
					xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST ii->getPktsV6OutPassedString());
					if (ii->getPktsV6OutPassedPerSecString())
						xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST ii->getPktsV6OutPassedPerSecString());

					if (ii->hasByteCounts()) {
						node5 = xmlNewNode(NULL, BAD_CAST "bytes");
						xmlAddChild(node4, node5);
						xmlNewChild(node5, NULL, BAD_CAST "in_blocked", BAD_CAST ii->getBytesV6InBlockedString());
						if (ii->getBytesV6InBlockedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "in_blocked_persec", BAD_CAST ii->getBytesV6InBlockedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "in_passed", BAD_CAST ii->getBytesV6InPassedString());
						if (ii->getBytesV6InPassedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "in_passed_persec", BAD_CAST ii->getBytesV6InPassedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "out_blocked", BAD_CAST ii->getBytesV6OutBlockedString());
						if (ii->getBytesV6OutBlockedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "out_blocked_persec", BAD_CAST ii->getBytesV6OutBlockedPerSecString());
						xmlNewChild(node5, NULL, BAD_CAST "out_passed", BAD_CAST ii->getBytesV6OutPassedString());
						if (ii->getBytesV6OutPassedPerSecString())
							xmlNewChild(node5, NULL, BAD_CAST "out_passed_persec", BAD_CAST ii->getBytesV6OutPassedPerSecString());
					}
				}
			}
		}
	}
}

xmlDocPtr
XSIBase::getSysInfoDocPtr(void)
{
	return (sysInfoDocument);
}

const char *
XSIBase::getUserCacheStatsString(void)
{
	static string	ret;
	char		buf[128];

	if (user_cache == NULL)
		XSI_Log::log()->err(__func__, "Internal error: XMLSysInfo is not properly initialized");

	snprintf(buf, sizeof(buf),
		 "%lu hits  %lu misses  %lu timeouts  %lu failures",
		 user_cache->getHits(), user_cache->getMisses(),
		 user_cache->getTimeouts(), user_cache->getFailures());

	ret = buf;
	return (ret.c_str());
}

void
XSIBase::resetCaches(void)
{
	user_cache->reset();
}

void
XSIBase::updateTimeKeeper(void)
{
	XSI_TimeKeeper::get()->triggerUpdate();
}

/* virtual */
int
XSIBase::tryOpenPrivilegedDevices(void)
{
	return (0);
}

/* virtual */
int
XSIBase::initializePacketfilters(void)
{
	return (0);
}

/* virtual */
void
XSIBase::updatePacketfilterStats(void)
{
	// empty
}

/* virtual */
void
XSIBase::collectCommonInfo(void)
{
	// empty
}

/* virtual */
void
XSIBase::updateCommonStats(void)
{
	// empty
}

/* virtual */
void
XSIBase::collectSysInfo(void)
{
	// empty
}

/* virtual */
void
XSIBase::updateStats(void)
{
	// empty
}
